/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "marshal_types.h"
#include "marshal_util.h"
#include "OptimizedGstBuffer.h"

GstBuffer_t *
get_marshallable_GstBuffer (const GstBuffer * buffer, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GstBuffer_t *tbuffer = NULL;

  tbuffer = (GstBuffer_t *) g_new0 (OptimizedGstBuffer_t, 1);

  INT64_TO_OCTET_STRING (&tbuffer->pts, G_GUINT64_FORMAT,
      GST_BUFFER_PTS (buffer), size, error, err);
  INT64_TO_OCTET_STRING (&tbuffer->dts, G_GUINT64_FORMAT,
      GST_BUFFER_DTS (buffer), size, error, err);
  INT64_TO_OCTET_STRING (&tbuffer->duration, G_GUINT64_FORMAT,
      GST_BUFFER_DURATION (buffer), size, error, err);
  INT64_TO_OCTET_STRING (&tbuffer->offset, G_GUINT64_FORMAT,
      GST_BUFFER_OFFSET (buffer), size, error, err);
  INT64_TO_OCTET_STRING (&tbuffer->offset_end, G_GUINT64_FORMAT,
      GST_BUFFER_OFFSET_END (buffer), size, error, err);

  tbuffer->flags.buf = calloc (2, 1);
  tbuffer->flags.size = 2;

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_LIVE))
    BIT_STRING_SET_FLAG (&tbuffer->flags, GstBufferFlags_gst_buffer_flag_live);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_DECODE_ONLY))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_decode_only);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_DISCONT))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_discont);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_RESYNC))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_resync);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_CORRUPTED))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_corrupted);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_MARKER))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_marker);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_HEADER))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_header);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_GAP))
    BIT_STRING_SET_FLAG (&tbuffer->flags, GstBufferFlags_gst_buffer_flag_gap);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_DROPPABLE))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_droppable);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_DELTA_UNIT))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_delta_unit);

  if (GST_BUFFER_FLAG_IS_SET (buffer, GST_BUFFER_FLAG_TAG_MEMORY))
    BIT_STRING_SET_FLAG (&tbuffer->flags,
        GstBufferFlags_gst_buffer_flag_tag_memory);

  ((OptimizedGstBuffer_t *) (tbuffer))->gstbuffer =
      gst_buffer_ref ((GstBuffer *) buffer);

  return tbuffer;

error:
  if (tbuffer != NULL)
    asn_DEF_OptimizedGstBuffer.free_struct (&asn_DEF_OptimizedGstBuffer,
        tbuffer, 0);

  return NULL;
}

static GstBuffer *
get_GstBuffer (GstBuffer_t * tbuffer, KurentoMarshalRules rules, GError ** err)
{
  GstBuffer *buffer = NULL;

  if (tbuffer->data.size > 0) {
    buffer = gst_buffer_new_wrapped (tbuffer->data.buf, tbuffer->data.size);
    tbuffer->data.buf = NULL;
    tbuffer->data.size = 0;
  } else {
    buffer = gst_buffer_new ();
  }

  GST_BUFFER_PTS (buffer) =
      OCTET_STRING_TO_INT64 (tbuffer->pts.buf, uint64, strtoull);
  GST_BUFFER_DTS (buffer) =
      OCTET_STRING_TO_INT64 (tbuffer->dts.buf, uint64, strtoull);
  GST_BUFFER_DURATION (buffer) =
      OCTET_STRING_TO_INT64 (tbuffer->duration.buf, uint64, strtoull);
  GST_BUFFER_OFFSET (buffer) =
      OCTET_STRING_TO_INT64 (tbuffer->offset.buf, uint64, strtoull);
  GST_BUFFER_OFFSET_END (buffer) =
      OCTET_STRING_TO_INT64 (tbuffer->offset_end.buf, uint64, strtoull);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_live))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_LIVE);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_decode_only))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DECODE_ONLY);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_discont))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DISCONT);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_resync))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_RESYNC);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_corrupted))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_CORRUPTED);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_marker))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_MARKER);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_header))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_HEADER);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_gap))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_GAP);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_droppable))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DROPPABLE);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_delta_unit))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_DELTA_UNIT);

  if (BIT_STRING_FLAG_IS_SET (&tbuffer->flags,
          GstBufferFlags_gst_buffer_flag_tag_memory))
    GST_BUFFER_FLAG_SET (buffer, GST_BUFFER_FLAG_TAG_MEMORY);

  return buffer;
}

gssize
enc_GstBuffer (KurentoMarshalRules rules, const GstBuffer * buffer,
    gchar * buf, gsize size, GError ** err)
{
  GstBuffer_t *tbuffer = NULL;
  gssize s;

  tbuffer = get_marshallable_GstBuffer (buffer, rules, buf, size, err);

  if (tbuffer == NULL)
    return -1;

  s = ENCODE_MARSHAL_TYPE (OptimizedGstBuffer, tbuffer, rules, buf, size, err);

  if (tbuffer != NULL)
    asn_DEF_OptimizedGstBuffer.free_struct (&asn_DEF_OptimizedGstBuffer,
        tbuffer, 0);

  return s;
}

gboolean
dec_GstBuffer (KurentoMarshalRules rules, const gchar * buf, gsize size,
    GstBuffer ** buffer, GError ** err)
{
  GstBuffer_t *tbuffer = NULL;
  asn_dec_rval_t rval;
  gboolean ret;

  rval = decode_from_buffer (NULL, &asn_DEF_GstBuffer, (void **) &tbuffer, buf,
      size, rules);

  if (!(ret = rval.code == RC_OK)) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "Could not decode GstBuffer");
    goto end;
  }

  *buffer = get_GstBuffer (tbuffer, rules, err);
  ret = (*buffer != NULL);

end:
  if (tbuffer != NULL)
    asn_DEF_GstBuffer.free_struct (&asn_DEF_GstBuffer, tbuffer, 0);

  return ret;
}
