/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "marshal_types.h"
#include "marshal_util.h"
#include "GstCap.h"

GstCaps_t *
get_marshallable_GstCaps (const GstCaps * caps, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GstCaps_t *tcaps;
  guint len, i;

  tcaps = g_new0 (GstCaps_t, 1);

  len = gst_caps_get_size (caps);
  for (i = 0; i < len; i++) {
    GstCapsFeatures *feature;
    GstStructure *structure;
    GstStructure_t *tstruct;
    GstCap_t *cap;

    cap = g_new0 (GstCap_t, 1);

    structure = gst_caps_get_structure (caps, i);
    tstruct = get_marshallable_gststructure (structure, rules, buf, size, err);
    if (tstruct == NULL) {
      asn_DEF_GstStructure.free_struct (&asn_DEF_GstStructure, tstruct, 0);
      break;
    }

    own_allocated_memory (tstruct, &cap->structure);
    asn_DEF_GstStructure.free_struct (&asn_DEF_GstStructure, tstruct, 0);

    feature = gst_caps_get_features (caps, i);

    if (feature != NULL) {
      gchar *string;

      string = gst_caps_features_to_string (feature);
      cap->feature = g_new0 (GstCapsFeatures_t, 1);

      if (OCTET_STRING_fromString (cap->feature, string) < 0) {
        g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
            "Could not encode feature: %s", string);
        asn_DEF_GstCap.free_struct (&asn_DEF_GstCap, cap, 0);
        g_free (string);
        break;
      }

      g_free (string);
    }

    ASN_SEQUENCE_ADD (tcaps, cap);
  }

  if (*err != NULL) {
    asn_DEF_GstCaps.free_struct (&asn_DEF_GstCaps, tcaps, 0);
    tcaps = NULL;
  }

  return tcaps;
}

static GstCaps *
get_GstCaps (const GstCaps_t * tcaps, KurentoMarshalRules rules, GError ** err)
{
  GstCaps *caps = NULL;
  gint i;

  caps = gst_caps_new_empty ();

  for (i = 0; i < tcaps->list.count; i++) {
    GstStructure *structure;
    const GstCap_t *cap;

    cap = tcaps->list.array[i];
    structure = get_GstStructure (&cap->structure, rules, err);

    if (structure == NULL)
      break;

    if (cap->feature != NULL) {
      GstCapsFeatures *features;

      features = gst_caps_features_from_string ((gchar *) cap->feature->buf);
      if (features == NULL) {
        g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
            "Could not create GstCapsFeatures from string: %s",
            (gchar *) cap->feature->buf);
        gst_structure_free (structure);
        break;
      }

      gst_caps_append_structure_full (caps, structure, features);
    } else {
      gst_caps_append_structure (caps, structure);
    }
  }

  if (*err != NULL) {
    gst_caps_unref (caps);
    caps = NULL;
  }

  return caps;
}

gssize
enc_GstCaps (KurentoMarshalRules rules, const GstCaps * caps, gchar * buf,
    gsize size, GError ** err)
{
  GstCaps_t *tcaps;
  gssize s;

  tcaps = get_marshallable_GstCaps (caps, rules, buf, size, err);
  if (tcaps == NULL)
    return -1;

  s = ENCODE_MARSHAL_TYPE (GstCaps, tcaps, rules, buf, size, err);
  asn_DEF_GstCaps.free_struct (&asn_DEF_GstCaps, tcaps, 0);

  return s;
}

gboolean
dec_GstCaps (KurentoMarshalRules rules, const gchar * buf, gsize size,
    GstCaps ** caps, GError ** err)
{
  GstCaps_t *tcaps = NULL;
  asn_dec_rval_t rval;
  gboolean ret;

  rval =
      decode_from_buffer (NULL, &asn_DEF_GstCaps, (void **) &tcaps, buf,
      size, rules);

  if (!(ret = rval.code == RC_OK)) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "Could not decode GstCaps");
    goto end;
  }

  *caps = get_GstCaps (tcaps, rules, err);
  ret = (*caps != NULL);

end:
  if (tcaps != NULL)
    asn_DEF_GstCaps.free_struct (&asn_DEF_GstCaps, tcaps, 0);

  return ret;
}
