/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "marshal_types.h"
#include "marshal_util.h"
#include "GstEvent.h"

static GstEventType
get_event_type (const GstEvent_t * event)
{
  switch (event->event_type) {
    case GstEventType_gst_event_flush_start:
      return GST_EVENT_FLUSH_START;
    case GstEventType_gst_event_flush_stop:
      return GST_EVENT_FLUSH_STOP;
    case GstEventType_gst_event_stream_start:
      return GST_EVENT_STREAM_START;
    case GstEventType_gst_event_caps:
      return GST_EVENT_CAPS;
    case GstEventType_gst_event_segment:
      return GST_EVENT_SEGMENT;
    case GstEventType_gst_event_tag:
      return GST_EVENT_TAG;
    case GstEventType_gst_event_buffersize:
      return GST_EVENT_BUFFERSIZE;
    case GstEventType_gst_event_message:
      return GST_EVENT_SINK_MESSAGE;
    case GstEventType_gst_event_eos:
      return GST_EVENT_EOS;
    case GstEventType_gst_event_toc:
      return GST_EVENT_TOC;
    case GstEventType_gst_event_segment_done:
      return GST_EVENT_SEGMENT_DONE;
    case GstEventType_gst_event_gap:
      return GST_EVENT_GAP;
    case GstEventType_gst_event_qos:
      return GST_EVENT_QOS;
    case GstEventType_gst_event_seek:
      return GST_EVENT_SEEK;
    case GstEventType_gst_event_navigation:
      return GST_EVENT_NAVIGATION;
    case GstEventType_gst_event_latency:
      return GST_EVENT_LATENCY;
    case GstEventType_gst_event_step:
      return GST_EVENT_STEP;
    case GstEventType_gst_event_reconfigure:
      return GST_EVENT_RECONFIGURE;
    case GstEventType_gst_event_toc_select:
      return GST_EVENT_TOC_SELECT;
    case GstEventType_gst_event_custom_upstream:
      return GST_EVENT_CUSTOM_UPSTREAM;
    case GstEventType_gst_event_custom_downstream:
      return GST_EVENT_CUSTOM_DOWNSTREAM;
    case GstEventType_gst_event_custom_downstream_oob:
      return GST_EVENT_CUSTOM_DOWNSTREAM_OOB;
    case GstEventType_gst_event_custom_downstream_sticky:
      return GST_EVENT_CUSTOM_DOWNSTREAM_STICKY;
    case GstEventType_gst_event_custom_both:
      return GST_EVENT_CUSTOM_BOTH;
    case GstEventType_gst_event_custom_both_oob:
      return GST_EVENT_CUSTOM_BOTH_OOB;
    default:
      return GST_EVENT_UNKNOWN;
  }
}

static GstEventType_t
get_marshallable_event_type (const GstEvent * event)
{
  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_FLUSH_START:
      return GstEventType_gst_event_flush_start;
    case GST_EVENT_FLUSH_STOP:
      return GstEventType_gst_event_flush_stop;
    case GST_EVENT_STREAM_START:
      return GstEventType_gst_event_stream_start;
    case GST_EVENT_CAPS:
      return GstEventType_gst_event_caps;
    case GST_EVENT_SEGMENT:
      return GstEventType_gst_event_segment;
    case GST_EVENT_TAG:
      return GstEventType_gst_event_tag;
    case GST_EVENT_BUFFERSIZE:
      return GstEventType_gst_event_buffersize;
    case GST_EVENT_SINK_MESSAGE:
      return GstEventType_gst_event_message;
    case GST_EVENT_EOS:
      return GstEventType_gst_event_eos;
    case GST_EVENT_TOC:
      return GstEventType_gst_event_toc;
    case GST_EVENT_SEGMENT_DONE:
      return GstEventType_gst_event_segment_done;
    case GST_EVENT_GAP:
      return GstEventType_gst_event_gap;
    case GST_EVENT_QOS:
      return GstEventType_gst_event_qos;
    case GST_EVENT_SEEK:
      return GstEventType_gst_event_seek;
    case GST_EVENT_NAVIGATION:
      return GstEventType_gst_event_navigation;
    case GST_EVENT_LATENCY:
      return GstEventType_gst_event_latency;
    case GST_EVENT_STEP:
      return GstEventType_gst_event_step;
    case GST_EVENT_RECONFIGURE:
      return GstEventType_gst_event_reconfigure;
    case GST_EVENT_TOC_SELECT:
      return GstEventType_gst_event_toc_select;
    case GST_EVENT_CUSTOM_UPSTREAM:
      return GstEventType_gst_event_custom_upstream;
    case GST_EVENT_CUSTOM_DOWNSTREAM:
      return GstEventType_gst_event_custom_downstream;
    case GST_EVENT_CUSTOM_DOWNSTREAM_OOB:
      return GstEventType_gst_event_custom_downstream_oob;
    case GST_EVENT_CUSTOM_DOWNSTREAM_STICKY:
      return GstEventType_gst_event_custom_downstream_sticky;
    case GST_EVENT_CUSTOM_BOTH:
      return GstEventType_gst_event_custom_both;
    case GST_EVENT_CUSTOM_BOTH_OOB:
      return GstEventType_gst_event_custom_both_oob;
    default:
      return GstEventType_gst_event_unknown;
  }
}

static GstEvent *
get_GstEvent (GstEvent_t * tevent, KurentoMarshalRules rules, GError ** err)
{
  GstStructure *structure = NULL;
  GstEventType type;
  GstEvent *event;

  if (tevent->structure != NULL) {
    structure = get_GstStructure (tevent->structure, rules, err);

    if (structure == NULL)
      return NULL;
  }

  type = get_event_type (tevent);
  event = gst_event_new_custom (type, structure);

  GST_EVENT_TIMESTAMP (event) = tevent->timestamp;

  return event;
}

gssize
enc_GstEvent (KurentoMarshalRules rules, const GstEvent * event, gchar * buf,
    gsize size, GError ** err)
{
  GstEvent_t *tevent = NULL;
  const GstStructure *structure;
  gssize s = -1;

  tevent = g_new0 (GstEvent_t, 1);
  tevent->event_type = get_marshallable_event_type (event);
  tevent->timestamp = GST_EVENT_TIMESTAMP (event);
  tevent->seqnum = GST_EVENT_SEQNUM (event);

  structure = gst_event_get_structure ((GstEvent *) event);
  if (structure != NULL) {
    tevent->structure = get_marshallable_gststructure (structure, rules, buf,
        size, err);
    if (tevent->structure == NULL)
      goto end;
  }

  s = ENCODE_MARSHAL_TYPE (GstEvent, tevent, rules, buf, size, err);

end:
  if (tevent != NULL)
    asn_DEF_GValue.free_struct (&asn_DEF_GstEvent, tevent, 0);

  return s;
}

gboolean
dec_GstEvent (KurentoMarshalRules rules, const gchar * buf, gsize size,
    GstEvent ** event, GError ** err)
{
  GstEvent_t *tevent = NULL;
  asn_dec_rval_t rval;
  gboolean ret;

  rval = decode_from_buffer (NULL, &asn_DEF_GstEvent, (void **) &tevent, buf,
      size, rules);

  if (!(ret = rval.code == RC_OK)) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "Could not decode GstEvent");
    goto end;
  }

  *event = get_GstEvent (tevent, rules, err);
  ret = (*event != NULL);

end:
  if (tevent != NULL)
    asn_DEF_GstEvent.free_struct (&asn_DEF_GstEvent, tevent, 0);

  return ret;
}
