/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "marshal_types.h"
#include "marshal_util.h"
#include "GstQuery.h"

static GstQueryType
get_query_type (const GstQuery_t * query)
{
  switch (query->query_type) {
    case GstQueryType_gst_query_position:
      return GST_QUERY_POSITION;
    case GstQueryType_gst_query_duration:
      return GST_QUERY_DURATION;
    case GstQueryType_gst_query_latency:
      return GST_QUERY_LATENCY;
    case GstQueryType_gst_query_jitter:
      return GST_QUERY_JITTER;
    case GstQueryType_gst_query_rate:
      return GST_QUERY_RATE;
    case GstQueryType_gst_query_seeking:
      return GST_QUERY_SEEKING;
    case GstQueryType_gst_query_segment:
      return GST_QUERY_SEGMENT;
    case GstQueryType_gst_query_convert:
      return GST_QUERY_CONVERT;
    case GstQueryType_gst_query_formats:
      return GST_QUERY_FORMATS;
    case GstQueryType_gst_query_buffering:
      return GST_QUERY_BUFFERING;
    case GstQueryType_gst_query_custom:
      return GST_QUERY_CUSTOM;
    case GstQueryType_gst_query_uri:
      return GST_QUERY_URI;
    case GstQueryType_gst_query_allocation:
      return GST_QUERY_ALLOCATION;
    case GstQueryType_gst_query_scheduling:
      return GST_QUERY_SCHEDULING;
    case GstQueryType_gst_query_accept_caps:
      return GST_QUERY_ACCEPT_CAPS;
    case GstQueryType_gst_query_caps:
      return GST_QUERY_CAPS;
    case GstQueryType_gst_query_drain:
      return GST_QUERY_DRAIN;
    case GstQueryType_gst_query_context:
      return GST_QUERY_CONTEXT;
    default:
      return GST_QUERY_UNKNOWN;
  }
}

static GstQueryType_t
get_marshallable_query_type (const GstQuery * query)
{
  switch (GST_QUERY_TYPE (query)) {
    case GST_QUERY_POSITION:
      return GstQueryType_gst_query_position;
    case GST_QUERY_DURATION:
      return GstQueryType_gst_query_duration;
    case GST_QUERY_LATENCY:
      return GstQueryType_gst_query_latency;
    case GST_QUERY_JITTER:
      return GstQueryType_gst_query_jitter;
    case GST_QUERY_RATE:
      return GstQueryType_gst_query_rate;
    case GST_QUERY_SEEKING:
      return GstQueryType_gst_query_seeking;
    case GST_QUERY_SEGMENT:
      return GstQueryType_gst_query_segment;
    case GST_QUERY_CONVERT:
      return GstQueryType_gst_query_convert;
    case GST_QUERY_FORMATS:
      return GstQueryType_gst_query_formats;
    case GST_QUERY_BUFFERING:
      return GstQueryType_gst_query_buffering;
    case GST_QUERY_CUSTOM:
      return GstQueryType_gst_query_custom;
    case GST_QUERY_URI:
      return GstQueryType_gst_query_uri;
    case GST_QUERY_ALLOCATION:
      return GstQueryType_gst_query_allocation;
    case GST_QUERY_SCHEDULING:
      return GstQueryType_gst_query_scheduling;
    case GST_QUERY_ACCEPT_CAPS:
      return GstQueryType_gst_query_accept_caps;
    case GST_QUERY_CAPS:
      return GstQueryType_gst_query_caps;
    case GST_QUERY_DRAIN:
      return GstQueryType_gst_query_drain;
    case GST_QUERY_CONTEXT:
      return GstQueryType_gst_query_context;
    default:
      return GstQueryType_gst_query_unknown;
  }
}

static GstQuery *
get_GstQuery (GstQuery_t * tquery, KurentoMarshalRules rules, GError ** err)
{
  GstStructure *structure = NULL;
  GstQueryType type;
  GstQuery *query;

  structure = get_GstStructure (&tquery->structure, rules, err);

  if (structure == NULL)
    return NULL;

  type = get_query_type (tquery);
  query = gst_query_new_custom (type, structure);

  return query;
}

gssize
enc_GstQuery (KurentoMarshalRules rules, const GstQuery * query, gchar * buf,
    gsize size, GError ** err)
{
  GstQuery_t *tquery = NULL;
  const GstStructure *structure;
  GstStructure_t *tstructure;
  gssize s = -1;

  structure = gst_query_get_structure ((GstQuery *) query);
  if (structure == NULL) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "Could not encode GstQuery. No structure found");
    goto end;
  }

  tstructure = get_marshallable_gststructure (structure, rules, buf, size, err);
  if (tstructure == NULL)
    goto end;

  tquery = g_new0 (GstQuery_t, 1);
  tquery->query_type = get_marshallable_query_type (query);
  own_allocated_memory (tstructure, &tquery->structure);
  asn_DEF_GstStructure.free_struct (&asn_DEF_GstStructure, tstructure, 0);

  s = ENCODE_MARSHAL_TYPE (GstQuery, tquery, rules, buf, size, err);

end:
  if (tquery != NULL)
    asn_DEF_GstQuery.free_struct (&asn_DEF_GstQuery, tquery, 0);

  return s;
}

gboolean
dec_GstQuery (KurentoMarshalRules rules, const gchar * buf, gsize size,
    GstQuery ** query, GError ** err)
{
  GstQuery_t *tquery = NULL;
  asn_dec_rval_t rval;
  gboolean ret;

  rval = decode_from_buffer (NULL, &asn_DEF_GstQuery, (void **) &tquery, buf,
      size, rules);

  if (!(ret = rval.code == RC_OK)) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "Could not decode GstQuery");
    goto end;
  }

  *query = get_GstQuery (tquery, rules, err);
  ret = (*query != NULL);

end:
  if (tquery != NULL)
    asn_DEF_GstQuery.free_struct (&asn_DEF_GstQuery, tquery, 0);

  return ret;
}
