/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "marshal_types.h"
#include "marshal_util.h"

#include "GstStructure.h"
#include "GstStructureField.h"

static void
fill_marshallable_structure_fields (const GstStructure * structure,
    GstStructure_t * tstruct, KurentoMarshalRules rules, gchar * buf,
    gsize size, GError ** err)
{
  gint len, i;

  len = gst_structure_n_fields (structure);

  for (i = 0; i < len; i++) {
    GstStructureField_t *field;
    const gchar *fieldname;
    const GValue *gvalue;
    GValue_t *tvalue;

    field = g_new0 (GstStructureField_t, 1);
    fieldname = gst_structure_nth_field_name (structure, i);

    if (OCTET_STRING_fromString (&field->name, fieldname) < 0) {
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
          "Could not encode field name string: (%s)", fieldname);
      asn_DEF_GstStructureField.free_struct (&asn_DEF_GstStructureField, field,
          0);
      break;
    }

    gvalue = gst_structure_get_value (structure, fieldname);
    tvalue = get_marshallable_GValue (gvalue, rules, buf, size, err);

    if (*err != NULL) {
      asn_DEF_GstStructureField.free_struct (&asn_DEF_GstStructureField, field,
          0);
      asn_DEF_GValue.free_struct (&asn_DEF_GValue, tvalue, 0);
      break;
    }

    field->value.value_type = tvalue->value_type;

    if (tvalue->value != NULL) {
      field->value.value = g_new0 (OCTET_STRING_t, 1);
      field->value.value->buf = tvalue->value->buf;
      field->value.value->size = tvalue->value->size;
      tvalue->value->buf = NULL;
      tvalue->value->size = 0;
    }

    asn_DEF_GValue.free_struct (&asn_DEF_GValue, tvalue, 0);

    ASN_SEQUENCE_ADD (&tstruct->fields, field);
  }
}

static void
fill_structure_fields (const GstStructure_t * tstruct, GstStructure * structure,
    KurentoMarshalRules rules, GError ** err)
{
  gint i;

  for (i = 0; i < tstruct->fields.list.count; i++) {
    const GstStructureField_t *field;
    const gchar *fieldname;
    GValue *value;

    field = tstruct->fields.list.array[i];
    fieldname = (gchar *) field->name.buf;
    value = get_GValue (&field->value, rules, err);

    if (value == NULL)
      return;

    gst_structure_set_value (structure, fieldname, value);

    g_value_unset (value);
    g_slice_free (GValue, value);
  }
}

GstStructure_t *
get_marshallable_gststructure (const GstStructure * structure,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstStructure_t *tstruct = NULL;
  const gchar *string;
  gsize s = 0;

  string = gst_structure_get_name (structure);
  if (string == NULL) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "GstStructure does not have a name");
    return NULL;
  }

  if ((s = strlen (string)) > size) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "GstStructure name exceeds buffer size");
    return NULL;
  }

  tstruct = g_new0 (GstStructure_t, 1);

  if (OCTET_STRING_fromString (&tstruct->name, string) < 0) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "Could not encode string value");
  } else {
    fill_marshallable_structure_fields (structure, tstruct, rules, buf, size,
        err);
  }

  if (*err != NULL) {
    asn_DEF_GstStructure.free_struct (&asn_DEF_GstStructure, tstruct, 0);
    tstruct = NULL;
  }

  return tstruct;
}

GstStructure *
get_GstStructure (const GstStructure_t * tstruct, KurentoMarshalRules rules,
    GError ** err)
{
  GstStructure *structure = NULL;
  gchar *name;

  name = (gchar *) tstruct->name.buf;
  structure = gst_structure_new_empty (name);

  fill_structure_fields (tstruct, structure, rules, err);

  if (*err != NULL) {
    gst_structure_free (structure);
    structure = NULL;
  }

  return structure;
}

gssize
enc_GstStructure (KurentoMarshalRules rules, const GstStructure * structure,
    gchar * buf, gsize size, GError ** err)
{
  GstStructure_t *tstruct = NULL;
  gssize s;

  tstruct = get_marshallable_gststructure (structure, rules, buf, size, err);
  if (tstruct == NULL)
    return -1;

  s = ENCODE_MARSHAL_TYPE (GstStructure, tstruct, rules, buf, size, err);
  asn_DEF_GstStructure.free_struct (&asn_DEF_GstStructure, tstruct, 0);

  return s;
}

gboolean
dec_GstStructure (KurentoMarshalRules rules, const gchar * buf, gsize size,
    GstStructure ** structure, GError ** err)
{
  GstStructure_t *tstruct = NULL;
  asn_dec_rval_t rval;
  gboolean ret;

  rval =
      decode_from_buffer (NULL, &asn_DEF_GstStructure, (void **) &tstruct, buf,
      size, rules);

  if (!(ret = rval.code == RC_OK)) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "Could not decode GstStructure");
    goto end;
  }

  *structure = get_GstStructure (tstruct, rules, err);
  ret = (*structure != NULL);

end:
  if (tstruct != NULL)
    asn_DEF_GstStructure.free_struct (&asn_DEF_GstStructure, tstruct, 0);

  return ret;
}
