/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "marshal_types.h"
#include "marshal_util.h"
#include "GTypeInt.h"
#include "GTypeInt64.h"
#include "GTypeUInt64.h"
#include "GTypeBoolean.h"
#include "GTypeFloat.h"
#include "GstTypeSegment.h"
#include "GTypeString.h"
#include "GstTypeFormat.h"
#include "GstTypeFraction.h"
#include "GstTypeIntegerRange.h"
#include "GstTypeDoubleRange.h"
#include "GstTypeFractionRange.h"
#include "GstTypeArray.h"
#include "GstTypeList.h"
#include "GstTypeStreamFlags.h"
#include "GstTypeQOSType.h"
#include "GTypeDouble.h"
#include "GstTypeSeekFlags.h"
#include "GstTypeSeekType.h"
#include "GTypeDate.h"
#include "GstTypeTag.h"
#include "GstTypeTagList.h"
#include "GstTypeDateTime.h"
#include "GstToc.h"
#include "GstTocEntry.h"
#include "GstTocEntryType.h"
#include "GstTocScope.h"
#include "GstBufferingMode.h"
#include "GstSchedulingFlags.h"
#include "GstContext.h"
#include "GstBuffer.h"
#include "OptimizedGstBuffer.h"

static GstTocEntry_t *get_marshallable_entry (const GstTocEntry *,
    KurentoMarshalRules, gchar *, gsize, GError **);

static GstTocEntry *get_entry (const GstTocEntry_t *, KurentoMarshalRules,
    GError **);

#define ENCODE_GVALUE(value, rules, type, buffer, size, err) ({            \
  GValue_t *__val = NULL;                                                  \
  gssize __size;                                                           \
  __val = get_marshallable_##type##_gvalue (value, rules, buf, size, err); \
  if (__val == NULL) {                                                     \
    __size = -1;                                                           \
  } else {                                                                 \
    __size = ENCODE_MARSHAL_TYPE (GValue, __val, rules, buf, size, err);   \
    asn_DEF_GValue.free_struct (&asn_DEF_GValue, __val, 0);                \
  }                                                                        \
  __size;                                                                  \
})

#define SET_MARSHAL_FORMAT(format, ptr) ({        \
  switch (format) {                               \
    case GST_FORMAT_UNDEFINED:                    \
      *ptr = GstTypeFormat_gst_format_undefined;  \
      break;                                      \
    case GST_FORMAT_DEFAULT:                      \
      *ptr = GstTypeFormat_gst_format_default;    \
      break;                                      \
    case GST_FORMAT_BYTES:                        \
      *ptr = GstTypeFormat_gst_format_bytes;      \
      break;                                      \
    case GST_FORMAT_TIME:                         \
      *ptr = GstTypeFormat_gst_format_time;       \
      break;                                      \
    case GST_FORMAT_BUFFERS:                      \
      *ptr = GstTypeFormat_gst_format_buffers;    \
      break;                                      \
    case GST_FORMAT_PERCENT:                      \
      *ptr = GstTypeFormat_gst_format_percent;    \
      break;                                      \
  }                                               \
})

#define SET_FORMAT(format, ptr, err, err_label) ({                    \
  switch (format) {                                                   \
    case GstTypeFormat_gst_format_undefined:                          \
      *ptr = GST_FORMAT_UNDEFINED;                                    \
      break;                                                          \
    case GstTypeFormat_gst_format_default:                            \
      *ptr = GST_FORMAT_DEFAULT;                                      \
      break;                                                          \
    case GstTypeFormat_gst_format_bytes:                              \
      *ptr = GST_FORMAT_BYTES;                                        \
      break;                                                          \
    case GstTypeFormat_gst_format_time:                               \
      *ptr = GST_FORMAT_TIME;                                         \
      break;                                                          \
    case GstTypeFormat_gst_format_buffers:                            \
      *ptr = GST_FORMAT_BUFFERS;                                      \
      break;                                                          \
    case GstTypeFormat_gst_format_percent:                            \
      *ptr = GST_FORMAT_PERCENT;                                      \
      break;                                                          \
    default:                                                          \
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR, \
        "Decoded invalid format (%ld)", format);                      \
      goto err_label;                                                 \
  }                                                                   \
})

#define NEW_MARSHALLABLE(desc_ptr, ptr, gtype, buf, size, rules, err)        \
({                                                                           \
  asn_enc_rval_t ret;                                                        \
  GValue_t *val = NULL;                                                      \
                                                                             \
  if (ptr == NULL) {                                                         \
    val = g_new0 (GValue_t, 1);                                              \
    val->value_type = gtype;                                                  \
  } else {                                                                   \
    ret = encode_to_buffer (&desc_ptr, ptr, buf, size, rules);               \
    if (ret.encoded == -1) {                                                 \
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,        \
        "Could not encode %s", desc_ptr.name);                               \
    } else {                                                                 \
      gssize s;                                                              \
                                                                             \
      s = CALCULATE_SIZE (ret, rules);                                       \
      val = create_marshallable_gvalue (gtype, buf, s, rules, err);          \
    }                                                                        \
  }                                                                          \
                                                                             \
  if (ptr != NULL)                                                           \
    desc_ptr.free_struct (&desc_ptr, ptr, 0);                                \
                                                                             \
  val;                                                                       \
})

static void
decode_marshallable_gvalue (struct asn_TYPE_descriptor_s *type_descriptor,
    const GValue_t * val, KurentoMarshalRules rules, void **struct_ptr,
    GError ** err)
{
  asn_dec_rval_t rval;

  if (val->value == NULL)
    return;

  rval = decode_from_buffer (NULL, type_descriptor, struct_ptr, val->value->buf,
      val->value->size, rules);

  if (rval.code != RC_OK) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "could not decode %s", type_descriptor->name);
  }
}

static GValue_t *
create_marshallable_gvalue (GType_t type, gchar * buff, gsize s,
    KurentoMarshalRules rules, GError ** err)
{
  GValue_t *val = NULL;

  val = g_new0 (GValue_t, 1);
  val->value_type = type;
  val->value = g_new0 (OCTET_STRING_t, 1);

  if (OCTET_STRING_fromBuf (val->value, buff, s) < 0) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "Could not encode opaque data in GValue");
    asn_DEF_GValue.free_struct (&asn_DEF_GValue, val, 0);
    val = NULL;
  }

  return val;
}

static GValue_t *
get_marshallable_int_gvalue (const GValue * value, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GTypeInt_t *i;

  i = g_new0 (GTypeInt_t, 1);
  *i = g_value_get_int (value);

  return NEW_MARSHALLABLE (asn_DEF_GTypeInt, i, GType_g_type_int, buf,
      size, rules, err);
}

static GValue_t *
get_marshallable_uint_gvalue (const GValue * value, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GTypeInt_t *i;

  i = g_new0 (GTypeInt_t, 1);
  *i = g_value_get_uint (value);

  return NEW_MARSHALLABLE (asn_DEF_GTypeInt, i, GType_g_type_uint, buf,
      size, rules, err);
}

static GValue_t *
get_marshallable_date_gvalue (const GValue * value, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GTypeDate_t *tdate = NULL;
  GDate *date;

  date = g_value_get_boxed (value);

  if (date != NULL) {
    tdate = g_new0 (GTypeDate_t, 1);
    *tdate = g_date_get_julian (date);
  }

  return NEW_MARSHALLABLE (asn_DEF_GTypeDate, tdate, GType_g_type_date, buf,
      size, rules, err);
}

static GValue_t *
get_marshallable_boolean_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GTypeBoolean_t *b;

  b = g_new0 (GTypeBoolean_t, 1);
  *b = g_value_get_boolean (value);

  return NEW_MARSHALLABLE (asn_DEF_GTypeBoolean, b, GType_g_type_boolean,
      buf, size, rules, err);
}

static GValue_t *
get_marshallable_format_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeFormat_t *tformat = NULL;
  GstFormat format;

  format = (GstFormat) g_value_get_enum (value);

  tformat = g_new0 (GstTypeFormat_t, 1);

  SET_MARSHAL_FORMAT (format, tformat);

  return NEW_MARSHALLABLE (asn_DEF_GstTypeFormat, tformat,
      GType_gst_type_format, buf, size, rules, err);
}

static GValue_t *
get_marshallable_qos_type_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeQOSType_t *tqos = NULL;
  GstQOSType qos;

  qos = (GstFormat) g_value_get_enum (value);

  tqos = g_new0 (GstTypeQOSType_t, 1);

  switch (qos) {
    case GST_QOS_TYPE_OVERFLOW:
      *tqos = GstTypeQOSType_gst_qos_type_overflow;
      break;
    case GST_QOS_TYPE_UNDERFLOW:
      *tqos = GstTypeQOSType_gst_qos_type_underflow;
      break;
    case GST_QOS_TYPE_THROTTLE:
      *tqos = GstTypeQOSType_gst_qos_type_throttle;
      break;
  }

  return NEW_MARSHALLABLE (asn_DEF_GstTypeQOSType, tqos,
      GType_gst_type_qos_type, buf, size, rules, err);
}

static GValue_t *
get_marshallable_stream_flags_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeStreamFlags_t *tflags = NULL;
  GstStreamFlags flags;

  flags = (GstStreamFlags) g_value_get_flags (value);

  tflags = g_new0 (GstTypeStreamFlags_t, 1);
  tflags->buf = calloc (1, 1);
  tflags->size = 1;
  /* Trim unused bits */
  tflags->bits_unused = 4;

  if (flags & GST_STREAM_FLAG_NONE)
    BIT_STRING_SET_FLAG (tflags, GstTypeStreamFlags_gst_stream_flag_none);

  if (flags & GST_STREAM_FLAG_SPARSE)
    BIT_STRING_SET_FLAG (tflags, GstTypeStreamFlags_gst_stream_flag_sparse);

  if (flags & GST_STREAM_FLAG_SELECT)
    BIT_STRING_SET_FLAG (tflags, GstTypeStreamFlags_gst_stream_flag_select);

  if (flags & GST_STREAM_FLAG_UNSELECT)
    BIT_STRING_SET_FLAG (tflags, GstTypeStreamFlags_gst_stream_flagunselect);

  return NEW_MARSHALLABLE (asn_DEF_GstTypeStreamFlags, tflags,
      GType_gst_type_streaming_flags, buf, size, rules, err);
}

static GValue_t *
get_marshallable_seek_flags_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeSeekFlags_t *tflags;
  GstSeekFlags flags;

  flags = (GstSeekFlags) g_value_get_flags (value);

  tflags = g_new0 (GstTypeSeekFlags_t, 1);
  tflags->buf = calloc (1, 1);
  tflags->size = 1;

  if (flags & GST_SEEK_FLAG_NONE)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_none);

  if (flags & GST_SEEK_FLAG_FLUSH)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_flush);

  if (flags & GST_SEEK_FLAG_ACCURATE)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_accurate);

  if (flags & GST_SEEK_FLAG_KEY_UNIT)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_key_unit);

  if (flags & GST_SEEK_FLAG_SEGMENT)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_segment);

  if (flags & GST_SEEK_FLAG_SKIP)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_skip);

  if (flags & GST_SEEK_FLAG_SNAP_BEFORE)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_snap_before);

  if (flags & GST_SEEK_FLAG_SNAP_AFTER)
    BIT_STRING_SET_FLAG (tflags, GstTypeSeekFlags_gst_seek_flag_snap_after);

  return NEW_MARSHALLABLE (asn_DEF_GstTypeSeekFlags, tflags,
      GType_gst_type_seek_flags, buf, size, rules, err);
}

static GValue_t *
get_marshallable_seek_type_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeSeekType_t *tseek = NULL;
  GstSeekType seek;

  seek = (GstSeekType) g_value_get_enum (value);

  tseek = g_new0 (GstTypeSeekType_t, 1);

  switch (seek) {
    case GST_SEEK_TYPE_NONE:
      *tseek = GstTypeSeekType_gst_seek_type_none;
      break;
    case GST_SEEK_TYPE_SET:
      *tseek = GstTypeSeekType_gst_seek_type_set;
      break;
    case GST_SEEK_TYPE_END:
      *tseek = GstTypeSeekType_gst_seek_type_end;
      break;
  }

  return NEW_MARSHALLABLE (asn_DEF_GstTypeSeekType, tseek,
      GType_gst_type_seek_type, buf, size, rules, err);
}

static GValue_t *
get_marshallable_scheduling_flags_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstSchedulingFlags_t *tflags;
  GstSchedulingFlags flags;

  flags = (GstSchedulingFlags) g_value_get_flags (value);

  tflags = g_new0 (GstSchedulingFlags_t, 1);
  tflags->buf = calloc (1, 1);
  tflags->size = 1;
  /* Trim unused bits */
  tflags->bits_unused = 5;

  if (flags & GST_SCHEDULING_FLAG_SEEKABLE)
    BIT_STRING_SET_FLAG (tflags,
        GstSchedulingFlags_gst_scheduling_flag_seekable);

  if (flags & GST_SCHEDULING_FLAG_SEQUENTIAL)
    BIT_STRING_SET_FLAG (tflags,
        GstSchedulingFlags_gst_scheduling_flag_sequential);

  if (flags & GST_SCHEDULING_FLAG_BANDWIDTH_LIMITED)
    BIT_STRING_SET_FLAG (tflags,
        GstSchedulingFlags_gst_scheduling_flag_bandwidth_limited);

  return NEW_MARSHALLABLE (asn_DEF_GstSchedulingFlags, tflags,
      GType_gst_type_scheduling_flags, buf, size, rules, err);
}

static GValue_t *
get_marshallable_float_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GTypeFloat_t *f;

  f = g_new0 (GTypeFloat_t, 1);
  *f = g_value_get_float (value);

  return NEW_MARSHALLABLE (asn_DEF_GTypeFloat, f, GType_g_type_float, buf,
      size, rules, err);
}

static GValue_t *
get_marshallable_double_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GTypeDouble_t *d;

  d = g_new0 (GTypeDouble_t, 1);
  *d = g_value_get_double (value);

  return NEW_MARSHALLABLE (asn_DEF_GTypeDouble, d, GType_g_type_double,
      buf, size, rules, err);
}

static GValue_t *
get_marshallable_uint64_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  return INT64_AS_STRING_VALUE (value, GTypeUInt64_t, asn_DEF_GTypeUInt64,
      uint64, G_GUINT64_FORMAT, GType_g_type_uint64, buf, size, rules, err);
}

static GValue_t *
get_marshallable_int64_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  return INT64_AS_STRING_VALUE (value, GTypeInt64_t, asn_DEF_GTypeInt64,
      int64, G_GINT64_FORMAT, GType_g_type_int64, buf, size, rules, err);
}

static GValue_t *
get_marshallable_string_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GTypeString_t *str = NULL;
  const gchar *string;

  string = g_value_get_string (value);

  if (string != NULL && strlen (string) > size) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "Maximum size exceeded for GTypeString type");
    return NULL;
  }

  str = g_new0 (GTypeString_t, 1);
  if (string != NULL) {
    str->string = g_new0 (IA5String_t, 1);

    if (OCTET_STRING_fromString (str->string, string) < 0) {
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
          "Could not encode string value");
      asn_DEF_GTypeString.free_struct (&asn_DEF_GTypeString, str, 0);
      return NULL;
    }
  }

  return NEW_MARSHALLABLE (asn_DEF_GTypeString, str, GType_g_type_string,
      buf, size, rules, err);
}

static GstTypeTagList_t *
get_marshallable_tag_list (const GstTagList * list, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GstTypeTagList_t *tlist;
  gint i, n;

  tlist = g_new0 (GstTypeTagList_t, 1);

  n = gst_tag_list_n_tags (list);

  for (i = 0; i < n; i++) {
    GstTypeTag_t *tag;
    const gchar *tagname;
    gint ii, j;

    tag = g_new0 (GstTypeTag_t, 1);
    tagname = gst_tag_list_nth_tag_name (list, i);

    if (OCTET_STRING_fromString (&tag->tag, tagname) < 0) {
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
          "Could not encode tag list value");
      asn_DEF_GstTypeTag.free_struct (&asn_DEF_GstTypeTag, tag, 0);
      goto error;
    }

    j = gst_tag_list_get_tag_size (list, tagname);

    for (ii = 0; ii < j; ii++) {
      const GValue *gval;
      GValue_t *v;

      gval = gst_tag_list_get_value_index (list, tagname, ii);
      v = get_marshallable_GValue (gval, rules, buf, size, err);

      if (v == NULL)
        goto error;

      ASN_SEQUENCE_ADD (&tag->value, v);
    }

    ASN_SEQUENCE_ADD (tlist, tag);
  }

  return tlist;

error:
  if (tlist != NULL)
    asn_DEF_GstTypeTagList.free_struct (&asn_DEF_GstTypeTagList, tlist, 0);

  return NULL;
}

static GValue_t *
get_marshallable_tag_list_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeTagList_t *tlist = NULL;
  GstTagList *list;

  list = g_value_get_boxed (value);

  if (list != NULL) {
    tlist = get_marshallable_tag_list (list, rules, buf, size, err);

    if (*err != NULL)
      return NULL;
  }

  return NEW_MARSHALLABLE (asn_DEF_GstTypeTagList, tlist,
      GType_gst_type_tag_list, buf, size, rules, err);
}

static GValue_t *
get_marshallable_fraction_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeFraction_t *fraction = NULL;

  fraction = g_new0 (GstTypeFraction_t, 1);
  fraction->numerator = gst_value_get_fraction_numerator (value);
  fraction->denominator = gst_value_get_fraction_denominator (value);

  return NEW_MARSHALLABLE (asn_DEF_GstTypeFraction, fraction,
      GType_gst_type_fraction, buf, size, rules, err);
}

static GValue_t *
get_marshallable_int_range_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeIntegerRange_t *range = NULL;

  range = g_new0 (GstTypeIntegerRange_t, 1);
  range->max = gst_value_get_int_range_max (value);
  range->min = gst_value_get_int_range_min (value);

  return NEW_MARSHALLABLE (asn_DEF_GstTypeIntegerRange, range,
      GType_gst_type_int_range, buf, size, rules, err);
}

static GValue_t *
get_marshallable_segment_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeSegment_t *tsegment = NULL;
  GstSegment *segment = NULL;

  segment = g_value_get_boxed (value);

  if (segment == NULL)
    goto done;

  tsegment = g_new0 (GstTypeSegment_t, 1);
  tsegment->flags.buf = calloc (1, 1);
  tsegment->flags.size = 1;
  /* Trim unused bits */
  tsegment->flags.bits_unused = 4;

  if (segment->flags == GST_SEGMENT_FLAG_NONE)
    BIT_STRING_SET_FLAG (&tsegment->flags,
        GstTypeSegmentFlags_gst_segment_flag_none);

  if (segment->flags & GST_SEGMENT_FLAG_RESET)
    BIT_STRING_SET_FLAG (&tsegment->flags,
        GstTypeSegmentFlags_gst_segment_flag_reset);

  if (segment->flags & GST_SEGMENT_FLAG_SKIP)
    BIT_STRING_SET_FLAG (&tsegment->flags,
        GstTypeSegmentFlags_gst_segment_flag_skip);

  if (segment->flags & GST_SEGMENT_FLAG_SEGMENT)
    BIT_STRING_SET_FLAG (&tsegment->flags,
        GstTypeSegmentFlags_gst_segment_flag_segment);

  tsegment->rate = segment->rate;
  tsegment->applied_rate = segment->applied_rate;

  SET_MARSHAL_FORMAT (segment->format, &tsegment->format);

  INT64_TO_OCTET_STRING (&tsegment->base, G_GUINT64_FORMAT, segment->base, size,
      error, err);
  INT64_TO_OCTET_STRING (&tsegment->offset, G_GUINT64_FORMAT, segment->offset,
      size, error, err);
  INT64_TO_OCTET_STRING (&tsegment->start, G_GUINT64_FORMAT, segment->start,
      size, error, err);
  INT64_TO_OCTET_STRING (&tsegment->stop, G_GUINT64_FORMAT, segment->stop, size,
      error, err);
  INT64_TO_OCTET_STRING (&tsegment->time, G_GUINT64_FORMAT, segment->time, size,
      error, err);
  INT64_TO_OCTET_STRING (&tsegment->position, G_GUINT64_FORMAT,
      segment->position, size, error, err);
  INT64_TO_OCTET_STRING (&tsegment->duration, G_GUINT64_FORMAT,
      segment->duration, size, error, err);

done:
  return NEW_MARSHALLABLE (asn_DEF_GstTypeSegment, tsegment,
      GType_gst_type_segment, buf, size, rules, err);

error:
  if (tsegment != NULL)
    asn_DEF_GstTypeSegment.free_struct (&asn_DEF_GstTypeSegment, tsegment, 0);

  return NULL;
}

static GValue_t *
get_marshallable_double_range_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeDoubleRange_t *range = NULL;

  range = g_new0 (GstTypeDoubleRange_t, 1);
  range->max = gst_value_get_double_range_max (value);
  range->min = gst_value_get_double_range_min (value);

  return NEW_MARSHALLABLE (asn_DEF_GstTypeDoubleRange, range,
      GType_gst_type_double_range, buf, size, rules, err);
}

static GValue_t *
get_marshallable_fraction_range_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeFractionRange_t *range = NULL;
  const GValue *min, *max;

  range = g_new0 (GstTypeFractionRange_t, 1);
  min = gst_value_get_fraction_range_min (value);
  range->min.numerator = gst_value_get_fraction_numerator (min);
  range->min.denominator = gst_value_get_fraction_denominator (min);
  max = gst_value_get_fraction_range_max (value);
  range->max.numerator = gst_value_get_fraction_numerator (max);
  range->max.denominator = gst_value_get_fraction_denominator (max);

  return NEW_MARSHALLABLE (asn_DEF_GstTypeFractionRange, range,
      GType_gst_type_fraction_range, buf, size, rules, err);
}

static GValue_t *
get_marshallable_array_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeArray_t *l = NULL;
  guint len, i;

  l = g_new0 (GstTypeArray_t, 1);
  len = gst_value_array_get_size (value);

  for (i = 0; i < len; i++) {
    const GValue *gval;
    GValue_t *v;

    gval = gst_value_array_get_value (value, i);
    v = get_marshallable_GValue (gval, rules, buf, size, err);

    if (v == NULL)
      goto error;

    ASN_SEQUENCE_ADD (l, v);
  }

  return NEW_MARSHALLABLE (asn_DEF_GstTypeArray, l, GType_gst_type_array,
      buf, size, rules, err);

error:
  if (l != NULL)
    asn_DEF_GstTypeArray.free_struct (&asn_DEF_GstTypeArray, l, 0);

  return NULL;
}

static GValue_t *
get_marshallable_list_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeList_t *l = NULL;
  guint len, i;

  l = g_new0 (GstTypeList_t, 1);
  len = gst_value_list_get_size (value);

  for (i = 0; i < len; i++) {
    const GValue *gval;
    GValue_t *v;

    gval = gst_value_list_get_value (value, i);
    v = get_marshallable_GValue (gval, rules, buf, size, err);

    if (v == NULL)
      goto error;

    ASN_SEQUENCE_ADD (l, v);
  }

  return NEW_MARSHALLABLE (asn_DEF_GstTypeList, l, GType_gst_type_list,
      buf, size, rules, err);

error:
  if (l != NULL)
    asn_DEF_GstTypeList.free_struct (&asn_DEF_GstTypeList, l, 0);

  return NULL;
}

static void
get_marshallable_entry_list (GList * entries, void **ptr,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  struct GstTocEntry *list = NULL;
  GList *l;

  list = g_new0 (struct GstTocEntry, 1);

  for (l = entries; l != NULL; l = l->next) {
    GstTocEntry *entry;
    GstTocEntry_t *tentry;

    entry = l->data;
    tentry = get_marshallable_entry (entry, rules, buf, size, err);

    if (*err != NULL) {
      asn_sequence_empty (list);
      g_free (list);
    }

    ASN_SEQUENCE_ADD (list, tentry);
  }

  *ptr = list;
}

static GstTocEntry_t *
get_marshallable_entry (const GstTocEntry * entry, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GstTocEntry_t *tentry;
  GstTagList *taglist;
  gint64 start, stop;
  GList *list;

  tentry = g_new0 (GstTocEntry_t, 1);
  if (OCTET_STRING_fromString (&tentry->id, gst_toc_entry_get_uid (entry)) < 0) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "Could not encode entry uid");
    goto error;
  }

  switch (gst_toc_entry_get_entry_type (entry)) {
    case GST_TOC_ENTRY_TYPE_ANGLE:
      tentry->type = GstTocEntryType_gst_toc_entry_type_angle;
      break;
    case GST_TOC_ENTRY_TYPE_VERSION:
      tentry->type = GstTocEntryType_gst_toc_entry_type_version;
      break;
    case GST_TOC_ENTRY_TYPE_EDITION:
      tentry->type = GstTocEntryType_gst_toc_entry_type_edition;
      break;
    case GST_TOC_ENTRY_TYPE_INVALID:
      tentry->type = GstTocEntryType_gst_toc_entry_type_invalid;
      break;
    case GST_TOC_ENTRY_TYPE_TITLE:
      tentry->type = GstTocEntryType_gst_toc_entry_type_title;
      break;
    case GST_TOC_ENTRY_TYPE_TRACK:
      tentry->type = GstTocEntryType_gst_toc_entry_type_track;
      break;
    case GST_TOC_ENTRY_TYPE_CHAPTER:
      tentry->type = GstTocEntryType_gst_toc_entry_type_chapter;
      break;
  }

  gst_toc_entry_get_start_stop_times (entry, &start, &stop);

  INT64_TO_OCTET_STRING (&tentry->start, G_GINT64_FORMAT, start, size, error,
      err);
  INT64_TO_OCTET_STRING (&tentry->stop, G_GINT64_FORMAT, stop, size, error,
      err);

  taglist = gst_toc_entry_get_tags (entry);
  if (taglist != NULL) {
    tentry->tags = get_marshallable_tag_list (taglist, rules, buf, size, err);

    if (*err != NULL)
      goto error;
  }

  list = gst_toc_entry_get_sub_entries (entry);
  if (list != NULL) {
    get_marshallable_entry_list (list, (void **) &tentry->sub_entries, rules,
        buf, size, err);

    if (*err != NULL)
      goto error;
  }

  return tentry;

error:
  if (tentry != NULL)
    asn_DEF_GstTocEntry.free_struct (&asn_DEF_GstTocEntry, tentry, 0);

  return NULL;
}

static GValue_t *
get_marshallable_toc_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstToc_t *ttoc = NULL;
  GstTagList *taglist;
  GstToc *toc;
  GList *list;

  toc = g_value_get_boxed (value);
  if (toc == NULL)
    goto done;

  ttoc = g_new0 (GstToc_t, 1);
  switch (gst_toc_get_scope (toc)) {
    case GST_TOC_SCOPE_GLOBAL:
      ttoc->scope = GstTocScope_gst_toc_scope_global;
      break;
    case GST_TOC_SCOPE_CURRENT:
      ttoc->scope = GstTocScope_gst_toc_scope_current;
      break;
  }

  taglist = gst_toc_get_tags (toc);
  if (taglist != NULL) {
    ttoc->tags = get_marshallable_tag_list (taglist, rules, buf, size, err);
    if (*err != NULL)
      goto error;
  }

  list = gst_toc_get_entries (toc);
  if (list != NULL) {
    get_marshallable_entry_list (list, (void **) &ttoc->entries, rules, buf,
        size, err);

    if (*err != NULL)
      goto error;
  }

done:
  return NEW_MARSHALLABLE (asn_DEF_GstToc, ttoc, GType_gst_type_toc, buf, size,
      rules, err);

error:
  if (ttoc != NULL)
    asn_DEF_GstToc.free_struct (&asn_DEF_GstToc, ttoc, 0);

  return NULL;
}

static GValue_t *
get_marshallable_caps_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstCaps_t *tcaps = NULL;
  GstCaps *caps;

  caps = g_value_get_boxed (value);

  if (caps != NULL)
    tcaps = get_marshallable_GstCaps (caps, rules, buf, size, err);

  if (*err != NULL)
    return NULL;

  return NEW_MARSHALLABLE (asn_DEF_GstCaps, tcaps, GType_gst_type_caps, buf,
      size, rules, err);
}

static GValue_t *
get_marshallable_context_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  const GstStructure *structure;
  GstContext_t *tctx = NULL;
  GstStructure_t *tstruct;
  const gchar *string;
  GstContext *ctx;
  gsize s = 0;

  ctx = g_value_get_boxed (value);
  if (ctx == NULL)
    goto done;

  string = gst_context_get_context_type (ctx);
  if (string == NULL) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "GstContext does not have a valid type");
    return NULL;
  }

  if ((s = strlen (string)) > size) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "GstContext type exceeds buffer size");
    return NULL;
  }

  tctx = g_new0 (GstContext_t, 1);

  if (OCTET_STRING_fromString (&tctx->context_type, string) < 0) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "Could not encode type value");
    goto error;
  }

  tctx->persistent = gst_context_is_persistent (ctx);

  structure = gst_context_get_structure (ctx);
  tstruct = get_marshallable_gststructure (structure, rules, buf, size, err);
  own_allocated_memory (tstruct, &tctx->structure);
  asn_DEF_GstStructure.free_struct (&asn_DEF_GstStructure, tstruct, 0);

done:
  return NEW_MARSHALLABLE (asn_DEF_GstContext, tctx, GType_gst_type_context,
      buf, size, rules, err);

error:
  if (tctx != NULL)
    asn_DEF_GstContext.free_struct (&asn_DEF_GstContext, tctx, 0);

  return NULL;
}

static GValue_t *
get_marshallable_buffer_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstBuffer_t *tbuffer = NULL;
  GstBuffer *buffer;

  buffer = g_value_get_boxed (value);
  if (buffer == NULL)
    goto done;

  tbuffer = get_marshallable_GstBuffer (buffer, rules, buf, size, err);

  if (tbuffer == NULL)
    return NULL;

done:
  return NEW_MARSHALLABLE (asn_DEF_OptimizedGstBuffer, tbuffer,
      GType_gst_type_buffer, buf, size, rules, err);

  if (tbuffer != NULL)
    asn_DEF_OptimizedGstBuffer.free_struct (&asn_DEF_OptimizedGstBuffer,
        tbuffer, 0);

  return NULL;
}

static GValue_t *
get_marshallable_date_time_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstTypeDateTime_t *tdate = NULL;
  GstDateTime *date;

  date = g_value_get_boxed (value);

  if (date == NULL)
    goto done;

  tdate = g_new0 (GstTypeDateTime_t, 1);

  tdate->tzoffset = gst_date_time_get_time_zone_offset (date);
  tdate->year = gst_date_time_get_year (date);
  tdate->month = gst_date_time_get_month (date);
  tdate->day = gst_date_time_get_day (date);
  tdate->hour = gst_date_time_get_hour (date);
  tdate->minute = gst_date_time_get_minute (date);
  tdate->seconds = gst_date_time_get_second (date);

done:
  return NEW_MARSHALLABLE (asn_DEF_GstTypeDateTime, tdate,
      GType_gst_type_date_time, buf, size, rules, err);
}

static GValue_t *
get_marshallable_buffering_mode_gvalue (const GValue * value,
    KurentoMarshalRules rules, gchar * buf, gsize size, GError ** err)
{
  GstBufferingMode_t *tbuffmode = NULL;
  GstBufferingMode buffmode;

  buffmode = (GstStreamFlags) g_value_get_enum (value);

  tbuffmode = g_new0 (GstBufferingMode_t, 1);
  switch (buffmode) {
    case GST_BUFFERING_STREAM:
      *tbuffmode = GstBufferingMode_gst_buffering_stream;
      break;
    case GST_BUFFERING_DOWNLOAD:
      *tbuffmode = GstBufferingMode_gst_buffering_download;
      break;
    case GST_BUFFERING_TIMESHIFT:
      *tbuffmode = GstBufferingMode_gst_buffering_timeshift;
      break;
    case GST_BUFFERING_LIVE:
      *tbuffmode = GstBufferingMode_gst_buffering_live;
      break;
  }

  return NEW_MARSHALLABLE (asn_DEF_GstBufferingMode, tbuffmode,
      GType_gst_type_buffering_mode, buf, size, rules, err);
}

GValue_t *
get_marshallable_GValue (const GValue * value, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  GValue_t *ret;

  if (value == NULL) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "NULL GValue provided");
    return NULL;
  }

  if (G_VALUE_HOLDS_INT (value))
    ret = get_marshallable_int_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_UINT (value))
    ret = get_marshallable_uint_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_INT64 (value))
    ret = get_marshallable_int64_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_UINT64 (value))
    ret = get_marshallable_uint64_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_BOOLEAN (value))
    ret = get_marshallable_boolean_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_DOUBLE (value))
    ret = get_marshallable_double_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_DATE (value))
    ret = get_marshallable_date_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_FLOAT (value))
    ret = get_marshallable_float_gvalue (value, rules, buf, size, err);
  else if (G_VALUE_HOLDS_STRING (value))
    ret = get_marshallable_string_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_FORMAT (value))
    ret = get_marshallable_format_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_FRACTION (value))
    ret = get_marshallable_fraction_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_INT_RANGE (value))
    ret = get_marshallable_int_range_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_CAPS (value))
    ret = get_marshallable_caps_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_DATE_TIME (value))
    ret = get_marshallable_date_time_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_DOUBLE_RANGE (value))
    ret = get_marshallable_double_range_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_QOS_TYPE (value))
    ret = get_marshallable_qos_type_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_FRACTION_RANGE (value))
    ret = get_marshallable_fraction_range_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_LIST (value))
    ret = get_marshallable_list_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_SEEK_FLAGS (value))
    ret = get_marshallable_seek_flags_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_SEEK_TYPE (value))
    ret = get_marshallable_seek_type_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_SEGMENT (value))
    ret = get_marshallable_segment_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_STREAM_FLAGS (value))
    ret = get_marshallable_stream_flags_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_TAG_LIST (value))
    ret = get_marshallable_tag_list_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_TOC (value))
    ret = get_marshallable_toc_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_ARRAY (value))
    ret = get_marshallable_array_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_BUFFERING_MODE (value))
    ret = get_marshallable_buffering_mode_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_SCHEDULING_FLAGS (value))
    ret =
        get_marshallable_scheduling_flags_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_CONTEXT (value))
    ret = get_marshallable_context_gvalue (value, rules, buf, size, err);
  else if (GST_VALUE_HOLDS_BUFFER (value))
    ret = get_marshallable_buffer_gvalue (value, rules, buf, size, err);
  else {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,
        "Not marshallable GValue: %s", g_type_name (G_VALUE_TYPE (value)));
    ret = NULL;
  }

  return ret;
}

static void
dec_int_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GTypeInt_t *n = NULL;

  decode_marshallable_gvalue (&asn_DEF_GTypeInt, val, rules, (void **) &n, err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, G_TYPE_INT);
  g_value_set_int (*value, *n);

end:
  if (n != NULL)
    asn_DEF_GTypeInt.free_struct (&asn_DEF_GTypeInt, n, 0);
}

static void
dec_uint_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GTypeInt_t *n = NULL;

  decode_marshallable_gvalue (&asn_DEF_GTypeInt, val, rules, (void **) &n, err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, G_TYPE_UINT);
  g_value_set_uint (*value, *n);

end:
  if (n != NULL)
    asn_DEF_GTypeInt.free_struct (&asn_DEF_GTypeInt, n, 0);
}

static void
dec_date_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GTypeDate_t *tdate = NULL;
  GDate *date = NULL;

  if (val->value != NULL) {
    decode_marshallable_gvalue (&asn_DEF_GTypeDate, val, rules,
        (void **) &tdate, err);

    if (*err != NULL)
      goto end;

    date = g_date_new_julian (*tdate);
  }

  *value = g_slice_new0 (GValue);
  g_value_init (*value, G_TYPE_DATE);
  g_value_set_boxed (*value, date);

end:
  if (date != NULL)
    g_date_free (date);

  if (tdate != NULL)
    asn_DEF_GTypeDate.free_struct (&asn_DEF_GTypeDate, tdate, 0);
}

static void
dec_boolean_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GTypeBoolean_t *b = NULL;

  decode_marshallable_gvalue (&asn_DEF_GTypeBoolean, val, rules, (void **) &b,
      err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, G_TYPE_BOOLEAN);
  g_value_set_boolean (*value, *b);

end:
  if (b != NULL)
    asn_DEF_GTypeBoolean.free_struct (&asn_DEF_GTypeBoolean, b, 0);
}

static void
dec_format_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeFormat_t *tformat = NULL;
  GstFormat format;

  decode_marshallable_gvalue (&asn_DEF_GstTypeFormat, val, rules,
      (void **) &tformat, err);

  if (*err != NULL)
    goto end;

  SET_FORMAT (*tformat, &format, err, end);

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_FORMAT);
  g_value_set_enum (*value, format);

end:
  if (tformat != NULL) {
    asn_DEF_GstTypeFormat.free_struct (&asn_DEF_GstTypeFormat, tformat, 0);
  }
}

static void
dec_qos_type_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeQOSType_t *tqos = NULL;
  GstQOSType qos;

  decode_marshallable_gvalue (&asn_DEF_GstTypeQOSType, val, rules,
      (void **) &tqos, err);

  if (*err != NULL)
    goto end;

  switch (*tqos) {
    case GstTypeQOSType_gst_qos_type_overflow:
      qos = GST_QOS_TYPE_OVERFLOW;
      break;
    case GstTypeQOSType_gst_qos_type_underflow:
      qos = GST_QOS_TYPE_UNDERFLOW;
      break;
    case GstTypeQOSType_gst_qos_type_throttle:
      qos = GST_QOS_TYPE_THROTTLE;
      break;
    default:
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
          "Decoded invalid qos type (%ld)", *tqos);
      goto end;
  }

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_QOS_TYPE);
  g_value_set_enum (*value, qos);

end:
  if (tqos != NULL) {
    asn_DEF_GstTypeQOSType.free_struct (&asn_DEF_GstTypeQOSType, tqos, 0);
  }
}

static void
dec_stream_flags_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeStreamFlags_t *tflags = NULL;
  GstStreamFlags flags = GST_STREAM_FLAG_NONE;

  decode_marshallable_gvalue (&asn_DEF_GstTypeStreamFlags, val, rules,
      (void **) &tflags, err);

  if (*err != NULL)
    goto end;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstTypeStreamFlags_gst_stream_flag_sparse))
    flags |= GST_STREAM_FLAG_SPARSE;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstTypeStreamFlags_gst_stream_flag_select))
    flags |= GST_STREAM_FLAG_SELECT;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstTypeStreamFlags_gst_stream_flagunselect))
    flags |= GST_STREAM_FLAG_UNSELECT;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_STREAM_FLAGS);
  g_value_set_flags (*value, flags);

end:
  if (tflags != NULL) {
    asn_DEF_GstTypeStreamFlags.free_struct (&asn_DEF_GstTypeStreamFlags, tflags,
        0);
  }
}

static void
dec_seek_flags_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeSeekFlags_t *tflags = NULL;
  GstSeekFlags flags = GST_STREAM_FLAG_NONE;

  decode_marshallable_gvalue (&asn_DEF_GstTypeSeekFlags, val, rules,
      (void **) &tflags, err);

  if (*err != NULL)
    goto end;

  if (BIT_STRING_FLAG_IS_SET (tflags, GstTypeSeekFlags_gst_seek_flag_none))
    flags |= GST_SEEK_FLAG_NONE;

  if (BIT_STRING_FLAG_IS_SET (tflags, GstTypeSeekFlags_gst_seek_flag_flush))
    flags |= GST_SEEK_FLAG_FLUSH;

  if (BIT_STRING_FLAG_IS_SET (tflags, GstTypeSeekFlags_gst_seek_flag_accurate))
    flags |= GST_SEEK_FLAG_ACCURATE;

  if (BIT_STRING_FLAG_IS_SET (tflags, GstTypeSeekFlags_gst_seek_flag_key_unit))
    flags |= GST_SEEK_FLAG_KEY_UNIT;

  if (BIT_STRING_FLAG_IS_SET (tflags, GstTypeSeekFlags_gst_seek_flag_segment))
    flags |= GST_SEEK_FLAG_SEGMENT;

  if (BIT_STRING_FLAG_IS_SET (tflags, GstTypeSeekFlags_gst_seek_flag_skip))
    flags |= GST_SEEK_FLAG_SKIP;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstTypeSeekFlags_gst_seek_flag_snap_before))
    flags |= GST_SEEK_FLAG_SNAP_BEFORE;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstTypeSeekFlags_gst_seek_flag_snap_after))
    flags |= GST_SEEK_FLAG_SNAP_AFTER;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_SEEK_FLAGS);
  g_value_set_flags (*value, flags);

end:
  if (tflags != NULL)
    asn_DEF_GstTypeSeekFlags.free_struct (&asn_DEF_GstTypeSeekFlags, tflags, 0);
}

static void
dec_seek_type_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeSeekType_t *tseek = NULL;
  GstSeekType seek;

  decode_marshallable_gvalue (&asn_DEF_GstTypeSeekType, val, rules,
      (void **) &tseek, err);

  if (*err != NULL)
    goto end;

  switch (*tseek) {
    case GstTypeSeekType_gst_seek_type_none:
      seek = GST_SEEK_TYPE_NONE;
      break;
    case GstTypeSeekType_gst_seek_type_set:
      seek = GST_SEEK_TYPE_SET;
      break;
    case GstTypeSeekType_gst_seek_type_end:
      seek = GST_SEEK_TYPE_END;
      break;
    default:
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
          "Decoded invalid seek type (%ld)", *tseek);
      goto end;
  }

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_SEEK_TYPE);
  g_value_set_enum (*value, seek);

end:
  if (tseek != NULL) {
    asn_DEF_GstTypeSeekType.free_struct (&asn_DEF_GstTypeSeekType, tseek, 0);
  }
}

static void
dec_float_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GTypeFloat_t *f = NULL;

  decode_marshallable_gvalue (&asn_DEF_GTypeFloat, val, rules, (void **) &f,
      err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, G_TYPE_FLOAT);
  g_value_set_float (*value, *f);

end:
  if (f != NULL)
    asn_DEF_GTypeFloat.free_struct (&asn_DEF_GTypeFloat, f, 0);
}

static void
dec_double_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GTypeDouble_t *d = NULL;

  decode_marshallable_gvalue (&asn_DEF_GTypeDouble, val, rules, (void **) &d,
      err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, G_TYPE_DOUBLE);
  g_value_set_double (*value, *d);

end:
  if (d != NULL)
    asn_DEF_GTypeDouble.free_struct (&asn_DEF_GTypeDouble, d, 0);
}

static void
dec_uint64_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  *value = STRING_AS_INT64_VALUE (GTypeUInt64_t, asn_DEF_GTypeUInt64, uint64,
      G_TYPE_UINT64, val, strtoull, rules, err);
}

static void
dec_int64_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  *value = STRING_AS_INT64_VALUE (GTypeInt64_t, asn_DEF_GTypeInt64, int64,
      G_TYPE_INT64, val, strtoll, rules, err);
}

static void
dec_string_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GTypeString_t *str = NULL;
  gchar *string = NULL;

  decode_marshallable_gvalue (&asn_DEF_GTypeString, val, rules, (void **) &str,
      err);

  if (*err != NULL)
    goto end;

  if (str->string != NULL)
    string = (gchar *) str->string->buf;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, G_TYPE_STRING);
  g_value_set_string (*value, string);

end:
  if (str != NULL)
    asn_DEF_GTypeString.free_struct (&asn_DEF_GTypeString, str, 0);
}

static void
dec_segment_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeSegment_t *tsegment = NULL;
  GstSegment *segment = NULL;

  if (val->value == NULL)
    goto done;

  decode_marshallable_gvalue (&asn_DEF_GstTypeSegment, val, rules,
      (void **) &tsegment, err);

  if (*err != NULL)
    goto end;

  segment = gst_segment_new ();

  if (BIT_STRING_FLAG_IS_SET (&tsegment->flags,
          GstTypeSegmentFlags_gst_segment_flag_none))
    segment->flags = GST_SEGMENT_FLAG_NONE;

  if (BIT_STRING_FLAG_IS_SET (&tsegment->flags,
          GstTypeSegmentFlags_gst_segment_flag_reset))
    segment->flags |= GST_SEGMENT_FLAG_RESET;

  if (BIT_STRING_FLAG_IS_SET (&tsegment->flags,
          GstTypeSegmentFlags_gst_segment_flag_skip))
    segment->flags |= GST_SEGMENT_FLAG_SKIP;

  if (BIT_STRING_FLAG_IS_SET (&tsegment->flags,
          GstTypeSegmentFlags_gst_segment_flag_segment))
    segment->flags |= GST_SEGMENT_FLAG_SEGMENT;

  segment->rate = tsegment->rate;
  segment->applied_rate = tsegment->applied_rate;

  SET_FORMAT (tsegment->format, &segment->format, err, end);

  segment->base = OCTET_STRING_TO_INT64 (tsegment->base.buf, uint64, strtoull);
  segment->offset =
      OCTET_STRING_TO_INT64 (tsegment->offset.buf, uint64, strtoull);
  segment->start =
      OCTET_STRING_TO_INT64 (tsegment->start.buf, uint64, strtoull);
  segment->stop = OCTET_STRING_TO_INT64 (tsegment->stop.buf, uint64, strtoull);
  segment->time = OCTET_STRING_TO_INT64 (tsegment->time.buf, uint64, strtoull);
  segment->position =
      OCTET_STRING_TO_INT64 (tsegment->position.buf, uint64, strtoull);
  segment->duration =
      OCTET_STRING_TO_INT64 (tsegment->duration.buf, uint64, strtoull);

done:
  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_SEGMENT);
  g_value_set_boxed (*value, segment);

end:
  if (segment != NULL)
    gst_segment_free (segment);

  if (tsegment != NULL)
    asn_DEF_GstTypeSegment.free_struct (&asn_DEF_GstTypeSegment, tsegment, 0);
}

static void
dec_fraction_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeFraction_t *fraction = NULL;

  decode_marshallable_gvalue (&asn_DEF_GstTypeFraction, val, rules,
      (void **) &fraction, err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_FRACTION);
  gst_value_set_fraction (*value, fraction->numerator, fraction->denominator);

end:
  if (fraction != NULL)
    asn_DEF_GstTypeFraction.free_struct (&asn_DEF_GstTypeFraction, fraction, 0);
}

static void
dec_int_range_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeIntegerRange_t *range = NULL;

  decode_marshallable_gvalue (&asn_DEF_GstTypeIntegerRange, val, rules,
      (void **) &range, err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_INT_RANGE);
  gst_value_set_int_range (*value, range->min, range->max);

end:
  if (range != NULL)
    asn_DEF_GstTypeIntegerRange.free_struct (&asn_DEF_GstTypeIntegerRange,
        range, 0);
}

static void
dec_date_time_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeDateTime_t *tdate = NULL;
  GstDateTime *date = NULL;

  if (val->value == NULL)
    goto done;

  decode_marshallable_gvalue (&asn_DEF_GstTypeDateTime, val, rules,
      (void **) &tdate, err);

  if (*err != NULL)
    goto end;

  date = gst_date_time_new (tdate->tzoffset, tdate->year, tdate->month,
      tdate->day, tdate->hour, tdate->minute, tdate->seconds);

done:
  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_DATE_TIME);
  g_value_set_boxed (*value, date);

end:
  if (date != NULL)
    gst_date_time_unref (date);

  if (tdate != NULL)
    asn_DEF_GstTypeDateTime.free_struct (&asn_DEF_GstTypeDateTime, tdate, 0);
}

static void
dec_double_range_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeDoubleRange_t *range = NULL;

  decode_marshallable_gvalue (&asn_DEF_GstTypeDoubleRange, val, rules,
      (void **) &range, err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_DOUBLE_RANGE);
  gst_value_set_double_range (*value, range->min, range->max);

end:
  if (range != NULL)
    asn_DEF_GstTypeDoubleRange.free_struct (&asn_DEF_GstTypeDoubleRange,
        range, 0);
}

static void
dec_caps_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstCaps *caps = NULL;

  if (val->value != NULL && !dec_GstCaps (rules,
          (const gchar *) val->value->buf, val->value->size, &caps, err)) {
    goto end;
  }

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_CAPS);
  g_value_set_boxed (*value, caps);

end:
  if (caps != NULL)
    gst_caps_unref (caps);
}

static void
dec_fraction_range_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeFractionRange_t *range = NULL;
  GValue *min = NULL, *max = NULL;

  decode_marshallable_gvalue (&asn_DEF_GstTypeFractionRange, val, rules,
      (void **) &range, err);

  if (*err != NULL)
    goto end;

  min = g_slice_new0 (GValue);
  max = g_slice_new0 (GValue);

  g_value_init (min, GST_TYPE_FRACTION);
  g_value_init (max, GST_TYPE_FRACTION);

  gst_value_set_fraction (min, range->min.numerator, range->min.denominator);
  gst_value_set_fraction (max, range->max.numerator, range->max.denominator);

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_FRACTION_RANGE);
  gst_value_set_fraction_range (*value, min, max);

end:
  if (range != NULL)
    asn_DEF_GstTypeFractionRange.free_struct (&asn_DEF_GstTypeFractionRange,
        range, 0);

  if (min != NULL)
    g_slice_free (GValue, min);

  if (max != NULL)
    g_slice_free (GValue, max);
}

static void
dec_list_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeList_t *l = NULL;
  guint i;

  decode_marshallable_gvalue (&asn_DEF_GstTypeList, val, rules, (void **) &l,
      err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_LIST);

  for (i = 0; i < l->list.count; i++) {
    GValue *gval;
    GValue_t *v;

    v = l->list.array[i];
    gval = get_GValue (v, rules, err);

    if (*err != NULL) {
      g_value_unset (*value);
      g_slice_free (GValue, *value);
      goto end;
    }

    gst_value_list_append_value (*value, gval);

    g_value_unset (gval);
    g_slice_free (GValue, gval);
  }

end:
  if (l != NULL)
    asn_DEF_GstTypeList.free_struct (&asn_DEF_GstTypeList, l, 0);
}

static void
dec_array_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeArray_t *l = NULL;
  guint i;

  decode_marshallable_gvalue (&asn_DEF_GstTypeArray, val, rules, (void **) &l,
      err);

  if (*err != NULL)
    goto end;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_ARRAY);

  for (i = 0; i < l->list.count; i++) {
    GValue *gval;
    GValue_t *v;

    v = l->list.array[i];
    gval = get_GValue (v, rules, err);

    if (*err != NULL) {
      g_value_unset (*value);
      g_slice_free (GValue, *value);
      goto end;
    }

    gst_value_array_append_value (*value, gval);

    g_value_unset (gval);
    g_slice_free (GValue, gval);
  }

end:
  if (l != NULL)
    asn_DEF_GstTypeArray.free_struct (&asn_DEF_GstTypeArray, l, 0);
}

static GstTagList *
get_tag_list (GstTypeTagList_t * tlist, KurentoMarshalRules rules,
    GError ** err)
{
  GstTagList *list = NULL;
  guint i;

  list = gst_tag_list_new_empty ();

  for (i = 0; i < tlist->list.count; i++) {
    GstTypeTag_t *tag;
    guint ii;

    tag = tlist->list.array[i];
    for (ii = 0; ii < tag->value.list.count; ii++) {
      GValue *gval;
      GValue_t *v;

      v = tag->value.list.array[ii];
      gval = get_GValue (v, rules, err);

      if (*err != NULL)
        goto end;

      gst_tag_list_add_value (list, GST_TAG_MERGE_APPEND,
          (const gchar *) tag->tag.buf, gval);

      g_value_unset (gval);
      g_slice_free (GValue, gval);
    }
  }

  return list;

end:
  if (list != NULL)
    gst_tag_list_unref (list);

  if (tlist != NULL)
    asn_DEF_GstTypeTagList.free_struct (&asn_DEF_GstTypeTagList, tlist, 0);

  return NULL;
}

static void
dec_tag_list_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstTypeTagList_t *tlist = NULL;
  GstTagList *list = NULL;

  if (val->value != NULL) {
    decode_marshallable_gvalue (&asn_DEF_GstTypeTagList, val, rules,
        (void **) &tlist, err);

    if (*err != NULL)
      goto end;

    list = get_tag_list (tlist, rules, err);

    if (*err != NULL)
      goto end;
  }

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_TAG_LIST);
  g_value_set_boxed (*value, list);

end:
  if (list != NULL)
    gst_tag_list_unref (list);

  if (tlist != NULL)
    asn_DEF_GstTypeTagList.free_struct (&asn_DEF_GstTypeTagList, tlist, 0);
}

static void
append_sub_entries (const GstTocEntry_t * tentry, GstTocEntry * entry,
    KurentoMarshalRules rules, GError ** err)
{
  guint i;

  if (tentry->sub_entries == NULL)
    return;

  for (i = 0; i < tentry->sub_entries->list.count; i++) {
    GstTocEntry_t *te;
    GstTocEntry *e;

    te = tentry->sub_entries->list.array[i];
    e = get_entry (te, rules, err);

    if (*err != NULL)
      return;

    gst_toc_entry_append_sub_entry (entry, e);
  }
}

static GstTocEntry *
get_entry (const GstTocEntry_t * tentry, KurentoMarshalRules rules,
    GError ** err)
{
  GstTocEntry *entry = NULL;
  GstTocEntryType type;
  gint64 start, stop;

  switch (tentry->type) {
    case GstTocEntryType_gst_toc_entry_type_angle:
      type = GST_TOC_ENTRY_TYPE_ANGLE;
      break;
    case GstTocEntryType_gst_toc_entry_type_version:
      type = GST_TOC_ENTRY_TYPE_VERSION;
      break;
    case GstTocEntryType_gst_toc_entry_type_edition:
      type = GST_TOC_ENTRY_TYPE_EDITION;
      break;
    case GstTocEntryType_gst_toc_entry_type_invalid:
      type = GST_TOC_ENTRY_TYPE_INVALID;
      break;
    case GstTocEntryType_gst_toc_entry_type_title:
      type = GST_TOC_ENTRY_TYPE_TITLE;
      break;
    case GstTocEntryType_gst_toc_entry_type_track:
      type = GST_TOC_ENTRY_TYPE_TRACK;
      break;
    case GstTocEntryType_gst_toc_entry_type_chapter:
      type = GST_TOC_ENTRY_TYPE_CHAPTER;
      break;
    default:
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
          "Decoded invalid toc entry type (%ld)", tentry->type);
      goto end;
  }

  start = OCTET_STRING_TO_INT64 (tentry->start.buf, int64, strtoll);
  stop = OCTET_STRING_TO_INT64 (tentry->stop.buf, int64, strtoll);

  entry = gst_toc_entry_new (type, (gchar *) tentry->id.buf);
  gst_toc_entry_set_start_stop_times (entry, start, stop);

  if (tentry->tags != NULL) {
    GstTagList *taglist;

    taglist = get_tag_list (tentry->tags, rules, err);

    if (*err != NULL)
      goto end;

    gst_toc_entry_set_tags (entry, taglist);
  }

  append_sub_entries (tentry, entry, rules, err);

  if (*err == NULL)
    return entry;

end:
  if (entry != NULL)
    gst_toc_entry_unref (entry);

  return NULL;
}

static void
append_entries (const GstToc_t * ttoc, GstToc * toc, KurentoMarshalRules rules,
    GError ** err)
{
  guint i;

  if (ttoc->entries == NULL)
    return;

  for (i = 0; i < ttoc->entries->list.count; i++) {
    GstTocEntry_t *tentry;
    GstTocEntry *entry;

    tentry = ttoc->entries->list.array[i];
    entry = get_entry (tentry, rules, err);

    if (*err != NULL)
      return;

    gst_toc_append_entry (toc, entry);
  }
}

static void
dec_toc_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstToc_t *ttoc = NULL;
  GstToc *toc = NULL;
  GstTocScope scope;

  if (val->value == NULL)
    goto done;

  decode_marshallable_gvalue (&asn_DEF_GstToc, val, rules,
      (void **) &ttoc, err);

  if (*err != NULL)
    goto end;

  switch (ttoc->scope) {
    case GstTocScope_gst_toc_scope_global:
      scope = GST_TOC_SCOPE_GLOBAL;
      break;
    case GstTocScope_gst_toc_scope_current:
      scope = GST_TOC_SCOPE_CURRENT;
      break;
    default:
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
          "Decoded invalid toc scope (%ld)", ttoc->scope);
      goto end;
  }

  toc = gst_toc_new (scope);

  if (ttoc->tags != NULL) {
    GstTagList *list;

    list = get_tag_list (ttoc->tags, rules, err);
    if (*err != NULL)
      goto end;

    gst_toc_set_tags (toc, list);
  }

  append_entries (ttoc, toc, rules, err);

  if (*err != NULL)
    goto end;

done:
  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_TOC);
  g_value_set_boxed (*value, toc);

end:
  if (toc != NULL)
    gst_toc_unref (toc);

  if (ttoc != NULL)
    asn_DEF_GstToc.free_struct (&asn_DEF_GstToc, ttoc, 0);
}

static void
dec_buffering_mode_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstBufferingMode_t *tbuffmode = NULL;
  GstBufferingMode buffmode;

  decode_marshallable_gvalue (&asn_DEF_GstBufferingMode, val, rules,
      (void **) &tbuffmode, err);

  if (*err != NULL)
    goto end;

  switch (*tbuffmode) {
    case GstBufferingMode_gst_buffering_stream:
      buffmode = GST_BUFFERING_STREAM;
      break;
    case GstBufferingMode_gst_buffering_download:
      buffmode = GST_BUFFERING_DOWNLOAD;
      break;
    case GstBufferingMode_gst_buffering_timeshift:
      buffmode = GST_BUFFERING_TIMESHIFT;
      break;
    case GstBufferingMode_gst_buffering_live:
      buffmode = GST_BUFFERING_LIVE;
      break;
    default:
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
          "Decoded invalid buffer mode (%ld)", *tbuffmode);
      goto end;
  }

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_BUFFERING_MODE);
  g_value_set_enum (*value, buffmode);

end:
  if (tbuffmode != NULL) {
    asn_DEF_GstBufferingMode.free_struct (&asn_DEF_GstBufferingMode, tbuffmode,
        0);
  }
}

static void
dec_scheduling_flags_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstSchedulingFlags_t *tflags = NULL;
  GstSchedulingFlags flags = 0;

  decode_marshallable_gvalue (&asn_DEF_GstSchedulingFlags, val, rules,
      (void **) &tflags, err);

  if (*err != NULL)
    goto end;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstSchedulingFlags_gst_scheduling_flag_seekable))
    flags |= GST_SCHEDULING_FLAG_SEEKABLE;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstSchedulingFlags_gst_scheduling_flag_sequential))
    flags |= GST_SCHEDULING_FLAG_SEQUENTIAL;

  if (BIT_STRING_FLAG_IS_SET (tflags,
          GstSchedulingFlags_gst_scheduling_flag_bandwidth_limited))
    flags |= GST_SCHEDULING_FLAG_BANDWIDTH_LIMITED;

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_SCHEDULING_FLAGS);
  g_value_set_flags (*value, flags);

end:
  if (tflags != NULL) {
    asn_DEF_GstSchedulingFlags.free_struct (&asn_DEF_GstSchedulingFlags, tflags,
        0);
  }
}

static gboolean
fill_gststructure (GQuark field_id, const GValue * value, gpointer user_data)
{
  GstStructure *structure = user_data;

  gst_structure_id_set_value (structure, field_id, value);
  return TRUE;
}

static void
dec_context_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstContext_t *tctx = NULL;
  GstContext *ctx = NULL;
  GstStructure *structure1 = NULL, *structure2;
  gchar *context_type;

  decode_marshallable_gvalue (&asn_DEF_GstContext, val, rules,
      (void **) &tctx, err);

  if (*err != NULL)
    goto end;

  structure1 = get_GstStructure (&tctx->structure, rules, err);
  if (*err != NULL)
    goto end;

  context_type = (gchar *) tctx->context_type.buf;
  ctx = gst_context_new (context_type, (tctx->persistent != 0));

  structure2 = gst_context_writable_structure (ctx);
  gst_structure_remove_all_fields (structure2);
  gst_structure_foreach (structure1, fill_gststructure, structure2);

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_CONTEXT);
  g_value_set_boxed (*value, ctx);

end:
  if (ctx != NULL)
    gst_context_unref (ctx);

  if (structure1 != NULL)
    gst_structure_free (structure1);

  if (tctx != NULL)
    asn_DEF_GstContext.free_struct (&asn_DEF_GstContext, tctx, 0);
}

static void
dec_buffer_gvalue (KurentoMarshalRules rules, const GValue_t * val,
    GValue ** value, GError ** err)
{
  GstBuffer_t *tbuffer = NULL;
  GstBuffer *buffer = NULL;

  if (val->value != NULL && !dec_GstBuffer (rules,
          (const gchar *) val->value->buf, val->value->size, &buffer, err)) {
    goto end;
  }

  *value = g_slice_new0 (GValue);
  g_value_init (*value, GST_TYPE_BUFFER);
  g_value_set_boxed (*value, buffer);

end:
  if (buffer != NULL)
    gst_buffer_unref (buffer);

  if (tbuffer != NULL)
    asn_DEF_GstBuffer.free_struct (&asn_DEF_GstBuffer, tbuffer, 0);
}

GValue *
get_GValue (const GValue_t * val, KurentoMarshalRules rules, GError ** err)
{
  GValue *value = NULL;

  switch (val->value_type) {
    case GType_g_type_int:
      dec_int_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_uint:
      dec_uint_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_int64:
      dec_int64_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_uint64:
      dec_uint64_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_boolean:
      dec_boolean_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_date:
      dec_date_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_double:
      dec_double_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_float:
      dec_float_gvalue (rules, val, &value, err);
      break;
    case GType_g_type_string:
      dec_string_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_format:
      dec_format_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_fraction:
      dec_fraction_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_int_range:
      dec_int_range_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_date_time:
      dec_date_time_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_double_range:
      dec_double_range_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_caps:
      dec_caps_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_fraction_range:
      dec_fraction_range_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_list:
      dec_list_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_qos_type:
      dec_qos_type_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_seek_flags:
      dec_seek_flags_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_seek_type:
      dec_seek_type_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_segment:
      dec_segment_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_streaming_flags:
      dec_stream_flags_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_tag_list:
      dec_tag_list_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_toc:
      dec_toc_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_buffering_mode:
      dec_buffering_mode_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_array:
      dec_array_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_scheduling_flags:
      dec_scheduling_flags_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_context:
      dec_context_gvalue (rules, val, &value, err);
      break;
    case GType_gst_type_buffer:
      dec_buffer_gvalue (rules, val, &value, err);
      break;
    default:
      g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_UNKNOWN_TYPE,
          "GValue holds an unknown type");
      break;
  };

  return value;
}

gssize
enc_GValue (KurentoMarshalRules rules, const GValue * value, gchar * buf,
    gsize size, GError ** err)
{
  if (G_VALUE_HOLDS_INT (value))
    return ENCODE_GVALUE (value, rules, int, buf, size, err);
  else if (G_VALUE_HOLDS_UINT (value))
    return ENCODE_GVALUE (value, rules, uint, buf, size, err);
  else if (G_VALUE_HOLDS_INT64 (value))
    return ENCODE_GVALUE (value, rules, int64, buf, size, err);
  else if (G_VALUE_HOLDS_UINT64 (value))
    return ENCODE_GVALUE (value, rules, uint64, buf, size, err);
  else if (G_VALUE_HOLDS_BOOLEAN (value))
    return ENCODE_GVALUE (value, rules, boolean, buf, size, err);
  else if (G_VALUE_HOLDS_DATE (value))
    return ENCODE_GVALUE (value, rules, date, buf, size, err);
  else if (G_VALUE_HOLDS_DOUBLE (value))
    return ENCODE_GVALUE (value, rules, double, buf, size, err);
  else if (G_VALUE_HOLDS_FLOAT (value))
    return ENCODE_GVALUE (value, rules, float, buf, size, err);
  else if (G_VALUE_HOLDS_STRING (value))
    return ENCODE_GVALUE (value, rules, string, buf, size, err);
  else if (GST_VALUE_HOLDS_FORMAT (value))
    return ENCODE_GVALUE (value, rules, format, buf, size, err);
  else if (GST_VALUE_HOLDS_FRACTION (value))
    return ENCODE_GVALUE (value, rules, fraction, buf, size, err);
  else if (GST_VALUE_HOLDS_INT_RANGE (value))
    return ENCODE_GVALUE (value, rules, int_range, buf, size, err);
  else if (GST_VALUE_HOLDS_CAPS (value))
    return ENCODE_GVALUE (value, rules, caps, buf, size, err);
  else if (GST_VALUE_HOLDS_DATE_TIME (value))
    return ENCODE_GVALUE (value, rules, date_time, buf, size, err);
  else if (GST_VALUE_HOLDS_DOUBLE_RANGE (value))
    return ENCODE_GVALUE (value, rules, double_range, buf, size, err);
  else if (GST_VALUE_HOLDS_QOS_TYPE (value))
    return ENCODE_GVALUE (value, rules, qos_type, buf, size, err);
  else if (GST_VALUE_HOLDS_SEGMENT (value))
    return ENCODE_GVALUE (value, rules, segment, buf, size, err);
  else if (GST_VALUE_HOLDS_SEEK_FLAGS (value))
    return ENCODE_GVALUE (value, rules, seek_flags, buf, size, err);
  else if (GST_VALUE_HOLDS_SEEK_TYPE (value))
    return ENCODE_GVALUE (value, rules, seek_type, buf, size, err);
  else if (GST_VALUE_HOLDS_TOC (value))
    return ENCODE_GVALUE (value, rules, toc, buf, size, err);
  else if (GST_VALUE_HOLDS_FRACTION_RANGE (value))
    return ENCODE_GVALUE (value, rules, fraction_range, buf, size, err);
  else if (GST_VALUE_HOLDS_LIST (value))
    return ENCODE_GVALUE (value, rules, list, buf, size, err);
  else if (GST_VALUE_HOLDS_TAG_LIST (value))
    return ENCODE_GVALUE (value, rules, tag_list, buf, size, err);
  else if (GST_VALUE_HOLDS_STREAM_FLAGS (value))
    return ENCODE_GVALUE (value, rules, stream_flags, buf, size, err);
  else if (GST_VALUE_HOLDS_ARRAY (value))
    return ENCODE_GVALUE (value, rules, array, buf, size, err);
  else if (GST_VALUE_HOLDS_BUFFERING_MODE (value))
    return ENCODE_GVALUE (value, rules, buffering_mode, buf, size, err);
  else if (GST_VALUE_HOLDS_SCHEDULING_FLAGS (value))
    return ENCODE_GVALUE (value, rules, scheduling_flags, buf, size, err);
  else if (GST_VALUE_HOLDS_CONTEXT (value))
    return ENCODE_GVALUE (value, rules, context, buf, size, err);
  else if (GST_VALUE_HOLDS_BUFFER (value))
    return ENCODE_GVALUE (value, rules, buffer, buf, size, err);
  else {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_UNKNOWN_TYPE,
        "GValue holds an unknown type");
    return -1;
  }
}

gboolean
dec_GValue (KurentoMarshalRules rules, const gchar * buf, gsize size,
    GValue ** value, GError ** err)
{
  GValue_t *val = NULL;
  asn_dec_rval_t rval;
  gboolean ret;

  rval = decode_from_buffer (NULL, &asn_DEF_GValue, (void **) &val, buf,
      size, rules);

  if (!(ret = rval.code == RC_OK)) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "Could not decode GValue");
    goto end;
  }

  *value = get_GValue (val, rules, err);
  ret = (*value != NULL);

end:
  if (val != NULL)
    asn_DEF_GValue.free_struct (&asn_DEF_GValue, val, 0);

  return ret;
}
