/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <string.h>
#include "kmsassembler.h"

GST_DEBUG_CATEGORY_STATIC (kms_assembler_debug);
#define GST_CAT_DEFAULT kms_assembler_debug

GType _kms_assembler_type = 0;

struct _KmsAssembler
{
  GstMiniObject obj;

  KurentoMarshalRules rules;

  GSList *messages;
  KmsMessageType message_type;
  KmsDataType data_type;

  guint req_id;
  guint fragments;
};

GST_DEFINE_MINI_OBJECT_TYPE (KmsAssembler, kms_assembler);

static void
_priv_kms_assembler_initialize (void)
{
  _kms_assembler_type = kms_assembler_get_type ();

  GST_DEBUG_CATEGORY_INIT (kms_assembler_debug, "protocolassembler", 0,
      "protocol assembler");
}

static KmsAssembler *
_kms_assembler_copy (KmsAssembler * assembler)
{
  KmsAssembler *copy;

  g_return_val_if_fail (assembler != NULL, NULL);

  GST_DEBUG ("copy");

  copy = kms_assembler_new (assembler->rules);
  /* TODO: Copy private stuff in the new object */

  return copy;
}

static void
_kms_assembler_free (KmsAssembler * assembler)
{
  g_return_if_fail (assembler != NULL);

  GST_DEBUG ("free");

  if (assembler->messages != NULL)
    g_slist_free_full (assembler->messages, (GDestroyNotify) kms_message_unref);

  g_slice_free1 (sizeof (KmsAssembler), assembler);
}

KmsAssembler *
kms_assembler_new (KurentoMarshalRules rules)
{
  KmsAssembler *assembler;

  assembler = g_slice_new0 (KmsAssembler);

  gst_mini_object_init (GST_MINI_OBJECT_CAST (assembler), 0,
      _kms_assembler_type,
      (GstMiniObjectCopyFunction) _kms_assembler_copy, NULL,
      (GstMiniObjectFreeFunction) _kms_assembler_free);

  assembler->rules = rules;

  return KMS_ASSEMBLER (assembler);
}

gboolean
kms_assembler_append_message (KmsAssembler * assembler, KmsMessage * message)
{
  g_return_val_if_fail (assembler != NULL, FALSE);

  if (assembler->messages == NULL) {
    if (kms_message_get_fragment_id (message) != 0)
      return FALSE;

    assembler->data_type = kms_message_get_message_data_type (message);
    assembler->message_type = kms_message_get_message_type (message);
    assembler->fragments = kms_message_get_total_fragments (message);
    assembler->req_id = kms_message_get_req_id (message);
  } else {
    KmsMessage *last_msg;
    GSList *last;

    if (assembler->req_id != kms_message_get_req_id (message) ||
        assembler->fragments != kms_message_get_total_fragments (message) ||
        assembler->message_type != kms_message_get_message_type (message) ||
        assembler->data_type != kms_message_get_message_data_type (message))
      return FALSE;

    last = g_slist_last (assembler->messages);
    last_msg = KMS_MESSAGE (last->data);

    if (kms_message_get_fragment_id (last_msg) + 1 !=
        kms_message_get_fragment_id (message))
      /* Fragment is not consecutive. So the previous buffer has been missed */
      return FALSE;
  }

  assembler->messages = g_slist_append (assembler->messages, message);

  return TRUE;
}

gboolean
kms_assembler_is_completed (KmsAssembler * assembler)
{
  g_return_val_if_fail (assembler != NULL, FALSE);

  if (assembler->messages == NULL)
    return FALSE;

  return (assembler->fragments == g_slist_length (assembler->messages));
}

static void
calculate_size (KmsMessage * message, gsize * size)
{
  KmsDataType type;
  const gchar *buf;
  gsize s;

  kms_message_get_data (message, &type, &buf, &s);
  *size += s;
}

void
kms_assembler_compose_buffer (KmsAssembler * assembler, gchar ** buf,
    gsize * size)
{
  GSList *l;
  guint index = 0;

  g_return_if_fail (assembler != NULL);

  *size = 0;
  g_slist_foreach (assembler->messages, (GFunc) calculate_size, size);

  *buf = g_malloc (*size);
  for (l = assembler->messages; l != NULL; l = l->next) {
    KmsDataType type;
    const gchar *buffer;
    gsize s;

    kms_message_get_data (KMS_MESSAGE (l->data), &type, &buffer, &s);
    memcpy (*buf + index, buffer, s);
    index += s;
  }
}

KmsDataType
kms_assembler_get_data_type (const KmsAssembler * assembler)
{
  g_return_val_if_fail (assembler != NULL, KMS_DATA_TYPE_UNKNOWN);
  return assembler->data_type;
}

KurentoMarshalRules
kms_assembler_get_encoding_rules (const KmsAssembler * assembler)
{
  g_return_val_if_fail (assembler != NULL, KURENTO_MARSHALL_BER);
  return assembler->rules;
}

static void _priv_kms_assembler_initialize (void)
    __attribute__ ((constructor));
