/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "kmsfragmenter.h"

#define MAX_TRIES 10

GST_DEBUG_CATEGORY_STATIC (kms_fragmenter_debug);
#define GST_CAT_DEFAULT kms_fragmenter_debug

#define FRAGMENT_DATA_STRUCTURE(fragmenter, data, enc_func, buf, size, err) ({ \
  gint _try, _tries = MAX_TRIES;                                               \
  GError *_err = NULL;                                                         \
  gssize _coded = -1;                                                          \
  gsize _size;                                                                 \
                                                                               \
  if (size > 0) {                                                              \
    _size = size + fragmenter->max;                                            \
    _tries = 1;                                                                \
  } else {                                                                     \
    _size = fragmenter->max;                                                   \
  }                                                                            \
                                                                               \
  for (_try = 0; _try < _tries; _try++) {                                      \
    GST_LOG ("Try %u, size %" G_GSIZE_FORMAT, _try, _size);                    \
                                                                               \
    if (_err != NULL) {                                                        \
      g_error_free (_err);                                                     \
      _err = NULL;                                                             \
    }                                                                          \
                                                                               \
    buf = g_malloc (_size);                                                    \
    _coded = enc_func (fragmenter->rules, data, buf, _size, &_err);            \
                                                                               \
    if (_coded >= 0)                                                           \
      break;                                                                   \
                                                                               \
    g_free (buf);                                                              \
                                                                               \
    if (g_error_matches (_err, KURENTO_MARSHAL_ERROR,                          \
        KURENTO_UNKNOWN_TYPE)) {                                               \
      break;                                                                   \
    }                                                                          \
                                                                               \
    /* Try using a bigger buffer */                                            \
    _size += fragmenter->max / 2;                                              \
  }                                                                            \
                                                                               \
  if (_coded < 0) {                                                            \
    if (_err != NULL)                                                          \
      *err = _err;                                                             \
    else                                                                       \
      g_set_error (err, KURENTO_MARSHAL_ERROR,                                 \
        KURENTO_ENCODED_ERROR, "Fragmentation error");                         \
    buf = NULL;                                                                \
  }                                                                            \
                                                                               \
  _size = _coded;                                                              \
})

GType _kms_fragmenter_type = 0;

struct _KmsFragmenter
{
  GstMiniObject obj;

  KurentoMarshalRules rules;
  gsize max;
  gsize max_data;
  gboolean serialized;

  GSList *messages;
  KmsMessageType message_type;

  gchar *buf;
  gsize size;
};

GST_DEFINE_MINI_OBJECT_TYPE (KmsFragmenter, kms_fragmenter);

static void
_priv_kms_fragmenter_initialize (void)
{
  _kms_fragmenter_type = kms_fragmenter_get_type ();

  GST_DEBUG_CATEGORY_INIT (kms_fragmenter_debug, "fragmenter", 0, "fragmenter");
}

static KmsFragmenter *
_kms_fragmenter_copy (KmsFragmenter * fragmenter)
{
  KmsFragmenter *copy;

  g_return_val_if_fail (fragmenter != NULL, NULL);

  GST_DEBUG ("copy");

  copy = kms_fragmenter_new (fragmenter->rules, fragmenter->max);

  return copy;
}

static void
_kms_fragmenter_free (KmsFragmenter * fragmenter)
{
  g_return_if_fail (fragmenter != NULL);

  GST_DEBUG ("free");

  if (fragmenter->messages != NULL)
    g_slist_free_full (fragmenter->messages,
        (GDestroyNotify) kms_message_unref);

  if (fragmenter->buf != NULL)
    g_free (fragmenter->buf);

  g_slice_free1 (sizeof (KmsFragmenter), fragmenter);
}

static void
kms_fragmenter_adjust_data_size (KmsFragmenter * fragmenter)
{
  KmsMessage *msg;
  GError *err = NULL;
  gchar *test_buf;
  gsize size;
  gssize coded;

  size = fragmenter->max + fragmenter->max / 2;
  test_buf = g_malloc (size);

  /* encode a test buffer to know how many bytes it might need */

  msg = kms_message_new (G_MAXUINT);
  kms_message_set_fragment_id (msg, G_MAXUINT);
  kms_message_set_total_fragments (msg, G_MAXUINT);
  kms_message_set_message_type (msg, GST_MARSHALL_REQUEST);
  kms_message_set_data (msg, KMS_DATA_TYPE_UNKNOWN, g_malloc (fragmenter->max),
      fragmenter->max, TRUE);

  coded = enc_KmsMessage (fragmenter->rules, msg, test_buf, size, &err);

  if (coded < 0) {
    g_warning ("%s", err->message);
    g_error_free (err);
  }

  fragmenter->max_data = fragmenter->max - (coded - fragmenter->max);

  /* OCTET STRING packs length when it is encoded, we make sure */
  /* that even a 64 bit integer can fit */

  fragmenter->max_data -= 8;    /* 8 bytes */

  kms_message_unref (msg);
  g_free (test_buf);
}

static void
kms_fragmenter_split_buffer (KmsFragmenter * fragmenter, guint id,
    KmsDataType type, gchar * buf, gsize size)
{
  guint fragments, fragment, byte;

  fragment = byte = 0;

  fragments = size / fragmenter->max_data;
  if (size % fragmenter->max_data != 0)
    fragments++;

  while (byte < size) {
    KmsMessage *msg;
    gsize s;

    if (byte + fragmenter->max_data > size)
      s = size - byte;
    else
      s = fragmenter->max_data;

    msg = kms_message_new (id);
    kms_message_set_fragment_id (msg, fragment++);
    kms_message_set_total_fragments (msg, fragments);
    kms_message_set_message_type (msg, fragmenter->message_type);
    kms_message_set_data (msg, type, buf + byte, s, FALSE);

    fragmenter->messages = g_slist_append (fragmenter->messages, msg);

    byte += s;
  }
}

KmsFragmenter *
kms_fragmenter_new (KurentoMarshalRules rules, gsize max)
{
  KmsFragmenter *fragmenter;

  fragmenter = g_slice_new0 (KmsFragmenter);

  gst_mini_object_init (GST_MINI_OBJECT_CAST (fragmenter), 0,
      _kms_fragmenter_type,
      (GstMiniObjectCopyFunction) _kms_fragmenter_copy, NULL,
      (GstMiniObjectFreeFunction) _kms_fragmenter_free);

  fragmenter->rules = rules;
  fragmenter->max = max;

  kms_fragmenter_adjust_data_size (fragmenter);

  return KMS_FRAGMENTER (fragmenter);
}

void
update_messages (KmsMessage * msg, KmsFragmenter * fragmenter)
{
  kms_message_set_message_type (msg, fragmenter->message_type);
}

void
kms_fragmenter_set_message_type (KmsFragmenter * fragmenter,
    KmsMessageType type)
{
  g_return_if_fail (fragmenter != NULL);

  fragmenter->message_type = type;
  g_slist_foreach (fragmenter->messages, (GFunc) update_messages, fragmenter);
}

gboolean
kms_fragmenter_event (KmsFragmenter * fragmenter, guint id,
    const GstEvent * event, GError ** err)
{
  gssize coded;
  gchar *buf;

  g_return_val_if_fail (fragmenter != NULL, FALSE);
  coded =
      FRAGMENT_DATA_STRUCTURE (fragmenter, event, enc_GstEvent, buf, 0, err);

  if (coded < 0)
    return FALSE;

  kms_fragmenter_split_buffer (fragmenter, id, KMS_DATA_TYPE_EVENT, buf, coded);

  fragmenter->buf = buf;
  fragmenter->size = coded;

  fragmenter->serialized = GST_EVENT_IS_SERIALIZED (event);

  return TRUE;
}

gboolean
kms_fragmenter_query (KmsFragmenter * fragmenter, guint id,
    const GstQuery * query, GError ** err)
{
  gssize coded;
  gchar *buf;

  g_return_val_if_fail (fragmenter != NULL, FALSE);
  coded =
      FRAGMENT_DATA_STRUCTURE (fragmenter, query, enc_GstQuery, buf, 0, err);

  if (coded < 0) {
    return FALSE;
  }

  kms_fragmenter_split_buffer (fragmenter, id, KMS_DATA_TYPE_QUERY, buf, coded);

  fragmenter->buf = buf;
  fragmenter->size = coded;

  fragmenter->serialized = GST_QUERY_IS_SERIALIZED (query);

  return TRUE;
}

gboolean
kms_fragmenter_buffer (KmsFragmenter * fragmenter, guint id,
    const GstBuffer * buffer, GError ** err)
{
  GstMapInfo map;
  gssize coded;
  gchar *buf;
  gsize size;

  g_return_val_if_fail (fragmenter != NULL, FALSE);

  gst_buffer_map ((GstBuffer *) buffer, &map, GST_MAP_READ);
  size = map.size;
  gst_buffer_unmap ((GstBuffer *) buffer, &map);

  coded = FRAGMENT_DATA_STRUCTURE (fragmenter, buffer, enc_GstBuffer, buf, size,
      err);

  if (coded < 0) {
    return FALSE;
  }

  kms_fragmenter_split_buffer (fragmenter, id, KMS_DATA_TYPE_BUFFER, buf,
      coded);

  fragmenter->buf = buf;
  fragmenter->size = coded;

  fragmenter->serialized = TRUE;

  return TRUE;
}

guint
kms_fragmenter_n_messages (const KmsFragmenter * fragmenter)
{
  g_return_val_if_fail (fragmenter != NULL, 0);
  return g_slist_length (fragmenter->messages);
}

const KmsMessage *
kms_fragmenter_nth_message (const KmsFragmenter * fragmenter, guint n)
{
  GSList *l;

  g_return_val_if_fail (fragmenter != NULL, NULL);
  l = g_slist_nth (fragmenter->messages, n);

  return l->data;
}

void
kms_fragmenter_for_each_message (const KmsFragmenter * fragmenter, GFunc func,
    gpointer user_data)
{
  g_return_if_fail (fragmenter != NULL);
  g_slist_foreach (fragmenter->messages, func, user_data);
}

guint
kms_fragmenter_coded_data_size (const KmsFragmenter * fragmenter)
{
  g_return_val_if_fail (fragmenter != NULL, 0);
  return fragmenter->size;
}

KurentoMarshalRules
kms_fragmenter_get_encoding_rules (const KmsFragmenter * fragmenter)
{
  g_return_val_if_fail (fragmenter != NULL, KURENTO_MARSHALL_BER);
  return fragmenter->rules;
}

gboolean
kms_fragmenter_is_serialized (const KmsFragmenter * fragmenter)
{
  g_return_val_if_fail (fragmenter != NULL, FALSE);
  return fragmenter->serialized;
}

static void _priv_kms_fragmenter_initialize (void)
    __attribute__ ((constructor));
