/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <string.h>
#include "kmsmessage.h"

GST_DEBUG_CATEGORY_STATIC (kms_message_debug);
#define GST_CAT_DEFAULT kms_message_debug

GType _kms_message_type = 0;

struct _KmsMessage
{
  GstMiniObject obj;

  guint id;
  guint fragment;
  guint n_fragments;

  KmsMessageType type;
  KmsDataType data_type;
  guint32 counter_type;

  gboolean own;
  gchar *buf;
  gsize size;
};

GST_DEFINE_MINI_OBJECT_TYPE (KmsMessage, kms_message);

static void
_priv_kms_message_initialize (void)
{
  _kms_message_type = kms_message_get_type ();

  GST_DEBUG_CATEGORY_INIT (kms_message_debug, "protocolmessage", 0,
      "protocol message");
}

static KmsMessage *
_kms_message_copy (KmsMessage * msg)
{
  KmsMessage *copy;

  g_return_val_if_fail (msg != NULL, NULL);

  GST_DEBUG ("copy");

  copy = kms_message_new (msg->id);
  copy->fragment = msg->fragment;
  copy->n_fragments = msg->n_fragments;
  copy->type = msg->type;
  copy->data_type = msg->data_type;
  copy->buf = g_memdup (msg->buf, msg->size);
  copy->size = msg->size;
  copy->own = TRUE;

  return copy;
}

static void
_kms_message_free (KmsMessage * msg)
{
  g_return_if_fail (msg != NULL);

  GST_DEBUG ("free");

  if (msg->own && msg->buf != NULL)
    g_free (msg->buf);

  g_slice_free1 (sizeof (KmsMessage), msg);
}

KmsMessage *
kms_message_new (guint id)
{
  KmsMessage *msg;

  msg = g_slice_new0 (KmsMessage);

  gst_mini_object_init (GST_MINI_OBJECT_CAST (msg), 0,
      _kms_message_type,
      (GstMiniObjectCopyFunction) _kms_message_copy, NULL,
      (GstMiniObjectFreeFunction) _kms_message_free);

  msg->id = id;

  return KMS_MESSAGE (msg);
}

guint
kms_message_get_req_id (const KmsMessage * msg)
{
  g_return_val_if_fail (msg != NULL, 0);
  return msg->id;
}

void
kms_message_set_fragment_id (KmsMessage * msg, guint id)
{
  g_return_if_fail (msg != NULL);
  msg->fragment = id;
}

guint
kms_message_get_fragment_id (const KmsMessage * msg)
{
  g_return_val_if_fail (msg != NULL, 0);
  return msg->fragment;
}

void
kms_message_set_total_fragments (KmsMessage * msg, guint n)
{
  g_return_if_fail (msg != NULL);
  msg->n_fragments = n;
}

guint
kms_message_get_total_fragments (const KmsMessage * msg)
{
  g_return_val_if_fail (msg != NULL, 0);
  return msg->n_fragments;
}

void
kms_message_set_message_type (KmsMessage * msg, KmsMessageType type)
{
  g_return_if_fail (msg != NULL);
  msg->type = type;
}

KmsMessageType
kms_message_get_message_type (const KmsMessage * msg)
{
  g_return_val_if_fail (msg != NULL, 0);
  return msg->type;
}

KmsDataType
kms_message_get_message_data_type (const KmsMessage * msg)
{
  g_return_val_if_fail (msg != NULL, KMS_DATA_TYPE_UNKNOWN);
  return msg->data_type;
}

void
kms_message_set_data (KmsMessage * msg, KmsDataType data_type, gchar * buf,
    gsize size, gboolean own)
{
  g_return_if_fail (msg != NULL);

  msg->data_type = data_type;

  if (msg->own && msg->buf != NULL)
    g_free (msg->buf);

  msg->buf = buf;
  msg->own = own;
  msg->size = size;
}

void
kms_message_get_data (const KmsMessage * msg, KmsDataType * data_type,
    const gchar ** buf, gsize * size)
{
  g_return_if_fail (msg != NULL);

  *data_type = msg->data_type;
  *buf = msg->buf;
  *size = msg->size;
}

gboolean
kms_message_is_equal (const KmsMessage * msg1, const KmsMessage * msg2)
{
  g_return_val_if_fail (msg1 != NULL, FALSE);
  g_return_val_if_fail (msg2 != NULL, FALSE);

  if (msg1->id != msg2->id)
    return FALSE;

  if (msg1->counter_type != msg2->counter_type)
    return FALSE;

  if (msg1->fragment != msg2->fragment)
    return FALSE;

  if (msg1->n_fragments != msg2->n_fragments)
    return FALSE;

  if (msg1->type != msg2->type)
    return FALSE;

  if (msg1->data_type != msg2->data_type)
    return FALSE;

  if (msg1->size != msg2->size)
    return FALSE;

  return (memcmp (msg1->buf, msg2->buf, msg1->size) == 0);
}

void
kms_message_set_type_counter (KmsMessage * msg, guint n)
{
  g_return_if_fail (msg != NULL);

  msg->counter_type = n;
}

guint
kms_message_get_type_counter (const KmsMessage * msg)
{
  g_return_val_if_fail (msg != NULL, 0);

  return msg->counter_type;
}

static void _priv_kms_message_initialize (void)
    __attribute__ ((constructor));
