/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include "marshal_types.h"
#include "marshal_util.h"

#include "KmsProtocolMessage.h"

static KmsProtocolMessage_t *
get_marshallable_message (const KmsMessage * msg, KurentoMarshalRules rules,
    gchar * buf, gsize size, GError ** err)
{
  KmsProtocolMessage_t *message;
  KmsDataType type;

  message = g_new0 (KmsProtocolMessage_t, 1);
  message->type_counter = kms_message_get_type_counter (msg);
  message->req_id = kms_message_get_req_id (msg);
  message->fragment.fragment = kms_message_get_fragment_id (msg);
  message->fragment.total = kms_message_get_total_fragments (msg);

  switch (kms_message_get_message_type (msg)) {
    case GST_MARSHALL_REQUEST:
      message->message = KmsMessageType_gst_type_request;
      break;
    case GST_MARSHALL_RESPONSE:
      message->message = KmsMessageType_gst_type_response;
      break;
  }

  kms_message_get_data (msg, &type, (const gchar **) &message->data.buf,
      (gsize *) & message->data.size);

  switch (type) {
    case KMS_DATA_TYPE_UNKNOWN:
      message->type = KmsDataType_gst_apdu_type_unknown;
      break;
    case KMS_DATA_TYPE_QUERY:
      message->type = KmsDataType_gst_apdu_type_query;
      break;
    case KMS_DATA_TYPE_EVENT:
      message->type = KmsDataType_gst_apdu_type_event;
      break;
    case KMS_DATA_TYPE_BUFFER:
      message->type = KmsDataType_gst_apdu_type_buffer;
      break;
  }

  return message;
}

static KmsMessage *
get_KmsMessage (const KmsProtocolMessage_t * message,
    KurentoMarshalRules rules, GError ** err)
{
  KmsMessageType type = GST_MARSHALL_REQUEST;
  KmsDataType data_type;
  KmsMessage *msg;

  msg = kms_message_new (message->req_id);

  kms_message_set_type_counter (msg, message->type_counter);
  kms_message_set_fragment_id (msg, message->fragment.fragment);
  kms_message_set_total_fragments (msg, message->fragment.total);

  if (message->message == GST_MARSHALL_RESPONSE)
    type = GST_MARSHALL_RESPONSE;

  kms_message_set_message_type (msg, type);

  switch (message->type) {
    case KmsDataType_gst_apdu_type_query:
      data_type = KMS_DATA_TYPE_QUERY;
      break;
    case KmsDataType_gst_apdu_type_event:
      data_type = KMS_DATA_TYPE_EVENT;
      break;
    case KmsDataType_gst_apdu_type_buffer:
      data_type = KMS_DATA_TYPE_BUFFER;
      break;
    default:
      data_type = KMS_DATA_TYPE_UNKNOWN;
      break;
  }

  /* Let's make the Message owner of the data allocated by ASN1 to avoid */
  /* useless malloc and release operations on memory for better eficency */
  kms_message_set_data (msg, data_type, (gchar *) message->data.buf,
      message->data.size, TRUE);

  return msg;
}

gssize
enc_KmsMessage (KurentoMarshalRules rules, const KmsMessage * msg,
    gchar * buf, gsize size, GError ** err)
{
  KmsProtocolMessage_t *message = NULL;
  gssize s;

  message = get_marshallable_message (msg, rules, buf, size, err);
  if (message == NULL)
    return -1;

  s = ENCODE_MARSHAL_TYPE (KmsProtocolMessage, message, rules, buf, size, err);

  /* Buffer is now property of KmsMessage so do not release it */
  message->data.buf = NULL;
  message->data.size = 0;

  asn_DEF_KmsProtocolMessage.free_struct (&asn_DEF_KmsProtocolMessage, message,
      0);

  return s;
}

gboolean
dec_KmsMessage (KurentoMarshalRules rules, const gchar * buf,
    gsize size, KmsMessage ** msg, GError ** err)
{
  KmsProtocolMessage_t *message = NULL;
  asn_dec_rval_t rval;
  gboolean ret;

  rval = decode_from_buffer (NULL, &asn_DEF_KmsProtocolMessage,
      (void **) &message, buf, size, rules);

  if (!(ret = rval.code == RC_OK)) {
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_DECODED_ERROR,
        "Could not decode KmsProtocolMessage");
    goto end;
  }

  *msg = get_KmsMessage (message, rules, err);

  /* Optimization of memory allocated by ASN1. So it is owned by Message we */
  /* don't want it to be released here. */
  message->data.buf = NULL;
  message->data.size = 0;

end:
  if (message != NULL) {
    asn_DEF_KmsProtocolMessage.free_struct (&asn_DEF_KmsProtocolMessage,
        message, 0);
  }

  return ret;
}
