/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <glib.h>
#include "marshal_util.h"

void
own_allocated_memory (GstStructure_t * orig, GstStructure_t * dest)
{
  if (orig == NULL || dest == NULL)
    return;

  dest->name.buf = orig->name.buf;
  dest->name.size = orig->name.size;

  dest->fields.list.array = orig->fields.list.array;
  dest->fields.list.count = orig->fields.list.count;
  dest->fields.list.size = orig->fields.list.size;
  dest->fields.list.free = orig->fields.list.free;

  orig->name.buf = NULL;
  orig->name.size = 0;

  orig->fields.list.array = NULL;
  orig->fields.list.count = 0;
  orig->fields.list.size = 0;
}

/*
 * Argument type and callback necessary for xer_encode_to_buffer().
 */
typedef struct enc_to_buf_arg
{
  void *buffer;
  size_t left;
} enc_to_buf_arg;

static int
encode_to_buffer_cb (const void *buffer, size_t size, void *key)
{
  enc_to_buf_arg *arg = (enc_to_buf_arg *) key;

  if (arg->left < size) {
    /* Data exceeds the available buffer size */
    return -1;
  }

  memcpy (arg->buffer, buffer, size);
  arg->buffer = ((char *) arg->buffer) + size;
  arg->left -= size;

  return 0;
}

/*
 * A variant of the xer_encode() which encodes the data into the provided buffer
 */
static asn_enc_rval_t
xer_encode_to_buffer (asn_TYPE_descriptor_t * type_descriptor, void *struct_ptr,
    void *buffer, size_t buffer_size)
{
  enc_to_buf_arg arg;
  asn_enc_rval_t ec;

  arg.buffer = buffer;
  arg.left = buffer_size;

  ec = xer_encode (type_descriptor, struct_ptr, XER_F_BASIC,
      encode_to_buffer_cb, &arg);

  return ec;
}

asn_dec_rval_t
per_decode (struct asn_codec_ctx_s * opt_codec_ctx,
    struct asn_TYPE_descriptor_s * type_descriptor, void **struct_ptr,
    const void *buffer, size_t size)
{
  return uper_decode (opt_codec_ctx, type_descriptor, struct_ptr, buffer,
      size, 0, 0);
}

static asn_enc_rval_t (*encoders[]) (struct asn_TYPE_descriptor_s *
    type_descriptor, void *struct_ptr, void *buffer, size_t buffer_size) = {
der_encode_to_buffer, xer_encode_to_buffer, uper_encode_to_buffer};

static asn_dec_rval_t (*decoders[]) (struct asn_codec_ctx_s * opt_codec_ctx,
    struct asn_TYPE_descriptor_s * type_descriptor, void **struct_ptr,
    const void *buffer, size_t size) = {
ber_decode, xer_decode, per_decode};

asn_enc_rval_t
encode_to_buffer (struct asn_TYPE_descriptor_s *type_descriptor,
    void *struct_ptr, void *buffer, size_t buffer_size,
    KurentoMarshalRules rules)
{
  if (rules >= KURENTO_MARSHALL_BER && rules <= KURENTO_MARSHALL_PER) {
    return encoders[rules] (type_descriptor, struct_ptr, buffer, buffer_size);
  } else {
    asn_enc_rval_t fail = {
      -1, type_descriptor, struct_ptr
    };
    g_warning ("Invalid encoding rules provided.");
    return fail;
  }
}

asn_dec_rval_t
decode_from_buffer (struct asn_codec_ctx_s * opt_codec_ctx,
    struct asn_TYPE_descriptor_s * type_descriptor, void **struct_ptr,
    const void *buffer, size_t size, KurentoMarshalRules rules)
{
  if (rules >= KURENTO_MARSHALL_BER && rules <= KURENTO_MARSHALL_PER) {
    return decoders[rules] (opt_codec_ctx, type_descriptor, struct_ptr, buffer,
        size);
  } else {
    asn_dec_rval_t fail = {
      RC_FAIL, 0
    };
    g_warning ("Invalid encoding rules provided.");
    return fail;
  }
}
