/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#ifndef _KURENTO_MARSHAL_UTIL_
#define _KURENTO_MARSHAL_UTIL_

#include "asn_application.h"
#include "marshal.h"
#include "GstStructure.h"

#define CALCULATE_SIZE(ret, rules) ({         \
  gssize __size;                              \
  if ((rules) == KURENTO_MARSHALL_PER) {      \
    /* PER returns the number of encoded */   \
    /* bits in the .encoded fiels        */   \
    __size = ((ret.encoded + 7) / 8);         \
  } else {                                    \
    __size = (ret).encoded;                   \
  }                                           \
  __size;                                     \
})

#define ENCODE_MARSHAL_TYPE(type, struct_ptr, rules, buf, size, err) ({     \
  asn_enc_rval_t __ret;                                                     \
  gssize __size;                                                            \
  __ret = encode_to_buffer (&asn_DEF_##type, struct_ptr, buf, size, rules); \
  if (__ret.encoded == -1) {                                                \
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,         \
        "Could not encode %s", asn_DEF_##type.name);                        \
    __size = -1;                                                            \
  } else {                                                                  \
    __size = CALCULATE_SIZE (__ret, rules);                                 \
  }                                                                         \
  __size;                                                                   \
})

/* Missed macros in gstreamer */
#define G_VALUE_HOLDS_DATE(x) (G_VALUE_HOLDS((x), G_TYPE_DATE))
#define GST_VALUE_HOLDS_STREAM_FLAGS(x) (G_VALUE_HOLDS((x), GST_TYPE_STREAM_FLAGS))
#define GST_VALUE_HOLDS_FORMAT(x) (G_VALUE_HOLDS((x), GST_TYPE_FORMAT))
#define GST_VALUE_HOLDS_SEGMENT(x) (G_VALUE_HOLDS((x), GST_TYPE_SEGMENT))
#define GST_VALUE_HOLDS_QOS_TYPE(x) (G_VALUE_HOLDS((x), GST_TYPE_QOS_TYPE))
#define GST_VALUE_HOLDS_SEEK_FLAGS(x) (G_VALUE_HOLDS((x), GST_TYPE_SEEK_FLAGS))
#define GST_VALUE_HOLDS_SEEK_TYPE(x) (G_VALUE_HOLDS((x), GST_TYPE_SEEK_TYPE))
#define GST_VALUE_HOLDS_TAG_LIST(x) (G_VALUE_HOLDS((x), GST_TYPE_TAG_LIST))
#define GST_VALUE_HOLDS_TOC(x) (G_VALUE_HOLDS((x), GST_TYPE_TOC))
#define GST_VALUE_HOLDS_BUFFERING_MODE(x) (G_VALUE_HOLDS((x), GST_TYPE_BUFFERING_MODE))
#define GST_VALUE_HOLDS_SCHEDULING_FLAGS(x) (G_VALUE_HOLDS((x), GST_TYPE_SCHEDULING_FLAGS))
#define GST_VALUE_HOLDS_CONTEXT(x) (G_VALUE_HOLDS((x), GST_TYPE_CONTEXT))

#define OCTET_STRING_TO_INT64(buf, name, func)            \
({                                                        \
  g##name ival;                                           \
  ival = g_ascii_##func ((const gchar *) buf, NULL, 10);  \
  ival;                                                   \
})

#define INT64_TO_OCTET_STRING(ptr, format, number, size, label, err)   \
({                                                                     \
  gchar *_str = NULL;                                                  \
  gsize s = 0;                                                         \
                                                                       \
  _str = g_strdup_printf ("%" format, number);                         \
                                                                       \
  if (_str != NULL && (s = strlen (_str)) > size) {                    \
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,    \
        "Maximum size exceeded for %" format, number);                 \
    g_free (_str);                                                     \
    goto label;                                                        \
  }                                                                    \
                                                                       \
  if (OCTET_STRING_fromString (ptr, _str) < 0) {                       \
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,    \
        "Could not encode %" format, number);                          \
    g_free (_str);                                                     \
    goto label;                                                        \
  }                                                                    \
                                                                       \
  g_free (_str);                                                       \
})

#define INT64_AS_STRING_VALUE(value, type, t_descriptor, name, format, t_marshal, buf, size, rules, err) \
({                                                                     \
  asn_enc_rval_t _ret;                                                 \
  GValue_t *_val = NULL;                                               \
  type *_integer = NULL;                                               \
  gssize _s = 0;                                                       \
                                                                       \
  _integer = g_new0 (type, 1);                                         \
                                                                       \
  INT64_TO_OCTET_STRING(_integer, format, g_value_get_##name (value),  \
    size, _end, err);                                                       \
                                                                       \
  _ret = encode_to_buffer (&t_descriptor, _integer, buf, size, rules); \
  if (_ret.encoded == -1) {                                            \
    g_set_error (err, KURENTO_MARSHAL_ERROR, KURENTO_ENCODED_ERROR,    \
        "Could not encode %s", g_type_name(G_VALUE_TYPE(value)));      \
    goto _end;                                                         \
  }                                                                    \
                                                                       \
  _s = CALCULATE_SIZE (_ret, rules);                                   \
                                                                       \
  _val = create_marshallable_gvalue (t_marshal, buf, _s, rules, err);  \
                                                                       \
_end:                                                                  \
  if (_integer != NULL)                                                \
    t_descriptor.free_struct (&t_descriptor, _integer, 0);             \
                                                                       \
  _val;                                                                \
})

#define STRING_AS_INT64_VALUE(type, t_descriptor, name, g_type, tvalue, func, rules, err) \
({                                                                             \
  type *_integer = NULL;                                                       \
  GValue *_value = NULL;                                                       \
  g##name _ival;                                                               \
                                                                               \
  decode_marshallable_gvalue (&t_descriptor, tvalue, rules,                    \
      (void **) &_integer, err);                                               \
                                                                               \
  if (*err != NULL)                                                            \
    goto _end;                                                                 \
                                                                               \
  _ival = OCTET_STRING_TO_INT64(_integer->buf, name, func);                    \
                                                                               \
  _value = g_slice_new0 (GValue);                                              \
  g_value_init (_value, g_type);                                               \
  g_value_set_##name (_value, _ival);                                          \
                                                                               \
_end:                                                                          \
  if (_integer != NULL)                                                        \
    t_descriptor.free_struct (&t_descriptor, _integer, 0);                     \
                                                                               \
  _value;                                                                      \
})

#define GET_BIT(bit) (1 << (7 - bit))

#define BIT_STRING_SET_FLAG(bstring, flag)     \
({                                             \
  guint _byte;                                 \
                                               \
  _byte = (flag) / 8;                          \
                                               \
  if (_byte < (bstring)->size) {               \
    guint _offset;                             \
                                               \
    _offset = (flag) % 8;                      \
    (bstring)->buf[_byte] |= GET_BIT(_offset); \
  }                                            \
})

#define BIT_STRING_FLAG_IS_SET(bstring, flag)          \
({                                                     \
  guint _byte;                                         \
  gboolean _ret = FALSE;                               \
                                                       \
  _byte = (flag) / 8;                                  \
                                                       \
  if (_byte < (bstring)->size) {                       \
    guint _offset;                                     \
                                                       \
    _offset = (flag) % 8;                              \
    _ret = ((bstring)->buf[_byte] & GET_BIT(_offset)); \
  }                                                    \
  _ret;                                                \
})

void own_allocated_memory (GstStructure_t * orig, GstStructure_t * dest);

asn_enc_rval_t encode_to_buffer (struct asn_TYPE_descriptor_s *type_descriptor,
    void *struct_ptr, void *buffer, size_t buffer_size,
    KurentoMarshalRules rules);

asn_dec_rval_t decode_from_buffer (struct asn_codec_ctx_s *opt_codec_ctx,
    struct asn_TYPE_descriptor_s *type_descriptor, void **struct_ptr,
    const void *buffer, size_t size, KurentoMarshalRules rules);

#endif /* _KURENTO_MARSHAL_UTIL_ */
