/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <math.h>
#include <gst/gst.h>
#include "marshal_test.h"
#include "marshal_util.h"

#define GST_CAT_DEFAULT kms_marshal_test
GST_DEBUG_CATEGORY_STATIC (GST_CAT_DEFAULT);
#define GST_DEFAULT_NAME "kms_marshal_test"

#define STR_SEGMENT_FORMAT(ptr) \
  "flags: %d, rate: %f, applied_rate: %f, format: %d, base: %"   \
  G_GUINT64_FORMAT ", offset: %" G_GUINT64_FORMAT ", start: %"   \
  G_GUINT64_FORMAT ", stop %" G_GUINT64_FORMAT ", time: %"       \
  G_GUINT64_FORMAT ", position %" G_GUINT64_FORMAT               \
  ", duration %" G_GUINT64_FORMAT, ptr->flags, ptr->rate,        \
  ptr->applied_rate, ptr->format, ptr->base, ptr->offset,        \
  ptr->start, ptr->stop, ptr->time, ptr->position, ptr->duration

static gboolean equal_entry_list (GList *, GList *);

static gboolean
equal_gavlue_int (const GValue * v1, const GValue * v2)
{
  gint n1, n2;

  if (!G_VALUE_HOLDS_INT (v2)) {
    GST_ERROR ("GValue does not hold an INT value");
    return FALSE;
  }

  n1 = g_value_get_int (v1);
  n2 = g_value_get_int (v2);

  GST_DEBUG ("INT values: %d == %d", n1, n2);

  return (n1 == n2);
}

static gboolean
equal_gvalue_uint (const GValue * v1, const GValue * v2)
{
  gint n1, n2;

  if (!G_VALUE_HOLDS_UINT (v2)) {
    GST_ERROR ("GValue does not hold an UINT value");
    return FALSE;
  }

  n1 = g_value_get_uint (v1);
  n2 = g_value_get_uint (v2);

  GST_DEBUG ("UINT values: %u == %u", n1, n2);

  return (n1 == n2);
}

static gboolean
equal_gvalue_date (const GValue * v1, const GValue * v2)
{
  GDate *d1, *d2;

  if (!G_VALUE_HOLDS_DATE (v2)) {
    GST_ERROR ("GValue does not hold an Date value");
    return FALSE;
  }

  d1 = g_value_get_boxed (v1);
  d2 = g_value_get_boxed (v2);

  if (d1 == NULL && d2 == NULL)
    return TRUE;

  if ((d1 == NULL && d2 != NULL) || (d1 != NULL && d2 == NULL))
    return FALSE;

  GST_DEBUG ("INT values: %d/%d/%d == %d/%d/%d ",
      g_date_get_day (d1), g_date_get_month (d1), g_date_get_year (d1),
      g_date_get_day (d2), g_date_get_month (d2), g_date_get_year (d2));

  return (g_date_get_julian (d1) == g_date_get_julian (d2));
}

static gboolean
equal_gavlue_uint64 (const GValue * v1, const GValue * v2)
{
  guint64 n1, n2;

  if (!G_VALUE_HOLDS_UINT64 (v2)) {
    GST_ERROR ("GValue does not hold a UINT64 value");
    return FALSE;
  }

  n1 = g_value_get_uint64 (v1);
  n2 = g_value_get_uint64 (v2);

  GST_DEBUG ("UINT64 values: %" G_GUINT64_FORMAT " == %" G_GUINT64_FORMAT, n1,
      n2);

  return (n1 == n2);
}

static gboolean
equal_gvalue_boolean (const GValue * v1, const GValue * v2)
{
  gint n1, n2;

  if (!G_VALUE_HOLDS_BOOLEAN (v2)) {
    GST_ERROR ("GValue does not hold an BOOLEAN value");
    return FALSE;
  }

  n1 = g_value_get_boolean (v1);
  n2 = g_value_get_boolean (v2);

  GST_DEBUG ("BOOLEAN values: %s == %s", (n1) ? "TRUE" : "FALSE",
      (n2) ? "TRUE" : "FALSE");

  return (n1 == n2);
}

static gboolean
equal_gvalue_float (const GValue * v1, const GValue * v2)
{
  gfloat f1, f2;

  if (!G_VALUE_HOLDS_FLOAT (v2)) {
    GST_ERROR ("GValue does not hold an FLOAT value");
    return FALSE;
  }

  f1 = g_value_get_float (v1);
  f2 = g_value_get_float (v2);

  GST_DEBUG ("FLOAT values: %5.3f == %5.3f", f1, f2);

  return f1 == f2;
}

static gboolean
equal_gvalue_string (const GValue * v1, const GValue * v2)
{
  const gchar *str1, *str2;

  if (!G_VALUE_HOLDS_STRING (v2)) {
    GST_ERROR ("GValue does not hold an STRING value");
    return FALSE;
  }

  str1 = g_value_get_string (v1);
  str2 = g_value_get_string (v2);

  GST_DEBUG ("String values: %s == %s", str1, str2);

  return g_strcmp0 (str1, str2) == 0;
}

static gboolean
equal_gvalue_fraction (const GValue * v1, const GValue * v2)
{
  gint num1, den1, num2, den2;

  if (!GST_VALUE_HOLDS_FRACTION (v2)) {
    GST_ERROR ("GValue does not hold an FRACTION value");
    return FALSE;
  }

  num1 = gst_value_get_fraction_numerator (v1);
  den1 = gst_value_get_fraction_denominator (v1);

  num2 = gst_value_get_fraction_numerator (v2);
  den2 = gst_value_get_fraction_denominator (v2);

  GST_DEBUG ("FRACTION values: %d/%d == %d/%d", num1, den1, num2, den2);

  return (num1 == num2 && den1 == den2);
}

static gboolean
equal_gvalue_int_range (const GValue * v1, const GValue * v2)
{
  gint min1, max1, min2, max2;

  if (!GST_VALUE_HOLDS_INT_RANGE (v2)) {
    GST_ERROR ("GValue does not hold an INT_RANGE value");
    return FALSE;
  }

  max1 = gst_value_get_int_range_max (v1);
  min1 = gst_value_get_int_range_min (v1);

  max2 = gst_value_get_int_range_max (v2);
  min2 = gst_value_get_int_range_min (v2);

  GST_DEBUG ("INTEGER RANGE values: [%d,%d] == [%d,%d]", min1, max1, min2,
      max2);

  return (min1 == min2 && max1 == max2);
}

static gboolean
fequal (double a, double b)
{
  double epsilon = 0.000000001;

  return fabs (a - b) < epsilon;
}

static gboolean
equal_gvalue_double (const GValue * v1, const GValue * v2)
{
  gdouble d1, d2;

  if (!G_VALUE_HOLDS_DOUBLE (v2)) {
    GST_ERROR ("GValue does not hold a DOUBLE value");
    return FALSE;
  }

  d1 = g_value_get_double (v1);
  d2 = g_value_get_double (v2);

  GST_DEBUG ("DOUBLE values: %f == %f", d1, d2);

  return fequal (d1, d2);
}

static gboolean
equal_gvalue_double_range (const GValue * v1, const GValue * v2)
{
  gdouble min1, max1, min2, max2;

  if (!GST_VALUE_HOLDS_DOUBLE_RANGE (v2)) {
    GST_ERROR ("GValue does not hold an DOUBLE_RANGE value");
    return FALSE;
  }

  max1 = gst_value_get_double_range_max (v1);
  min1 = gst_value_get_double_range_min (v1);

  max2 = gst_value_get_double_range_max (v2);
  min2 = gst_value_get_double_range_min (v2);

  GST_DEBUG ("DOUBLE RANGE values: [%lf,%lf] == [%lf,%lf]", min1, max1, min2,
      max2);

  return (fequal (min1, min2) && fequal (max1, max2));
}

static gboolean
equal_gvalue_fraction_range (const GValue * v1, const GValue * v2)
{
  gint min_num1, max_num1, min_den1, max_den1;
  gint min_num2, max_num2, min_den2, max_den2;
  const GValue *min, *max;

  if (!GST_VALUE_HOLDS_FRACTION_RANGE (v2)) {
    GST_ERROR ("GValue does not hold an FRACTION_RANGE value");
    return FALSE;
  }

  min = gst_value_get_fraction_range_min (v1);
  min_num1 = gst_value_get_fraction_numerator (min);
  min_den1 = gst_value_get_fraction_denominator (min);
  max = gst_value_get_fraction_range_max (v1);
  max_num1 = gst_value_get_fraction_numerator (max);
  max_den1 = gst_value_get_fraction_denominator (max);

  min = gst_value_get_fraction_range_min (v2);
  min_num2 = gst_value_get_fraction_numerator (min);
  min_den2 = gst_value_get_fraction_denominator (min);
  max = gst_value_get_fraction_range_max (v2);
  max_num2 = gst_value_get_fraction_numerator (max);
  max_den2 = gst_value_get_fraction_denominator (max);

  GST_DEBUG ("FRACTION RANGE values: [%d/%d,%d/%d] == [%d/%d,%d/%d]", min_num1,
      min_den1, max_num1, max_den1, min_num2, min_den2, max_num2, max_den2);

  return (gst_value_compare (v1, v2) == GST_VALUE_EQUAL);
}

static gboolean
equal_gvalue_array (const GValue * v1, const GValue * v2)
{
  gboolean ret = TRUE;
  guint i;

  if (!GST_VALUE_HOLDS_ARRAY (v2)) {
    GST_ERROR ("GValue does not hold an ARRAY value");
    return FALSE;
  }

  if (gst_value_array_get_size (v1) != gst_value_array_get_size (v2)) {
    GST_ERROR ("Arrays have different length");
    return FALSE;
  }

  for (i = 0; i < gst_value_array_get_size (v1); i++) {
    const GValue *val1, *val2;

    val1 = gst_value_array_get_value (v1, i);
    val2 = gst_value_array_get_value (v2, i);
    ret &= gvalue_are_equal (val1, val2);
  }

  return ret;
}

static gboolean
equal_gvalue_list (const GValue * v1, const GValue * v2)
{
  gboolean ret = TRUE;
  guint i;

  if (!GST_VALUE_HOLDS_LIST (v2)) {
    GST_ERROR ("GValue does not hold a LIST  value");
    return FALSE;
  }

  if (gst_value_list_get_size (v1) != gst_value_list_get_size (v2)) {
    GST_ERROR ("Lists have different length");
    return FALSE;
  }

  for (i = 0; i < gst_value_list_get_size (v1); i++) {
    const GValue *val1, *val2;

    val1 = gst_value_list_get_value (v1, i);
    val2 = gst_value_list_get_value (v2, i);
    ret &= gvalue_are_equal (val1, val2);
  }

  return ret;
}

static gboolean
equal_gavlue_int64 (const GValue * v1, const GValue * v2)
{
  gint64 n1, n2;

  if (!G_VALUE_HOLDS_INT64 (v2)) {
    GST_ERROR ("GValue does not hold an INT64 value");
    return FALSE;
  }

  n1 = g_value_get_int64 (v1);
  n2 = g_value_get_int64 (v2);

  GST_DEBUG ("INT64 values: %" G_GINT64_FORMAT " == %" G_GINT64_FORMAT, n1, n2);

  return (n1 == n2);
}

static gboolean
equal_gvalue_stream_flags (const GValue * v1, const GValue * v2)
{
  GstStreamFlags f1, f2;

  if (!GST_VALUE_HOLDS_STREAM_FLAGS (v2)) {
    GST_ERROR ("GValue does not hold Stream flags value");
    return FALSE;
  }

  f1 = (GstStreamFlags) g_value_get_flags (v1);
  f2 = (GstStreamFlags) g_value_get_flags (v2);

  GST_DEBUG ("Stream Flags values: %d == %d", f1, f2);

  return (f1 == f2);
}

static gboolean
equal_gvalue_segment (const GValue * v1, const GValue * v2)
{
  GstSegment *s1, *s2;

  if (!GST_VALUE_HOLDS_SEGMENT (v2)) {
    GST_ERROR ("GValue does not hold Stream flags value");
    return FALSE;
  }

  s1 = g_value_get_boxed (v1);
  s2 = g_value_get_boxed (v2);

  if (s1 == NULL && s2 == NULL)
    return TRUE;

  if ((s1 == NULL && s2 != NULL) || (s1 != NULL && s2 == NULL))
    return FALSE;

  GST_DEBUG ("(Segment) " STR_SEGMENT_FORMAT (s1));
  GST_DEBUG ("(Segment) " STR_SEGMENT_FORMAT (s2));

  return (s1->flags == s1->flags && fequal (s1->rate, s2->rate) &&
      fequal (s1->applied_rate, s2->applied_rate) && s1->format == s2->format &&
      s1->base == s2->base && s1->offset == s2->offset &&
      s1->start == s2->start && s1->stop == s2->stop && s1->time == s2->time &&
      s1->position == s2->position && s1->duration == s2->duration);
}

static gboolean
equal_gvalue_format (const GValue * v1, const GValue * v2)
{
  GstFormat f1, f2;

  if (!GST_VALUE_HOLDS_FORMAT (v2)) {
    GST_ERROR ("GValue does not hold FORMAT value");
    return FALSE;
  }

  f1 = (GstFormat) g_value_get_enum (v1);
  f2 = (GstFormat) g_value_get_enum (v2);

  GST_DEBUG ("Format values %d == %d", f1, f2);

  return f1 == f2;
}

static gboolean
equal_gvalue_qos_type (const GValue * v1, const GValue * v2)
{
  GstQOSType q1, q2;

  if (!GST_VALUE_HOLDS_QOS_TYPE (v2)) {
    GST_ERROR ("GValue does not hold QOS Type value");
    return FALSE;
  }

  q1 = (GstQOSType) g_value_get_enum (v1);
  q2 = (GstQOSType) g_value_get_enum (v2);

  GST_DEBUG ("QOS type values %d == %d", q1, q2);

  return q1 == q2;
}

static gboolean
equal_gvalue_seek_flags_type (const GValue * v1, const GValue * v2)
{
  GstSeekFlags f1, f2;

  if (!GST_VALUE_HOLDS_SEEK_FLAGS (v2)) {
    GST_ERROR ("GValue does not hold seek flags value");
    return FALSE;
  }

  f1 = (GstSeekFlags) g_value_get_flags (v1);
  f2 = (GstSeekFlags) g_value_get_flags (v2);

  GST_DEBUG ("Seek Flag values %d == %d", f1, f2);

  return f1 == f2;
}

static gboolean
equal_gvalue_seek_type (const GValue * v1, const GValue * v2)
{
  GstSeekType st1, st2;

  if (!GST_VALUE_HOLDS_SEEK_TYPE (v2)) {
    GST_ERROR ("GValue does not hold a Seek Type value");
    return FALSE;
  }

  st1 = (GstSeekType) g_value_get_enum (v1);
  st2 = (GstSeekType) g_value_get_enum (v2);

  GST_DEBUG ("Seek type values %d == %d", st1, st2);

  return st1 == st2;
}

static gboolean
equal_gvalue_caps (const GValue * v1, const GValue * v2)
{
  GstCaps *c1, *c2;

  if (!GST_VALUE_HOLDS_CAPS (v2)) {
    GST_ERROR ("GValue does not hold CAPs value");
    return FALSE;
  }

  c1 = g_value_get_boxed (v1);
  c2 = g_value_get_boxed (v2);

  GST_DEBUG ("Cap Values %" GST_PTR_FORMAT " ==  %" GST_PTR_FORMAT, c1, c2);

  if (c1 == NULL && c2 == NULL)
    return TRUE;

  if ((c1 == NULL && c2 != NULL) || (c1 != NULL && c2 == NULL))
    return FALSE;

  return gst_caps_is_equal (c1, c2);
}

static gboolean
are_tag_lists_equal (GstTagList * l1, GstTagList * l2)
{
  gchar *str1, *str2;

  str1 = gst_tag_list_to_string (l1);
  str2 = gst_tag_list_to_string (l2);

  GST_DEBUG ("Tag list Values %s ==  %s", str1, str2);

  g_free (str1);
  g_free (str2);

  return gst_tag_list_is_equal (l1, l2);
}

static gboolean
equal_gvalue_tag_list (const GValue * v1, const GValue * v2)
{
  GstTagList *l1, *l2;

  if (!GST_VALUE_HOLDS_TAG_LIST (v2)) {
    GST_ERROR ("GValue does not hold a tag list value");
    return FALSE;
  }

  l1 = g_value_get_boxed (v1);
  l2 = g_value_get_boxed (v2);

  if (l1 == NULL && l2 == NULL)
    return TRUE;

  if ((l1 == NULL && l2 != NULL) || (l1 != NULL && l2 == NULL))
    return FALSE;

  return are_tag_lists_equal (l1, l2);
}

static gboolean
equal_gvalue_date_time (const GValue * v1, const GValue * v2)
{
  GstDateTime *d1, *d2;
  gchar *str1, *str2;

  if (!GST_VALUE_HOLDS_DATE_TIME (v2)) {
    GST_ERROR ("GValue does not hold date time value");
    return FALSE;
  }

  d1 = g_value_get_boxed (v1);
  d2 = g_value_get_boxed (v2);

  if (d1 == NULL && d2 == NULL)
    return TRUE;

  if ((d1 == NULL && d2 != NULL) || (d1 != NULL && d2 == NULL))
    return FALSE;

  str1 = gst_date_time_to_iso8601_string (d1);
  str2 = gst_date_time_to_iso8601_string (d2);

  GST_DEBUG ("Date Time Values %s ==  %s", str1, str2);

  g_free (str1);
  g_free (str2);

  return (gst_date_time_get_year (d1) == gst_date_time_get_year (d2) &&
      gst_date_time_get_month (d1) == gst_date_time_get_month (d2) &&
      gst_date_time_get_day (d1) == gst_date_time_get_day (d2) &&
      gst_date_time_get_hour (d1) == gst_date_time_get_hour (d2) &&
      gst_date_time_get_minute (d1) == gst_date_time_get_minute (d2) &&
      gst_date_time_get_second (d1) == gst_date_time_get_second (d2) &&
      fequal (gst_date_time_get_time_zone_offset (d1),
          gst_date_time_get_time_zone_offset (d2)));
}

static gboolean
equal_entries (const GstTocEntry * e1, const GstTocEntry * e2)
{
  const gchar *uid1, *uid2;
  GstTagList *tl1, *tl2;
  GList *l1, *l2;
  gint64 start1, start2, stop1, stop2;
  GstTocEntryType type1, type2;

  type1 = gst_toc_entry_get_entry_type (e1);
  type2 = gst_toc_entry_get_entry_type (e2);

  GST_DEBUG ("Entry type %d == %d", type1, type2);

  if (type1 != type2)
    return FALSE;

  uid1 = gst_toc_entry_get_uid (e1);
  uid2 = gst_toc_entry_get_uid (e2);

  GST_DEBUG ("Entry uids %s == %s", uid1, uid2);

  if (g_strcmp0 (uid1, uid2) != 0)
    return FALSE;

  gst_toc_entry_get_start_stop_times (e1, &start1, &stop1);
  gst_toc_entry_get_start_stop_times (e2, &start2, &stop2);

  GST_DEBUG ("Entry start %" G_GINT64_FORMAT " == %" G_GINT64_FORMAT,
      start1, start2);

  GST_DEBUG ("Entry stop %" G_GINT64_FORMAT " == %" G_GINT64_FORMAT,
      stop1, stop2);

  if (start1 != start2 || stop1 != stop2)
    return FALSE;

  tl1 = gst_toc_entry_get_tags (e1);
  tl2 = gst_toc_entry_get_tags (e2);

  if (tl1 == NULL && tl2 != NULL)
    return FALSE;

  if (tl1 != NULL && tl2 == NULL)
    return FALSE;

  if (tl1 != NULL && tl2 != NULL) {
    if (!are_tag_lists_equal (tl1, tl2))
      return FALSE;
  }

  l1 = gst_toc_entry_get_sub_entries (e1);
  l2 = gst_toc_entry_get_sub_entries (e2);

  if (l1 == NULL && l2 == NULL)
    return TRUE;

  if (l1 != NULL && l2 != NULL)
    return equal_entry_list (l1, l2);

  return FALSE;
}

static gboolean
equal_entry_list (GList * list1, GList * list2)
{
  GList *l1, *l2;

  if (g_list_length (list1) != g_list_length (list2))
    return FALSE;

  for (l1 = list1, l2 = list2; l1 != NULL && l2 != NULL; l1 = l1->next,
      l2 = l2->next) {
    GstTocEntry *e1, *e2;

    e1 = l1->data;
    e2 = l2->data;

    if (!equal_entries (e1, e2))
      return FALSE;
  }

  return TRUE;
}

static gboolean
equal_gvalue_toc (const GValue * v1, const GValue * v2)
{
  GstToc *t1, *t2;
  GstTagList *tl1, *tl2;
  GList *l1, *l2;

  if (!GST_VALUE_HOLDS_TOC (v2)) {
    GST_ERROR ("GValue does not hold toc value");
    return FALSE;
  }

  t1 = g_value_get_boxed (v1);
  t2 = g_value_get_boxed (v2);

  if (t1 == NULL && t2 == NULL)
    return TRUE;

  if ((t1 == NULL && t2 != NULL) || (t1 != NULL && t2 == NULL))
    return FALSE;

  GST_DEBUG ("TOC scope %u == %u", gst_toc_get_scope (t1),
      gst_toc_get_scope (t2));
  if (gst_toc_get_scope (t1) != gst_toc_get_scope (t2))
    return FALSE;

  tl1 = gst_toc_get_tags (t1);
  tl2 = gst_toc_get_tags (t2);

  if (tl1 == NULL && tl2 != NULL)
    return FALSE;

  if (tl1 != NULL && tl2 == NULL)
    return FALSE;

  if (tl1 != NULL && tl2 != NULL) {
    if (!are_tag_lists_equal (tl1, tl2))
      return FALSE;
  }

  l1 = gst_toc_get_entries (t1);
  l2 = gst_toc_get_entries (t2);

  if (l1 == NULL && l2 == NULL)
    return TRUE;

  if (l1 != NULL && l2 != NULL)
    return equal_entry_list (l1, l2);

  return FALSE;
}

static gboolean
equal_gvalue_buffering_mode (const GValue * v1, const GValue * v2)
{
  GstBufferingMode b1, b2;

  if (!GST_VALUE_HOLDS_BUFFERING_MODE (v2)) {
    GST_ERROR ("GValue does not hold GstBufferingMode value");
    return FALSE;
  }

  b1 = (GstBufferingMode) g_value_get_enum (v1);
  b2 = (GstBufferingMode) g_value_get_enum (v2);

  GST_DEBUG ("GstBufferingMode values %d == %d", b1, b2);

  return b1 == b2;
}

static gboolean
equal_gvalue_scheduling_flags (const GValue * v1, const GValue * v2)
{
  GstSchedulingFlags f1, f2;

  if (!GST_VALUE_HOLDS_SCHEDULING_FLAGS (v2)) {
    GST_ERROR ("GValue does not hold GstSchedulingFlags value");
    return FALSE;
  }

  f1 = (GstStreamFlags) g_value_get_flags (v1);
  f2 = (GstStreamFlags) g_value_get_flags (v2);

  GST_DEBUG ("Scheduling Flags values: %d == %d", f1, f2);

  return (f1 == f2);
}

static gboolean
equal_structures (const GstStructure * st1, const GstStructure * st2)
{
  gint i, fields;

  if (st1 == NULL && st2 == NULL)
    return TRUE;

  if ((st1 != NULL && st2 == NULL) || (st1 == NULL && st2 != NULL))
    return FALSE;

  if (g_strcmp0 (gst_structure_get_name (st1),
          gst_structure_get_name (st2)) != 0)
    return FALSE;

  if ((fields = gst_structure_n_fields (st1)) != gst_structure_n_fields (st2))
    return FALSE;

  for (i = 0; i < fields; i++) {
    const gchar *key;
    const GValue *v1, *v2;

    key = gst_structure_nth_field_name (st1, i);

    if (!gst_structure_has_field (st2, key))
      return FALSE;

    v1 = gst_structure_get_value (st1, key);
    v2 = gst_structure_get_value (st2, key);

    if (!gvalue_are_equal (v1, v2))
      return FALSE;
  }

  return TRUE;
}

static gboolean
equal_gvalue_context (const GValue * v1, const GValue * v2)
{
  GstContext *c1, *c2;
  const GstStructure *s1, *s2;

  if (!GST_VALUE_HOLDS_CONTEXT (v2)) {
    GST_ERROR ("GValue does not hold GstContext value");
    return FALSE;
  }

  c1 = g_value_get_boxed (v1);
  c2 = g_value_get_boxed (v2);

  if (c1 == NULL && c2 == NULL)
    return TRUE;

  if ((c1 == NULL && c2 != NULL) || (c1 != NULL && c2 == NULL))
    return FALSE;

  if (g_strcmp0 (gst_context_get_context_type (c1),
          gst_context_get_context_type (c2)) != 0)
    return FALSE;

  if (gst_context_is_persistent (c1) != gst_context_is_persistent (c2)) {
    GST_ERROR ("Persistan %u != %u", gst_context_is_persistent (c1),
        gst_context_is_persistent (c2));
    return FALSE;
  }

  s1 = gst_context_get_structure (c1);
  s2 = gst_context_get_structure (c2);

  return (equal_structures (s1, s2));
}

static gboolean
equal_gvalue_buffer (const GValue * v1, const GValue * v2)
{
  GstBuffer *b1, *b2;
  GstMapInfo m1, m2;
  gboolean ret;

  if (!GST_VALUE_HOLDS_BUFFER (v2)) {
    GST_ERROR ("GValue does not hold GstBuffer value");
    return FALSE;
  }

  b1 = g_value_get_boxed (v1);
  b2 = g_value_get_boxed (v2);

  if (b1 == NULL && b2 == NULL)
    return TRUE;

  if ((b1 == NULL && b2 != NULL) || (b1 != NULL && b2 == NULL))
    return FALSE;

  gst_buffer_map (b1, &m1, GST_MAP_READ);
  gst_buffer_map (b2, &m2, GST_MAP_READ);

  ret = (m1.size == m2.size);
  if (!ret || m1.size == 0)
    goto mem_checked;

  ret = (memcmp (m1.data, m2.data, m1.size) == 0);

mem_checked:
  gst_buffer_unmap (b1, &m1);
  gst_buffer_unmap (b2, &m2);

  if (!ret)
    return FALSE;

  GST_DEBUG ("GstBuffer %" GST_PTR_FORMAT " == %" GST_PTR_FORMAT, b1, b2);

  return (GST_BUFFER_PTS (b1) == GST_BUFFER_PTS (b2) &&
      GST_BUFFER_DTS (b1) == GST_BUFFER_DTS (b2) &&
      GST_BUFFER_DURATION (b1) == GST_BUFFER_DURATION (b2) &&
      GST_BUFFER_OFFSET (b1) == GST_BUFFER_OFFSET (b2) &&
      GST_BUFFER_OFFSET_END (b1) == GST_BUFFER_OFFSET_END (b2) &&
      GST_BUFFER_FLAGS (b1) == GST_BUFFER_FLAGS (b2));
}

gboolean
gvalue_are_equal (const GValue * v1, const GValue * v2)
{
  if (G_VALUE_HOLDS_INT (v1))
    return equal_gavlue_int (v1, v2);
  else if (G_VALUE_HOLDS_UINT (v1))
    return equal_gvalue_uint (v1, v2);
  if (G_VALUE_HOLDS_INT64 (v1))
    return equal_gavlue_int64 (v1, v2);
  else if (G_VALUE_HOLDS_BOOLEAN (v1))
    return equal_gvalue_boolean (v1, v2);
  else if (G_VALUE_HOLDS_DATE (v1))
    return equal_gvalue_date (v1, v2);
  else if (G_VALUE_HOLDS_DOUBLE (v1))
    return equal_gvalue_double (v1, v2);
  else if (G_VALUE_HOLDS_FLOAT (v1))
    return equal_gvalue_float (v1, v2);
  else if (G_VALUE_HOLDS_STRING (v1))
    return equal_gvalue_string (v1, v2);
  if (G_VALUE_HOLDS_UINT64 (v1))
    return equal_gavlue_uint64 (v1, v2);
  else if (GST_VALUE_HOLDS_FORMAT (v1))
    return equal_gvalue_format (v1, v2);
  else if (GST_VALUE_HOLDS_FRACTION (v1))
    return equal_gvalue_fraction (v1, v2);
  else if (GST_VALUE_HOLDS_INT_RANGE (v1))
    return equal_gvalue_int_range (v1, v2);
  else if (GST_VALUE_HOLDS_CAPS (v1))
    return equal_gvalue_caps (v1, v2);
  else if (GST_VALUE_HOLDS_DATE_TIME (v1))
    return equal_gvalue_date_time (v1, v2);
  else if (GST_VALUE_HOLDS_DOUBLE_RANGE (v1))
    return equal_gvalue_double_range (v1, v2);
  else if (GST_VALUE_HOLDS_FRACTION_RANGE (v1))
    return equal_gvalue_fraction_range (v1, v2);
  else if (GST_VALUE_HOLDS_ARRAY (v1))
    return equal_gvalue_array (v1, v2);
  else if (GST_VALUE_HOLDS_LIST (v1))
    return equal_gvalue_list (v1, v2);
  else if (GST_VALUE_HOLDS_SEGMENT (v1))
    return equal_gvalue_segment (v1, v2);
  else if (GST_VALUE_HOLDS_STREAM_FLAGS (v1))
    return equal_gvalue_stream_flags (v1, v2);
  else if (GST_VALUE_HOLDS_QOS_TYPE (v1))
    return equal_gvalue_qos_type (v1, v2);
  else if (GST_VALUE_HOLDS_SEEK_FLAGS (v1))
    return equal_gvalue_seek_flags_type (v1, v2);
  else if (GST_VALUE_HOLDS_SEEK_TYPE (v1))
    return equal_gvalue_seek_type (v1, v2);
  else if (GST_VALUE_HOLDS_TAG_LIST (v1))
    return equal_gvalue_tag_list (v1, v2);
  else if (GST_VALUE_HOLDS_TOC (v1))
    return equal_gvalue_toc (v1, v2);
  else if (GST_VALUE_HOLDS_BUFFERING_MODE (v1))
    return equal_gvalue_buffering_mode (v1, v2);
  else if (GST_VALUE_HOLDS_SCHEDULING_FLAGS (v1))
    return equal_gvalue_scheduling_flags (v1, v2);
  else if (GST_VALUE_HOLDS_CONTEXT (v1))
    return equal_gvalue_context (v1, v2);
  else if (GST_VALUE_HOLDS_BUFFER (v1))
    return equal_gvalue_buffer (v1, v2);
  else {
    GST_ERROR ("Unknown value %s", g_type_name (G_VALUE_TYPE (v1)));
    return FALSE;
  }
}

static void init_debug (void) __attribute__ ((constructor));

static void
init_debug (void)
{
  GST_DEBUG_CATEGORY_INIT (GST_CAT_DEFAULT, GST_DEFAULT_NAME, 0,
      GST_DEFAULT_NAME);
}
