/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include <glib.h>
#include "marshal_types.h"

#define BUFFER_LENGTH (1024 * 6)        /* bytes */
#define TEST_BUFFER "This is just for testing purposes"

static gboolean
equal_buffers (GstBuffer * b1, GstBuffer * b2)
{
  GstMapInfo m1, m2;
  gboolean ret;

  if (b1 == NULL && b2 == NULL)
    return TRUE;

  if ((b1 == NULL && b2 != NULL) || (b1 != NULL && b2 == NULL))
    return FALSE;

  gst_buffer_map (b1, &m1, GST_MAP_READ);
  gst_buffer_map (b2, &m2, GST_MAP_READ);

  ret = (m1.size == m2.size);
  if (!ret || m1.size == 0)
    goto mem_checked;

  ret = (memcmp (m1.data, m2.data, m1.size) == 0);

mem_checked:
  gst_buffer_unmap (b1, &m1);
  gst_buffer_unmap (b2, &m2);

  if (!ret)
    return FALSE;

  GST_DEBUG ("GstBuffer %" GST_PTR_FORMAT " == %" GST_PTR_FORMAT, b1, b2);

  return (GST_BUFFER_PTS (b1) == GST_BUFFER_PTS (b2) &&
      GST_BUFFER_DTS (b1) == GST_BUFFER_DTS (b2) &&
      GST_BUFFER_DURATION (b1) == GST_BUFFER_DURATION (b2) &&
      GST_BUFFER_OFFSET (b1) == GST_BUFFER_OFFSET (b2) &&
      GST_BUFFER_OFFSET_END (b1) == GST_BUFFER_OFFSET_END (b2) &&
      GST_BUFFER_FLAGS (b1) == GST_BUFFER_FLAGS (b2));
}

static void
test_marshal_GstBuffer (KurentoMarshalRules rules)
{
  GstBuffer *b1 = NULL, *b2 = NULL;
  GError *err = NULL;
  gchar buf[BUFFER_LENGTH];
  gsize size;

  b1 = gst_buffer_new_wrapped (g_strdup (TEST_BUFFER), strlen (TEST_BUFFER));

  size = enc_GstBuffer (rules, b1, buf, BUFFER_LENGTH, &err);

  if (err != NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed encoding GValue holding a list value");
    goto end;
  }

  /* Encode GVal 2 */
  if (!dec_GstBuffer (rules, buf, size, &b2, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GValue holding a list value");
    goto end;
  }

  if (!equal_buffers (b1, b2)) {
    fail ("Decoded buffers are not equal");
  }

  GST_DEBUG ("DONE");

end:
  if (err != NULL)
    g_error_free (err);

  if (b1 != NULL)
    gst_buffer_unref (b1);

  if (b2 != NULL)
    gst_buffer_unref (b2);
}

GST_START_TEST (marshal_buffer_test)
{
  test_marshal_GstBuffer (KURENTO_MARSHALL_BER);
  test_marshal_GstBuffer (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* gstbuffer_marshalling test suite    */
/***************************************/
static Suite *
gstcaps_marshalling_suite (void)
{
  Suite *s = suite_create ("gstcaps");
  TCase *tc_chain = tcase_create ("marshalling");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, marshal_buffer_test);

  return s;
}

GST_CHECK_MAIN (gstcaps_marshalling);
