/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include <glib.h>
#include "marshal_types.h"

#define BUFFER_LENGTH (1024 * 6)        /* bytes */

static const gchar non_simple_caps_string[] =
    "video/x-raw, format=(string)I420, framerate=(fraction)[ 1/100, 100 ], "
    "width=(int)[ 16, 4096 ], height=(int)[ 16, 4096 ]; video/x-raw, "
    "format=(string)YUY2, framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
    "height=(int)[ 16, 4096 ]; video/x-raw, format=(string)RGB8_PALETTED, "
    "framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
    "height=(int)[ 16, 4096 ]; video/x-raw, "
    "format=(string){ I420, YUY2, YV12 }, width=(int)[ 16, 4096 ], "
    "height=(int)[ 16, 4096 ], framerate=(fraction)[ 1/100, 100 ]";

static void
test1_marshal_GstCaps (KurentoMarshalRules rules)
{
  GstCaps *c1 = NULL, *c2 = NULL;
  gchar buf[BUFFER_LENGTH];
  GError *err = NULL;
  gsize size;

  c1 = gst_caps_from_string (non_simple_caps_string);

  /* Encode GVal 1 */
  size = enc_GstCaps (rules, c1, buf, BUFFER_LENGTH, &err);
  if (err != NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed encoding GValue holding an string value");
    goto end;
  }

  /* Encode GVal 2 */
  if (!dec_GstCaps (rules, buf, size, &c2, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GValue holding an integer value");
    goto end;
  }

  if (!gst_caps_is_equal (c1, c2)) {
    GST_ERROR ("Value encoded and value decoded are not the same");
    fail ("Marshal did not succeed");
  }

  GST_DEBUG ("capabalities encoded: %" GST_PTR_FORMAT, c1);
  GST_DEBUG ("capabalities decoded: %" GST_PTR_FORMAT, c2);

end:
  if (err != NULL)
    g_error_free (err);

  if (c1 != NULL)
    gst_caps_unref (c1);

  if (c2 != NULL)
    gst_caps_unref (c2);
}

GST_START_TEST (marshal_gstcaps_test1)
{
  test1_marshal_GstCaps (KURENTO_MARSHALL_BER);
  test1_marshal_GstCaps (KURENTO_MARSHALL_PER);
}

GST_END_TEST static void
test_marshal_GstCapsFeatures (KurentoMarshalRules rules)
{
  GstCaps *c1, *c2;
  GstStructure *s1;
  GstCapsFeatures *f1;
  gchar buf[BUFFER_LENGTH];
  GError *err = NULL;
  gsize size;

  c1 = gst_caps_from_string
      ("audio/x-raw, format=(string) { S16_LE, S16_BE, U16_LE, U16_BE }, rate=(int)[ 1, 2147483647 ], channels=(int)[ 1, 2 ]; audio/x-raw, format=(string) { S16_LE, S16_BE, U16_LE, U16_BE }, rate=(int)[ 1, 2147483647 ], channels=(int)[ 1, 11 ]; audio/x-raw, format=(string) { S16_LE, S16_BE, U16_LE, U16_BE }, rate=(int)[ 1, 2147483647 ], channels=(int)[ 1, 11 ]");

  s1 = gst_structure_new ("video/x-raw", "width", G_TYPE_INT, 320, "height",
      GST_TYPE_INT_RANGE, 240, 260, NULL);

  f1 = gst_caps_features_new ("memory:EGLImage", NULL);

  gst_caps_append_structure_full (c1, s1, f1);

  /* Encode GVal 1 */
  size = enc_GstCaps (rules, c1, buf, BUFFER_LENGTH, &err);
  if (err != NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed encoding GValue holding an string value");
    goto end;
  }

  /* Encode GVal 2 */
  if (!dec_GstCaps (rules, buf, size, &c2, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GValue holding an integer value");
    goto end;
  }

  if (!gst_caps_is_equal (c1, c2)) {
    GST_ERROR ("Value encoded and value decoded are not the same");
    fail ("Marshal did not succeed");
  }

  GST_DEBUG ("capabalities encoded: %" GST_PTR_FORMAT, c1);
  GST_DEBUG ("capabalities decoded: %" GST_PTR_FORMAT, c2);

end:
  if (err != NULL)
    g_error_free (err);

  if (c1 != NULL)
    gst_caps_unref (c1);

  if (c2 != NULL)
    gst_caps_unref (c2);
}

GST_START_TEST (marshal_gstcaps_features)
{
  test_marshal_GstCapsFeatures (KURENTO_MARSHALL_BER);
  test_marshal_GstCapsFeatures (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* gstcaps_marshalling test suite */
/***************************************/
static Suite *
gstcaps_marshalling_suite (void)
{
  Suite *s = suite_create ("gstcaps");
  TCase *tc_chain = tcase_create ("marshalling");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, marshal_gstcaps_test1);
  tcase_add_test (tc_chain, marshal_gstcaps_features);

  return s;
}

GST_CHECK_MAIN (gstcaps_marshalling);
