/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include "marshal_types.h"
#include "marshal_test.h"

#define BUFFER_LENGTH (1024 * 6)        /* bytes */

#define GST_QUARK_FLAGS 46
#define GST_QUARK_EVENT_STREAM_START 160
#define GST_QUARK_STREAM_ID 161

static guint id = 0;

static GstTocEntryType types[] = {
  GST_TOC_ENTRY_TYPE_ANGLE,
  GST_TOC_ENTRY_TYPE_VERSION,
  GST_TOC_ENTRY_TYPE_EDITION,
  GST_TOC_ENTRY_TYPE_INVALID,
  GST_TOC_ENTRY_TYPE_TITLE,
  GST_TOC_ENTRY_TYPE_TRACK,
  GST_TOC_ENTRY_TYPE_CHAPTER
};

static gboolean
are_equal (GstEvent * event1, GstEvent * event2)
{
  const GstStructure *st1, *st2;
  gint i, fields;

  GST_DEBUG ("Event 1: %" GST_PTR_FORMAT, event1);
  GST_DEBUG ("Event 2: %" GST_PTR_FORMAT, event2);

  if (GST_EVENT_TIMESTAMP (event1) != GST_EVENT_TIMESTAMP (event2))
    return FALSE;

  st1 = gst_event_get_structure (event1);
  st2 = gst_event_get_structure (event2);

  if (st1 == NULL && st2 == NULL)
    return TRUE;

  if (st1 != NULL && st2 == NULL)
    return FALSE;

  if (st1 == NULL && st2 != NULL)
    return FALSE;

  /* gst_structure_is_equal fails managing certain kind of object like */
  /* GstSegments so we need to do this stuff manually */

  if (g_strcmp0 (gst_structure_get_name (st1),
          gst_structure_get_name (st2)) != 0)
    return FALSE;

  if ((fields = gst_structure_n_fields (st1)) != gst_structure_n_fields (st2))
    return FALSE;

  for (i = 0; i < fields; i++) {
    const gchar *key;
    const GValue *v1, *v2;

    key = gst_structure_nth_field_name (st1, i);

    if (!gst_structure_has_field (st2, key))
      return FALSE;

    v1 = gst_structure_get_value (st1, key);
    v2 = gst_structure_get_value (st2, key);
    if (!gvalue_are_equal (v1, v2))
      return FALSE;
  }

  return TRUE;
}

static void
test_event (GstEvent * event1, KurentoMarshalRules rules)
{
  GstEvent *event2 = NULL;
  gchar buf[BUFFER_LENGTH];
  GError *err = NULL;
  gsize size;

  /* Encode GVal 1 */
  size = enc_GstEvent (rules, event1, buf, BUFFER_LENGTH, &err);
  if (err != NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed encoding GstEvent %" GST_PTR_FORMAT, event1);
    goto end;
  }

  /* Encode GVal 2 */
  if (!dec_GstEvent (rules, buf, size, &event2, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GstEvent");
    goto end;
  }

  if (!are_equal (event1, event2)) {
    GST_ERROR ("Event encoded and event decoded are not the same");
    fail ("Marshal did not succeed");
  }

end:
  if (err != NULL)
    g_error_free (err);

  if (event2 != NULL)
    gst_event_unref (event2);
}

static void
test_stream_start_event_flags (KurentoMarshalRules rules)
{
  GstEvent *event;
  GstStructure *s;
  const gchar *stream_id = "test-start";

  event = gst_event_new_stream_start (stream_id);
  test_event (event, rules);
  gst_event_unref (event);

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, stream_id,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_SPARSE, NULL);
  event = gst_event_new_custom (GST_EVENT_STREAM_START, s);
  test_event (event, rules);
  gst_event_unref (event);

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, stream_id,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_SELECT, NULL);
  event = gst_event_new_custom (GST_EVENT_STREAM_START, s);
  test_event (event, rules);
  gst_event_unref (event);

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, stream_id,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_UNSELECT, NULL);
  event = gst_event_new_custom (GST_EVENT_STREAM_START, s);
  test_event (event, rules);
  gst_event_unref (event);

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, stream_id,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_SPARSE |
      GST_STREAM_FLAG_SELECT | GST_STREAM_FLAG_UNSELECT, NULL);
  event = gst_event_new_custom (GST_EVENT_STREAM_START, s);
  test_event (event, rules);
  gst_event_unref (event);
}

static void
test_event_new_segment (KurentoMarshalRules rules)
{
  GstSegment *segment;
  GstEvent *event;

  segment = gst_segment_new ();
  gst_segment_init (segment, GST_FORMAT_BUFFERS);
  event = gst_event_new_segment (segment);
  test_event (event, rules);
  gst_event_unref (event);
  gst_segment_free (segment);

  segment = gst_segment_new ();
  gst_segment_init (segment, GST_FORMAT_BYTES);
  segment->rate = 2.5;
  segment->applied_rate = 2.6;
  segment->base = 55;
  segment->offset = G_MAXUINT64;
  segment->start = 34;
  segment->position = 234;
  segment->flags |= GST_SEGMENT_FLAG_SKIP;
  event = gst_event_new_segment (segment);
  test_event (event, rules);
  gst_event_unref (event);
  gst_segment_free (segment);

  segment = gst_segment_new ();
  gst_segment_init (segment, GST_FORMAT_TIME);
  segment->rate = G_MAXDOUBLE;
  segment->applied_rate = G_MAXDOUBLE;
  segment->base = G_MAXUINT64;
  segment->offset = G_MAXUINT64;
  segment->start = G_MAXUINT64;
  segment->position = G_MAXUINT64;
  segment->flags |= GST_SEGMENT_FLAG_SEGMENT;
  event = gst_event_new_segment (segment);
  test_event (event, rules);
  gst_event_unref (event);
  gst_segment_free (segment);

  segment = gst_segment_new ();
  gst_segment_init (segment, GST_FORMAT_PERCENT);
  segment->rate = G_MINDOUBLE;
  segment->applied_rate = G_MINDOUBLE;
  segment->base = 0;
  segment->offset = 0;
  segment->start = 0;
  segment->position = 0;
  segment->flags |= GST_SEGMENT_FLAG_SEGMENT | GST_SEGMENT_FLAG_RESET;
  event = gst_event_new_segment (segment);
  test_event (event, rules);
  gst_event_unref (event);
  gst_segment_free (segment);
}

static void
test_event_navigation (KurentoMarshalRules rules)
{
  GstEvent *event;
  GstStructure *s;
  const gchar *str = "test-navigation";

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, str,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_SPARSE, NULL);
  event = gst_event_new_navigation (s);
  test_event (event, rules);
  gst_event_unref (event);

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, str,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_SELECT, NULL);
  event = gst_event_new_navigation (s);
  test_event (event, rules);
  gst_event_unref (event);

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, str,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_UNSELECT, NULL);
  event = gst_event_new_navigation (s);
  test_event (event, rules);
  gst_event_unref (event);

  s = gst_structure_new_id (GST_QUARK_EVENT_STREAM_START,
      GST_QUARK_STREAM_ID, G_TYPE_STRING, str,
      GST_QUARK_FLAGS, GST_TYPE_STREAM_FLAGS, GST_STREAM_FLAG_SPARSE |
      GST_STREAM_FLAG_SELECT | GST_STREAM_FLAG_UNSELECT, NULL);
  event = gst_event_new_navigation (s);
  test_event (event, rules);
  gst_event_unref (event);
}

static void
test_event_tag (KurentoMarshalRules rules)
{
  GstEvent *event;
  GstTagList *list;
  GDate *date;

  date = g_date_new_dmy (23, 1, 1982);

  list = gst_tag_list_new_empty ();

  gst_tag_list_add (list, GST_TAG_MERGE_APPEND, GST_TAG_TITLE, "test1",
      GST_TAG_TITLE, "test2", GST_TAG_TITLE_SORTNAME, "test-title",
      GST_TAG_ALBUM, "test-album", GST_TAG_DATE, date, NULL);

  event = gst_event_new_tag (list);
  test_event (event, rules);
  gst_event_unref (event);
  g_date_free (date);
}

static GstTagList *
create_tag_list ()
{
  GstTagList *taglist;
  GstDateTime *datetime;
  GDate *date;
  gchar *title, *name, *album;
  guint n = id++;

  title = g_strdup_printf ("title-%d", n);
  name = g_strdup_printf ("name-%d", n);
  album = g_strdup_printf ("album-%d", n);

  datetime = gst_date_time_new_now_utc ();
  date = g_date_new_dmy (gst_date_time_get_day (datetime),
      gst_date_time_get_month (datetime), gst_date_time_get_year (datetime));

  taglist = gst_tag_list_new_empty ();
  gst_tag_list_add (taglist, GST_TAG_MERGE_APPEND,
      GST_TAG_TITLE, title,
      GST_TAG_TITLE_SORTNAME, name,
      GST_TAG_ALBUM, album,
      GST_TAG_DATE, date,
      GST_TAG_DATE_TIME, datetime,
      GST_TAG_SHOW_SEASON_NUMBER, 2,
      GST_TAG_SHOW_EPISODE_NUMBER, G_MAXUINT,
      GST_TAG_IMAGE_ORIENTATION, "flip-rotate-270",
      GST_TAG_USER_RATING, 10,
      GST_TAG_GEO_LOCATION_HORIZONTAL_ERROR, 3.0, NULL);

  g_free (title);
  g_free (name);
  g_free (album);

  gst_date_time_unref (datetime);
  g_date_free (date);

  return taglist;
}

static GstTocEntry *
create_toc_entry ()
{
  GstTocEntry *entry;
  GstTocEntryType type;
  gchar *uid;
  guint n = id++;

  uid = g_strdup_printf ("uid-%d", n);

  type = types[n % G_N_ELEMENTS (types)];

  entry = gst_toc_entry_new (type, uid);
  gst_toc_entry_set_start_stop_times (entry, n, n + 500);

  if (n % 2 == 0) {
    GstTagList *taglist;

    taglist = create_tag_list ();
    gst_toc_entry_set_tags (entry, taglist);
  }

  g_free (uid);

  return entry;
}

static void
test_event_toc (KurentoMarshalRules rules)
{
  GstTagList *taglist;
  GstTocEntry *entry;
  GstEvent *event;
  GstToc *toc;

  toc = gst_toc_new (GST_TOC_SCOPE_GLOBAL);
  taglist = create_tag_list ();
  gst_toc_set_tags (toc, taglist);

  taglist = create_tag_list ();
  entry = create_toc_entry ();

  gst_toc_entry_set_tags (entry, taglist);
  gst_toc_append_entry (toc, entry);

  taglist = create_tag_list (1);
  entry = create_toc_entry (1);

  gst_toc_entry_set_tags (entry, taglist);
  gst_toc_append_entry (toc, entry);

  event = gst_event_new_toc (toc, TRUE);
  test_event (event, rules);

  gst_event_unref (event);
  gst_toc_unref (toc);
}

static void
marshal_GstEvent (KurentoMarshalRules rules)
{
  GstEvent *event;
  GstCaps *caps;

  event = gst_event_new_flush_start ();
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_flush_stop (FALSE);
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_flush_stop (TRUE);
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_eos ();
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_gap (1, 2);
  test_event (event, rules);
  gst_event_unref (event);

  test_stream_start_event_flags (rules);

  test_event_new_segment (rules);

  test_event_tag (rules);

  event = gst_event_new_segment_done (GST_FORMAT_DEFAULT, G_MAXINT64);
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_segment_done (GST_FORMAT_BYTES, 0);
  test_event (event, rules);
  gst_event_unref (event);

  event =
      gst_event_new_buffer_size (GST_FORMAT_BUFFERS, G_MAXINT64, G_MAXINT64,
      TRUE);
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_buffer_size (GST_FORMAT_PERCENT, 0, G_MAXINT64, FALSE);
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_qos (GST_QOS_TYPE_OVERFLOW, 1.0, 5, 6);
  test_event (event, rules);
  gst_event_unref (event);

  event =
      gst_event_new_qos (GST_QOS_TYPE_THROTTLE, G_MAXDOUBLE, G_MAXINT64,
      G_MAXINT64);
  test_event (event, rules);
  gst_event_unref (event);

  event =
      gst_event_new_seek (G_MAXDOUBLE, GST_FORMAT_DEFAULT, GST_SEEK_FLAG_NONE,
      GST_SEEK_TYPE_SET, G_MAXINT64, GST_SEEK_TYPE_END, G_MAXINT64);
  test_event (event, rules);
  gst_event_unref (event);

  test_event_navigation (rules);

  event = gst_event_new_latency (G_MAXUINT64);
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_latency (0);
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_step (GST_FORMAT_TIME, 6, 2.4, TRUE, FALSE);
  test_event (event, rules);
  gst_event_unref (event);

  event =
      gst_event_new_step (GST_FORMAT_BYTES, G_MAXUINT64, G_MAXDOUBLE, FALSE,
      FALSE);
  test_event (event, rules);
  gst_event_unref (event);

  /*TODO: Evaluate if gst_event_new_sink_message should be marshalled */

  event = gst_event_new_reconfigure ();
  test_event (event, rules);
  gst_event_unref (event);

  caps = gst_caps_new_simple ("video/x-raw",
      "format", G_TYPE_STRING, "I420",
      "width", G_TYPE_INT, 384,
      "height", G_TYPE_INT, 288, "framerate", GST_TYPE_FRACTION, 25, 1, NULL);
  event = gst_event_new_caps (caps);
  test_event (event, rules);
  gst_event_unref (event);
  gst_caps_unref (caps);

  test_event_toc (rules);

  event = gst_event_new_toc_select ("toc-select");
  test_event (event, rules);
  gst_event_unref (event);

  event = gst_event_new_segment_done (GST_FORMAT_BYTES, 346);
  test_event (event, rules);
  gst_event_unref (event);
}

GST_START_TEST (marshal_gstevents_test)
{
  marshal_GstEvent (KURENTO_MARSHALL_BER);
  /* Some tests need a bigger buffer to be encoded using XER */
//   marshal_GstEvent (KURENTO_MARSHALL_XER);
  marshal_GstEvent (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* gstevents_marshalling test suite */
/***************************************/
static Suite *
gstcaps_marshalling_suite (void)
{
  Suite *s = suite_create ("gstevents");
  TCase *tc_chain = tcase_create ("marshalling");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, marshal_gstevents_test);

  return s;
}

GST_CHECK_MAIN (gstcaps_marshalling);
