/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include "marshal_types.h"
#include "marshal_test.h"

#define BUFFER_LENGTH (1024 * 6)        /* bytes */

static const gchar non_simple_caps_string[] =
    "video/x-raw, format=(string)I420, framerate=(fraction)[ 1/100, 100 ], "
    "width=(int)[ 16, 4096 ], height=(int)[ 16, 4096 ]; video/x-raw, "
    "format=(string)YUY2, framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
    "height=(int)[ 16, 4096 ]; video/x-raw, format=(string)RGB8_PALETTED, "
    "framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
    "height=(int)[ 16, 4096 ]; video/x-raw, "
    "format=(string){ I420, YUY2, YV12 }, width=(int)[ 16, 4096 ], "
    "height=(int)[ 16, 4096 ], framerate=(fraction)[ 1/100, 100 ]";

static gboolean
are_equal (GstQuery * query1, GstQuery * query2)
{
  const GstStructure *st1, *st2;
  gint i, fields;

  GST_DEBUG ("Query 1: %" GST_PTR_FORMAT, query1);
  GST_DEBUG ("Query 2: %" GST_PTR_FORMAT, query2);

  if (GST_QUERY_TYPE (query1) != GST_QUERY_TYPE (query2))
    return FALSE;

  st1 = gst_query_get_structure (query1);
  st2 = gst_query_get_structure (query2);

  if (st1 == NULL && st2 == NULL)
    return TRUE;

  if (st1 != NULL && st2 == NULL)
    return FALSE;

  if (st1 == NULL && st2 != NULL)
    return FALSE;

  /* gst_structure_is_equal fails managing certain kind of object like */
  /* GstSegments so we need to do this stuff manually */

  if (g_strcmp0 (gst_structure_get_name (st1),
          gst_structure_get_name (st2)) != 0)
    return FALSE;

  if ((fields = gst_structure_n_fields (st1)) != gst_structure_n_fields (st2))
    return FALSE;

  for (i = 0; i < fields; i++) {
    const gchar *key;
    const GValue *v1, *v2;

    key = gst_structure_nth_field_name (st1, i);

    if (!gst_structure_has_field (st2, key))
      return FALSE;

    v1 = gst_structure_get_value (st1, key);
    v2 = gst_structure_get_value (st2, key);
    if (!gvalue_are_equal (v1, v2))
      return FALSE;
  }

  return TRUE;
}

static void
test_query (GstQuery * query1, KurentoMarshalRules rules)
{
  GstQuery *query2 = NULL;
  gchar buf[BUFFER_LENGTH];
  GError *err = NULL;
  gsize size;

  /* Encode GVal 1 */
  size = enc_GstQuery (rules, query1, buf, BUFFER_LENGTH, &err);
  if (err != NULL) {
    GST_ERROR ("Failed encoding GstQuery %" GST_PTR_FORMAT, query1);
    fail ("Error %s", err->message);
    goto end;
  }

  /* Encode GVal 2 */
  if (!dec_GstQuery (rules, buf, size, &query2, &err)) {
    GST_ERROR ("Failed decoding GstQuery");
    fail ("%s", err->message);
    goto end;
  }

  if (!are_equal (query1, query2)) {
    GST_ERROR ("Query encoded and query decoded are not the same");
    fail ("Marshal did not succeed");
  }

end:
  if (err != NULL)
    g_error_free (err);

  if (query2 != NULL)
    gst_query_unref (query2);
}

static void
marshal_GstQuery (KurentoMarshalRules rules)
{
//   GstBufferPool *pool;
  GstQuery *query;
  GstCaps *caps;

  query = gst_query_new_convert (GST_FORMAT_BYTES, G_MAXINT64,
      GST_FORMAT_BUFFERS);
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_position (GST_FORMAT_PERCENT);
  test_query (query, rules);
  gst_query_set_position (query, GST_FORMAT_PERCENT, G_MAXINT64);
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_duration (GST_FORMAT_BUFFERS);
  test_query (query, rules);
  gst_query_set_duration (query, GST_FORMAT_BUFFERS, G_MININT64);
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_latency ();
  test_query (query, rules);
  gst_query_set_latency (query, TRUE, 0, G_MAXUINT64);
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_seeking (GST_FORMAT_TIME);
  test_query (query, rules);
  gst_query_set_seeking (query, GST_FORMAT_TIME, FALSE, G_MININT64, G_MAXINT64);
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_formats ();
  test_query (query, rules);
  gst_query_set_formats (query, 3, GST_FORMAT_TIME, GST_FORMAT_BYTES,
      GST_FORMAT_BUFFERS);
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_segment (GST_FORMAT_TIME);
  test_query (query, rules);
  gst_query_set_segment (query, G_MAXDOUBLE, GST_FORMAT_BYTES, G_MININT64,
      G_MAXINT64);
  test_query (query, rules);
  gst_query_unref (query);

  caps = gst_caps_from_string (non_simple_caps_string);
  query = gst_query_new_caps (caps);
  test_query (query, rules);
  gst_query_set_caps_result (query, caps);
  test_query (query, rules);
  gst_query_unref (query);
  gst_caps_unref (caps);

  caps = gst_caps_new_simple ("video/x-raw",
      "format", G_TYPE_STRING, "I420",
      "width", G_TYPE_INT, 384,
      "height", G_TYPE_INT, 288, "framerate", GST_TYPE_FRACTION, 25, 1, NULL);
  query = gst_query_new_caps (caps);
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_accept_caps (caps);
  test_query (query, rules);
  gst_query_unref (query);
  gst_caps_unref (caps);

  query = gst_query_new_buffering (GST_FORMAT_BYTES);
  test_query (query, rules);
  gst_query_set_buffering_percent (query, TRUE, 5);
  test_query (query, rules);
  gst_query_set_buffering_stats (query, GST_BUFFERING_STREAM, 4, 6, 5432);
  test_query (query, rules);
  gst_query_set_buffering_range (query, GST_FORMAT_BYTES, G_MININT64,
      G_MAXINT64, G_MAXINT64);
  test_query (query, rules);
  /* Next option will set an opaque GArray of internal structures */
  /* If we need this feature in real life we will have to think about */
  /* how we are going to deal with this stuff */
//   gst_query_add_buffering_range (query, 0, 2);
//   test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_uri ();
  test_query (query, rules);
  gst_query_set_uri (query, "http://test/uri");
  test_query (query, rules);
  gst_query_set_uri_redirection (query, "http://test/uri-redirection");
  test_query (query, rules);
  gst_query_set_uri_redirection_permanent (query, TRUE);
  test_query (query, rules);
  gst_query_unref (query);

  caps = gst_caps_from_string (non_simple_caps_string);
//   pool = gst_buffer_pool_new ();
  query = gst_query_new_allocation (caps, TRUE);
  test_query (query, rules);
//   gst_query_add_allocation_pool (query, pool,10, 0, G_MAXUINT);
//   test_query (query, rules);
  gst_query_unref (query);
  gst_caps_unref (caps);
//   gst_object_unref (pool);

  query = gst_query_new_scheduling ();
  test_query (query, rules);
  gst_query_set_scheduling (query, GST_SCHEDULING_FLAG_BANDWIDTH_LIMITED,
      G_MININT, G_MAXINT, 5);
  test_query (query, rules);
  /* Next option will set an opaque GArray of modes. If we need this */
  /* feature in real life we will have to think about how we are */
  /* to deal with this stuff */
//   gst_query_add_scheduling_mode (query, GST_PAD_MODE_NONE);
//   test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_drain ();
  test_query (query, rules);
  gst_query_unref (query);

  query = gst_query_new_context ("test-context");
  test_query (query, rules);
  gst_query_unref (query);
}

GST_START_TEST (marshal_query)
{
  marshal_GstQuery (KURENTO_MARSHALL_BER);
  /* Some tests need a bigger buffer to be encoded using XER */
//   marshal_GstQuery (KURENTO_MARSHALL_XER);
  marshal_GstQuery (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* query_marshalling test suite */
/***************************************/
static Suite *
query_marshalling_suite (void)
{
  Suite *s = suite_create ("query");
  TCase *tc_chain = tcase_create ("marshalling");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, marshal_query);

  return s;
}

GST_CHECK_MAIN (query_marshalling);
