/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include <glib.h>
#include "marshal_types.h"

#define BUFFER_LENGTH (1024 * 6)        /* bytes */

static void
test1_marshal_GstStructure (KurentoMarshalRules rules)
{
  gchar buf[BUFFER_LENGTH];
  GError *err = NULL;
  guint i;

  const gchar *strings[] = {
    "video/x-raw, width = (int) 123456",
    "video/x-raw, stride = (int) -123456",
    "video/x-raw, red_mask = (int) 0xFFFF",
    "video/x-raw, red_mask = (int) 0x0000FFFF",
    "video/x-raw, red_mask = (int) 0x7FFFFFFF",
    "video/x-raw, red_mask = (int) 0x80000000",
    "video/x-raw, red_mask = (int) 0xFF000000, discard=(boolean) TRUE",
    "video/x-raw, format=(string)I420, width=(int)384, height=(int)288, framerate=(fraction)25/1"
  };

  for (i = 0; i < G_N_ELEMENTS (strings); ++i) {
    GstStructure *structure1 = NULL, *structure2 = NULL;
    const gchar *s;
    gsize size;

    s = strings[i];
    structure1 = gst_structure_from_string (s, NULL);

    if (structure1 == NULL) {
      fail ("Could not get structure from string %s", s);
      return;
    }

    /* Encode GVal 1 */
    size = enc_GstStructure (rules, structure1, buf, BUFFER_LENGTH, &err);
    if (err != NULL) {
      GST_ERROR ("%s", err->message);
      fail ("Failed encoding GValue holding an string value");
      gst_structure_free (structure1);
      break;
    }

    /* Encode GVal 2 */
    if (!dec_GstStructure (rules, buf, size, &structure2, &err)) {
      GST_ERROR ("%s", err->message);
      fail ("Failed decoding GValue holding an integer value");
      gst_structure_free (structure1);
      break;
    }

    if (!gst_structure_is_equal (structure1, structure1)) {
      GST_ERROR ("Value encoded and value decoded are not the same");
      fail ("Marshal did not succeed");
    }

    gst_structure_free (structure1);
    gst_structure_free (structure2);
  }

  if (err != NULL)
    g_error_free (err);
}

GST_START_TEST (marshal_gststructure_test1)
{
  test1_marshal_GstStructure (KURENTO_MARSHALL_BER);
  test1_marshal_GstStructure (KURENTO_MARSHALL_XER);
  test1_marshal_GstStructure (KURENTO_MARSHALL_PER);
}

GST_END_TEST static GValue *
get_string_list ()
{
  GValue *list, *val;

  list = g_slice_new0 (GValue);
  g_value_init (list, GST_TYPE_LIST);

  /* Add string values */
  val = g_slice_new0 (GValue);
  g_value_init (val, G_TYPE_STRING);

  g_value_set_static_string (val, NULL);
  gst_value_list_append_value (list, val);

  g_value_set_static_string (val, "");
  gst_value_list_append_value (list, val);

  g_value_set_static_string (val, "hello");
  gst_value_list_append_value (list, val);

  g_value_set_static_string (val, "fdslk'0(9&456:=235476");
  gst_value_list_append_value (list, val);

  g_value_unset (val);
  g_slice_free (GValue, val);

  return list;
}

static GValue *
get_int_list ()
{
  GValue *list, *val;

  list = g_slice_new0 (GValue);
  g_value_init (list, GST_TYPE_LIST);

  /* Add integer values */
  val = g_slice_new0 (GValue);
  g_value_init (val, G_TYPE_INT);

  g_value_set_int (val, 15);
  gst_value_list_append_value (list, val);

  g_value_set_int (val, G_MAXINT);
  gst_value_list_append_value (list, val);

  g_value_set_int (val, G_MININT);
  gst_value_list_append_value (list, val);

  g_value_set_int (val, 0);
  gst_value_list_append_value (list, val);

  g_value_set_int (val, -3452);
  gst_value_list_append_value (list, val);

  g_value_unset (val);
  g_slice_free (GValue, val);

  return list;
}

static GValue *
get_fraction_range_array ()
{
  GValue *array, *val, *min, *max;

  array = g_slice_new0 (GValue);
  g_value_init (array, GST_TYPE_ARRAY);

  /* Add fraction range values */
  min = g_slice_new0 (GValue);
  max = g_slice_new0 (GValue);

  g_value_init (min, GST_TYPE_FRACTION);
  g_value_init (max, GST_TYPE_FRACTION);

  gst_value_set_fraction (min, 6, -G_MAXINT);
  gst_value_set_fraction (max, G_MAXINT, 6);

  val = g_slice_new0 (GValue);
  g_value_init (val, GST_TYPE_FRACTION_RANGE);
  gst_value_set_fraction_range (val, min, max);

  gst_value_array_append_value (array, val);

  gst_value_set_fraction (min, G_MAXINT, -G_MAXINT);
  gst_value_set_fraction (max, 24, 3);
  gst_value_set_fraction_range (val, min, max);
  gst_value_array_append_value (array, val);

  gst_value_set_fraction (min, 1, 5);
  gst_value_set_fraction (max, 9, 3);
  gst_value_set_fraction_range (val, min, max);
  gst_value_array_append_value (array, val);

  g_value_unset (min);
  g_slice_free (GValue, min);

  g_value_unset (max);
  g_slice_free (GValue, max);

  g_value_unset (val);
  g_slice_free (GValue, val);

  return array;
}

static void
test2_marshal_GstStructure (KurentoMarshalRules rules)
{
  GstStructure *structure1 = NULL, *structure2 = NULL;
  gchar buf[BUFFER_LENGTH];
  GError *err = NULL;
  GValue *list;
  gsize size;

  structure1 = gst_structure_new ("test-structure",
      "key", G_TYPE_STRING, "value",
      "bool", G_TYPE_BOOLEAN, TRUE,
      "fraction", GST_TYPE_FRACTION, 1, 5,
      "int", G_TYPE_INT, G_MAXINT,
      "float", G_TYPE_FLOAT, -4568.543,
      "string", G_TYPE_STRING, "video/x-raw, width = (int) 123456",
      "int-range", GST_TYPE_INT_RANGE, G_MININT, G_MAXINT,
      "double-range", GST_TYPE_DOUBLE_RANGE, -123.543, 54326.67,
      "fraction-range", GST_TYPE_FRACTION_RANGE, 1, 4, 6, 8, NULL);

  list = get_string_list ();
  gst_structure_set_value (structure1, "string-list", list);
  g_value_unset (list);
  g_slice_free (GValue, list);

  list = get_int_list ();
  gst_structure_set_value (structure1, "int-list", list);
  g_value_unset (list);
  g_slice_free (GValue, list);

  list = get_fraction_range_array ();
  gst_structure_set_value (structure1, "fraction-range-array", list);
  g_value_unset (list);
  g_slice_free (GValue, list);

  /* Encode GVal 1 */
  size = enc_GstStructure (rules, structure1, buf, BUFFER_LENGTH, &err);
  if (err != NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed encoding GValue holding an string value");
    goto end;
  }

  /* Encode GVal 2 */
  if (!dec_GstStructure (rules, buf, size, &structure2, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GValue holding an integer value");
    goto end;
  }

  if (!gst_structure_is_equal (structure1, structure2)) {
    GST_ERROR ("Value encoded and value decoded are not the same");
    fail ("Marshal did not succeed");
  }

end:
  if (structure1 != NULL)
    gst_structure_free (structure1);

  if (structure2 != NULL)
    gst_structure_free (structure2);
}

GST_START_TEST (marshal_gststructure_test2)
{
  test2_marshal_GstStructure (KURENTO_MARSHALL_BER);
  /* XER encode needs a bigger buffer. */
//   test2_marshal_GstStructure (KURENTO_MARSHALL_XER);
  test2_marshal_GstStructure (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* gststructure_marshalling test suite */
/***************************************/
static Suite *
gststructure_marshalling_suite (void)
{
  Suite *s = suite_create ("gststructure");
  TCase *tc_chain = tcase_create ("marshalling");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, marshal_gststructure_test1);
  tcase_add_test (tc_chain, marshal_gststructure_test2);

  return s;
}

GST_CHECK_MAIN (gststructure_marshalling);
