/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include "marshal_types.h"
#include "marshal_test.h"

#define ENTRIES 2
#define DEEP 4

#define BUFFER_LENGTH (1024 * 30)       /* bytes */

static guint id = 0;

static GstTocEntryType types[] = {
  GST_TOC_ENTRY_TYPE_ANGLE,
  GST_TOC_ENTRY_TYPE_VERSION,
  GST_TOC_ENTRY_TYPE_EDITION,
  GST_TOC_ENTRY_TYPE_INVALID,
  GST_TOC_ENTRY_TYPE_TITLE,
  GST_TOC_ENTRY_TYPE_TRACK,
  GST_TOC_ENTRY_TYPE_CHAPTER
};

static GstTagList *
create_tag_list ()
{
  GstTagList *taglist;
  GstDateTime *datetime;
  GDate *date;
  gchar *title, *name, *album;
  guint n = id++;

  title = g_strdup_printf ("title-%d", n);
  name = g_strdup_printf ("name-%d", n);
  album = g_strdup_printf ("album-%d", n);

  datetime = gst_date_time_new_now_utc ();
  date = g_date_new_dmy (gst_date_time_get_day (datetime),
      gst_date_time_get_month (datetime), gst_date_time_get_year (datetime));

  taglist = gst_tag_list_new_empty ();
  gst_tag_list_add (taglist, GST_TAG_MERGE_APPEND,
      GST_TAG_TITLE, title,
      GST_TAG_TITLE_SORTNAME, name,
      GST_TAG_ALBUM, album,
      GST_TAG_DATE, date,
      GST_TAG_DATE_TIME, datetime,
      GST_TAG_SHOW_SEASON_NUMBER, 2,
      GST_TAG_SHOW_EPISODE_NUMBER, G_MAXUINT,
      GST_TAG_IMAGE_ORIENTATION, "flip-rotate-270",
      GST_TAG_USER_RATING, 10,
      GST_TAG_GEO_LOCATION_HORIZONTAL_ERROR, 3.0, NULL);

  g_free (title);
  g_free (name);
  g_free (album);

  gst_date_time_unref (datetime);
  g_date_free (date);

  return taglist;
}

static GstTocEntry *
create_toc_entry ()
{
  GstTocEntry *entry;
  GstTocEntryType type;
  gchar *uid;
  guint n = id++;

  uid = g_strdup_printf ("uid-%d", n);

  type = types[n % G_N_ELEMENTS (types)];

  entry = gst_toc_entry_new (type, uid);
  gst_toc_entry_set_start_stop_times (entry, n, n + 500);

  if (n % 2 == 0) {
    GstTagList *taglist;

    taglist = create_tag_list ();
    gst_toc_entry_set_tags (entry, taglist);
  }

  g_free (uid);

  return entry;
}

static void
test_empty_toc_GValue (KurentoMarshalRules rules)
{
  GstToc *toc;

  toc = gst_toc_new (GST_TOC_SCOPE_GLOBAL);
  TEST_BASE_CASE (rules, CREATE_SINGLE_GVALUE (GST_TYPE_TOC,
          g_value_set_boxed, toc), BUFFER_LENGTH, 0);
  gst_toc_unref (toc);

  toc = gst_toc_new (GST_TOC_SCOPE_CURRENT);
  TEST_BASE_CASE (rules, CREATE_SINGLE_GVALUE (GST_TYPE_TOC,
          g_value_set_boxed, toc), BUFFER_LENGTH, 1);
  gst_toc_unref (toc);
}

static void
test_toc_GValue (KurentoMarshalRules rules)
{
  GstTagList *taglist;
  GstTocEntry *entry;
  GstToc *toc;

  taglist = create_tag_list ();
  entry = create_toc_entry ();

  gst_toc_entry_set_tags (entry, taglist);

  toc = gst_toc_new (GST_TOC_SCOPE_GLOBAL);
  taglist = create_tag_list ();

  gst_toc_set_tags (toc, taglist);
  gst_toc_append_entry (toc, entry);

  TEST_BASE_CASE (rules, CREATE_SINGLE_GVALUE (GST_TYPE_TOC,
          g_value_set_boxed, toc), BUFFER_LENGTH, 0);

  gst_toc_unref (toc);
}

static void
add_deep_entries (GstTocEntry * entry, guint deep)
{
  guint i;

  if (deep <= 0)
    return;

  for (i = 0; i < ENTRIES; i++) {
    GstTocEntry *e;

    e = create_toc_entry ();
    gst_toc_entry_append_sub_entry (entry, e);
    add_deep_entries (e, deep - 1);
  }
}

static void
test_complex_toc (KurentoMarshalRules rules)
{
  GstToc *toc;
  GstTagList *taglist;
  guint i;

  toc = gst_toc_new (GST_TOC_SCOPE_GLOBAL);
  taglist = create_tag_list (0);
  gst_toc_set_tags (toc, taglist);

  for (i = 0; i < ENTRIES; i++) {
    GstTocEntry *entry;

    entry = create_toc_entry ();
    gst_toc_append_entry (toc, entry);
    add_deep_entries (entry, DEEP);
  }

  TEST_BASE_CASE (rules, CREATE_SINGLE_GVALUE (GST_TYPE_TOC,
          g_value_set_boxed, toc), BUFFER_LENGTH, 0);

  gst_toc_unref (toc);
}

static void
toc_tests (KurentoMarshalRules rules)
{
  test_empty_toc_GValue (rules);
  test_toc_GValue (rules);
  test_complex_toc (rules);
}

GST_START_TEST (marshal_toc)
{
  toc_tests (KURENTO_MARSHALL_BER);
//  toc_tests (KURENTO_MARSHALL_XER);
  toc_tests (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* toc_marshalling test suite */
/***************************************/
static Suite *
toc_marshalling_suite (void)
{
  Suite *s = suite_create ("toc");
  TCase *tc_chain = tcase_create ("marshalling");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, marshal_toc);

  return s;
}

GST_CHECK_MAIN (toc_marshalling);
