/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include <glib.h>
#include "marshal_types.h"

static guint id = 0;

#define BUFFER_LENGTH (1024 * 6)        /* bytes */
#define TEST_BUFFER "buffer for testing purpose"
#define TEST_BUFFER_SIZE strlen(TEST_BUFFER)

static guint
get_uid (void)
{
  return id++;
}

static void
marshal_KmsMessage (KurentoMarshalRules rules, KmsMessage * message)
{
  KmsMessage *m2 = NULL;
  gchar buf[BUFFER_LENGTH];
  GError *err = NULL;
  gsize size;

  size = enc_KmsMessage (rules, message, buf, BUFFER_LENGTH, &err);
  if (err != NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed encoding KmsMessage");
    goto end;
  }

  /* Encode GVal 2 */
  if (!dec_KmsMessage (rules, buf, size, &m2, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding KmsMessage");
    goto end;
  }

  if (!kms_message_is_equal (message, m2)) {
    GST_ERROR ("Value encoded and value decoded are not the same");
    fail ("Marshal did not succeed");
  }

end:
  if (err != NULL)
    g_error_free (err);

  if (m2 != NULL)
    kms_message_unref (m2);
}

static void
test_marshal_KmsMessage (KurentoMarshalRules rules)
{
  KmsMessage *msg;

  msg = kms_message_new (get_uid ());

  kms_message_set_type_counter (msg, 3);

  kms_message_set_fragment_id (msg, 0);
  kms_message_set_total_fragments (msg, 1);
  kms_message_set_message_type (msg, GST_MARSHALL_REQUEST);
  kms_message_set_data (msg, KMS_DATA_TYPE_UNKNOWN, g_strdup (TEST_BUFFER),
      TEST_BUFFER_SIZE, TRUE);

  marshal_KmsMessage (rules, msg);

  /* Check no leaks */
  kms_message_set_data (msg, KMS_DATA_TYPE_UNKNOWN, TEST_BUFFER,
      TEST_BUFFER_SIZE, FALSE);

  marshal_KmsMessage (rules, msg);

  kms_message_unref (msg);
}

GST_START_TEST (test_kmsmessage)
{
  test_marshal_KmsMessage (KURENTO_MARSHALL_BER);
  test_marshal_KmsMessage (KURENTO_MARSHALL_XER);
  test_marshal_KmsMessage (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* kmsmessage_marshalling test suite */
/***************************************/
static Suite *
kmsmessage_marshalling_suite (void)
{
  Suite *s = suite_create ("protocol");
  TCase *tc_chain = tcase_create ("messages");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, test_kmsmessage);

  return s;
}

GST_CHECK_MAIN (kmsmessage_marshalling);
