/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include "marshal_types.h"
#include "kmsfragmenter.h"
#include "kmsassembler.h"

#define TEST1_BUFFER_SIZE 100
#define TEST2_BUFFER_SIZE 150

#define TEST3_BUFFER "buffer for testing purpose"
#define TEST3_BUFFER_SIZE strlen(TEST3_BUFFER)

#define TEST4_BUFFER \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"

static GstEvent *
create_event_caps (void)
{
  GstEvent *event;
  GstCaps *caps;

  caps = gst_caps_new_simple ("video/x-raw",
      "format", G_TYPE_STRING, "I420",
      "width", G_TYPE_INT, 384,
      "height", G_TYPE_INT, 288, "framerate", GST_TYPE_FRACTION, 25, 1, NULL);

  event = gst_event_new_caps (caps);
  gst_caps_unref (caps);

  return event;
}

static KmsFragmenter *
get_fragmented_event (KurentoMarshalRules rules, GError ** err)
{
  KmsFragmenter *fragmenter = NULL;
  GstEvent *event;

  event = create_event_caps ();

  fragmenter = kms_fragmenter_new (rules, TEST1_BUFFER_SIZE);

  kms_fragmenter_set_message_type (fragmenter, GST_MARSHALL_REQUEST);

  if (!kms_fragmenter_event (fragmenter, 0, event, err)) {
    kms_fragmenter_unref (fragmenter);
    fragmenter = NULL;
  }

  gst_event_unref (event);

  return fragmenter;
}

static void
append_message (KmsMessage * msg, KmsAssembler * assembler)
{
  KmsMessage *copy;

  GST_DEBUG ("Adding fragment %u", kms_message_get_fragment_id (msg));
  copy = kms_message_copy (msg);
  kms_assembler_append_message (assembler, copy);
}

static gboolean
unpack_fragmented_event (KmsAssembler * assembler, GError ** err)
{
  GstEvent *event = NULL;
  gchar *buf;
  gsize size;
  gboolean ret;

  kms_assembler_compose_buffer (assembler, &buf, &size);

  if (!(ret = dec_GstEvent (kms_assembler_get_encoding_rules (assembler), buf,
              size, &event, err)))
    goto end;

  GST_DEBUG ("Decoded event %" GST_PTR_FORMAT, event);

end:
  g_free (buf);
  gst_event_unref (event);

  return ret;
}

static void
test_event_KmsAssembler (KurentoMarshalRules rules)
{
  KmsAssembler *assembler = NULL;
  KmsFragmenter *fragmenter = NULL;
  GError *err = NULL;

  fragmenter = get_fragmented_event (rules, &err);
  if (fragmenter == NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed fragmenting event");
    goto end;
  }

  assembler = kms_assembler_new (rules);
  kms_fragmenter_for_each_message (fragmenter, (GFunc) append_message,
      assembler);
  if (!kms_assembler_is_completed (assembler)) {
    fail ("Assembler did not got all fragments");
    goto end;
  }

  if (kms_assembler_get_data_type (assembler) != KMS_DATA_TYPE_EVENT) {
    fail ("Assembler does not hold a GstEvent type");
    goto end;
  }

  if (!unpack_fragmented_event (assembler, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed unpacking event");
  }

end:
  if (err != NULL)
    g_error_free (err);

  if (fragmenter != NULL)
    kms_fragmenter_unref (fragmenter);

  if (assembler != NULL)
    kms_assembler_unref (assembler);
}

GST_START_TEST (test_assembler_event)
{
  test_event_KmsAssembler (KURENTO_MARSHALL_BER);
  test_event_KmsAssembler (KURENTO_MARSHALL_PER);
}

GST_END_TEST static GstQuery *
create_query_caps (void)
{
  GstQuery *query;
  GstCaps *caps;
  gchar non_simple_caps_string[] =
      "video/x-raw, format=(string)I420, framerate=(fraction)[ 1/100, 100 ], "
      "width=(int)[ 16, 4096 ], height=(int)[ 16, 4096 ]; video/x-raw, "
      "format=(string)YUY2, framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
      "height=(int)[ 16, 4096 ]; video/x-raw, format=(string)RGB8_PALETTED, "
      "framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
      "height=(int)[ 16, 4096 ]; video/x-raw, "
      "format=(string){ I420, YUY2, YV12 }, width=(int)[ 16, 4096 ], "
      "height=(int)[ 16, 4096 ], framerate=(fraction)[ 1/100, 100 ]";

  caps = gst_caps_from_string (non_simple_caps_string);
  query = gst_query_new_caps (caps);

  gst_caps_unref (caps);

  return query;
}

static KmsFragmenter *
get_fragmented_query (KurentoMarshalRules rules, GError ** err)
{
  KmsFragmenter *fragmenter = NULL;
  GstQuery *query;

  query = create_query_caps ();

  fragmenter = kms_fragmenter_new (rules, TEST2_BUFFER_SIZE);

  kms_fragmenter_set_message_type (fragmenter, GST_MARSHALL_REQUEST);

  if (!kms_fragmenter_query (fragmenter, 0, query, err)) {
    kms_fragmenter_unref (fragmenter);
    fragmenter = NULL;
  }

  gst_query_unref (query);

  return fragmenter;
}

static gboolean
unpack_fragmented_query (KmsAssembler * assembler, GError ** err)
{
  GstQuery *query = NULL;
  gchar *buf;
  gsize size;
  gboolean ret = FALSE;

  kms_assembler_compose_buffer (assembler, &buf, &size);

  if (!(ret = dec_GstQuery (kms_assembler_get_encoding_rules (assembler), buf,
              size, &query, err)))
    goto end;

  GST_DEBUG ("Decoded query %" GST_PTR_FORMAT, query);

end:
  g_free (buf);
  gst_query_unref (query);

  return ret;
}

static void
test_query_KmsAssembler (KurentoMarshalRules rules)
{
  KmsAssembler *assembler = NULL;
  KmsFragmenter *fragmenter = NULL;
  GError *err = NULL;

  fragmenter = get_fragmented_query (rules, &err);
  if (fragmenter == NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed fragmenting event");
    goto end;
  }

  assembler = kms_assembler_new (rules);
  kms_fragmenter_for_each_message (fragmenter, (GFunc) append_message,
      assembler);
  if (!kms_assembler_is_completed (assembler)) {
    fail ("Assembler did not got all fragments");
    goto end;
  }

  if (kms_assembler_get_data_type (assembler) != KMS_DATA_TYPE_QUERY) {
    fail ("Assembler does not hold a GstQuery type");
    goto end;
  }

  if (!unpack_fragmented_query (assembler, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed unpacking event");
  }

end:
  if (err != NULL)
    g_error_free (err);

  if (fragmenter != NULL)
    kms_fragmenter_unref (fragmenter);

  if (assembler != NULL)
    kms_assembler_unref (assembler);
}

GST_START_TEST (test_assembler_query)
{
  test_query_KmsAssembler (KURENTO_MARSHALL_BER);
  test_query_KmsAssembler (KURENTO_MARSHALL_PER);
}

GST_END_TEST static gboolean
unpack_fragmented_buffer (KmsAssembler * assembler, GError ** err)
{
  GstBuffer *buffer = NULL;
  gchar *buf = NULL;
  gsize size;
  gboolean ret = FALSE;

  kms_assembler_compose_buffer (assembler, &buf, &size);

  if (!(ret =
          dec_GstBuffer (kms_assembler_get_encoding_rules (assembler), buf,
              size, &buffer, err))) {
    goto end;
  }

  GST_DEBUG ("Decoded buffer %" GST_PTR_FORMAT, buffer);

  ret =
      (gst_buffer_memcmp (buffer, 0, TEST4_BUFFER, strlen (TEST4_BUFFER)) == 0);

end:
  g_free (buf);

  if (buffer != NULL)
    gst_buffer_unref (buffer);

  return ret;
}

static KmsFragmenter *
get_fragmented_buffer (KurentoMarshalRules rules, GError ** err)
{
  KmsFragmenter *fragmenter = NULL;
  GstBuffer *buf;
  gsize size;

  size = strlen (TEST4_BUFFER);
  buf = gst_buffer_new_wrapped (g_memdup (TEST4_BUFFER, size), size);

  fragmenter = kms_fragmenter_new (rules, TEST2_BUFFER_SIZE);
  kms_fragmenter_set_message_type (fragmenter, GST_MARSHALL_REQUEST);

  if (!kms_fragmenter_buffer (fragmenter, 0, buf, err)) {
    kms_fragmenter_unref (fragmenter);
    fragmenter = NULL;
  }

  gst_buffer_unref (buf);

  return fragmenter;
}

static void
test_buffer_KmsAssembler (KurentoMarshalRules rules)
{
  KmsAssembler *assembler = NULL;
  KmsFragmenter *fragmenter = NULL;
  GError *err = NULL;

  fragmenter = get_fragmented_buffer (rules, &err);
  if (fragmenter == NULL) {
    GST_ERROR ("%s", err->message);
    fail ("Failed fragmenting buffer");
    goto end;
  }

  assembler = kms_assembler_new (rules);
  kms_fragmenter_for_each_message (fragmenter, (GFunc) append_message,
      assembler);
  if (!kms_assembler_is_completed (assembler)) {
    fail ("Assembler did not got all fragments");
    goto end;
  }

  if (kms_assembler_get_data_type (assembler) != KMS_DATA_TYPE_BUFFER) {
    fail ("Assembler does not hold a GstBuffer type");
    goto end;
  }

  if (!unpack_fragmented_buffer (assembler, &err)) {
    fail ("Failed unpacking buffer");
  }

end:
  if (err != NULL)
    g_error_free (err);

  if (fragmenter != NULL)
    kms_fragmenter_unref (fragmenter);

  if (assembler != NULL)
    kms_assembler_unref (assembler);
}

GST_START_TEST (test_assembler_buffer)
{
  test_buffer_KmsAssembler (KURENTO_MARSHALL_BER);
  test_buffer_KmsAssembler (KURENTO_MARSHALL_PER);
}

GST_END_TEST static void
test_wrong_KmsAssembler (KurentoMarshalRules rules)
{
  KmsAssembler *assembler = NULL;
  KmsMessage *msg1, *msg2, *msg3, *msg4 = NULL;

  msg1 = kms_message_new (0);
  kms_message_set_fragment_id (msg1, 0);
  kms_message_set_total_fragments (msg1, 3);
  kms_message_set_message_type (msg1, GST_MARSHALL_REQUEST);
  kms_message_set_data (msg1, KMS_DATA_TYPE_UNKNOWN, g_strdup (TEST3_BUFFER),
      TEST3_BUFFER_SIZE, TRUE);

  msg2 = kms_message_new (0);
  kms_message_set_fragment_id (msg2, 1);
  kms_message_set_total_fragments (msg2, 3);
  kms_message_set_message_type (msg2, GST_MARSHALL_REQUEST);
  kms_message_set_data (msg2, KMS_DATA_TYPE_UNKNOWN, g_strdup (TEST3_BUFFER),
      TEST3_BUFFER_SIZE, TRUE);

  msg3 = kms_message_new (0);
  kms_message_set_fragment_id (msg3, 2);
  kms_message_set_total_fragments (msg3, 3);
  kms_message_set_message_type (msg3, GST_MARSHALL_REQUEST);
  kms_message_set_data (msg3, KMS_DATA_TYPE_UNKNOWN, g_strdup (TEST3_BUFFER),
      TEST3_BUFFER_SIZE, TRUE);

  msg4 = kms_message_new (1);
  kms_message_set_fragment_id (msg4, 2);
  kms_message_set_total_fragments (msg4, 3);
  kms_message_set_message_type (msg4, GST_MARSHALL_REQUEST);
  kms_message_set_data (msg4, KMS_DATA_TYPE_UNKNOWN, g_strdup (TEST3_BUFFER),
      TEST3_BUFFER_SIZE, TRUE);

  assembler = kms_assembler_new (rules);
  if (kms_assembler_append_message (assembler, msg2)) {
    fail ("Fragment 2 can't be added before fragment 1");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (!kms_assembler_append_message (assembler, msg1)) {
    fail ("Fragment 1 can't be added");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (kms_assembler_is_completed (assembler)) {
    fail ("Assembler should not be completed yet");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (kms_assembler_append_message (assembler, msg3)) {
    fail ("Fragment 3 can't be added before fragment 2");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (kms_assembler_is_completed (assembler)) {
    fail ("Assembler should not be completed yet");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (!kms_assembler_append_message (assembler, msg2)) {
    fail ("Fragment 2 can't be added");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (kms_assembler_is_completed (assembler)) {
    fail ("Assembler should not be completed yet");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (kms_assembler_append_message (assembler, msg4)) {
    fail ("Fragment 4 belongs to other request");
    kms_message_unref (msg3);
    goto end;
  }

  if (!kms_assembler_append_message (assembler, msg3)) {
    fail ("Fragment 3 can't be added");
    kms_message_unref (msg1);
    kms_message_unref (msg2);
    kms_message_unref (msg3);
    goto end;
  }

  if (!kms_assembler_is_completed (assembler)) {
    fail ("Assembler should be completed");
    goto end;
  }

end:
  if (assembler != NULL)
    kms_assembler_unref (assembler);

  if (msg4 != NULL)
    kms_message_unref (msg4);
}

GST_START_TEST (test_assembler_wrong_cases)
{
  test_wrong_KmsAssembler (KURENTO_MARSHALL_BER);
  test_wrong_KmsAssembler (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* kmsassembler_tests test suite */
/***************************************/
static Suite *
kmsassembler_tests_suite (void)
{
  Suite *s = suite_create ("assembler");
  TCase *tc_chain = tcase_create ("assembler");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, test_assembler_event);
  tcase_add_test (tc_chain, test_assembler_query);
  tcase_add_test (tc_chain, test_assembler_buffer);
  tcase_add_test (tc_chain, test_assembler_wrong_cases);

  return s;
}

GST_CHECK_MAIN (kmsassembler_tests);
