/*
 * (C) Copyright 2014 Kurento (http://kurento.org/)
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 */

#include <string.h>
#include <gst/check/gstcheck.h>
#include <gst/gst.h>
#include <glib.h>
#include "marshal_types.h"
#include "kmsfragmenter.h"

#define MAX_BUFFER 150
#define TEST_BUFFER \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \
  "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"

struct composed_data
{
  gchar *buf;
  guint size;
  guint index;
};

static void
append_data (KmsMessage * msg, struct composed_data *data)
{
  KmsDataType type;
  const gchar *buf;
  gsize size;

  kms_message_get_data (msg, &type, &buf, &size);
  memcpy (data->buf + data->index, buf, size);
  data->index += size;
}

static void
check_fragmented_event (KmsFragmenter * fragmenter)
{
  struct composed_data data;
  GError *err = NULL;
  GstEvent *event = NULL;

  data.size = kms_fragmenter_coded_data_size (fragmenter);
  data.buf = g_malloc (data.size);
  data.index = 0;

  kms_fragmenter_for_each_message (fragmenter, (GFunc) append_data, &data);
  if (!dec_GstEvent (kms_fragmenter_get_encoding_rules (fragmenter), data.buf,
          data.size, &event, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GstEvent");
    g_error_free (err);
  } else {
    GST_DEBUG ("Decoded event %" GST_PTR_FORMAT, event);
  }

  g_free (data.buf);

  if (event != NULL)
    gst_event_unref (event);
}

static void
test_event_KmsFragmenter (KurentoMarshalRules rules)
{
  KmsFragmenter *fragmenter = NULL;
  GError *err = NULL;
  GstEvent *event;
  GstCaps *caps;

  caps = gst_caps_new_simple ("video/x-raw",
      "format", G_TYPE_STRING, "I420",
      "width", G_TYPE_INT, 384,
      "height", G_TYPE_INT, 288, "framerate", GST_TYPE_FRACTION, 25, 1, NULL);
  event = gst_event_new_caps (caps);

  fragmenter = kms_fragmenter_new (rules, MAX_BUFFER);

  kms_fragmenter_set_message_type (fragmenter, GST_MARSHALL_REQUEST);

  if (!kms_fragmenter_event (fragmenter, 0, event, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed fragmenting event");
    goto end;
  }

  fail_unless (GST_EVENT_IS_SERIALIZED (event) ==
      kms_fragmenter_is_serialized (fragmenter));

  check_fragmented_event (fragmenter);

end:
  if (err != NULL)
    g_error_free (err);

  if (fragmenter != NULL)
    kms_fragmenter_unref (fragmenter);

  gst_event_unref (event);
  gst_caps_unref (caps);
}

GST_START_TEST (test_kmsfragmenter_event)
{
  test_event_KmsFragmenter (KURENTO_MARSHALL_BER);
  test_event_KmsFragmenter (KURENTO_MARSHALL_PER);
}

GST_END_TEST static void
check_fragmented_query (KmsFragmenter * fragmenter)
{
  struct composed_data data;
  GError *err = NULL;
  GstQuery *query = NULL;

  data.size = kms_fragmenter_coded_data_size (fragmenter);
  data.buf = g_malloc (data.size);
  data.index = 0;

  kms_fragmenter_for_each_message (fragmenter, (GFunc) append_data, &data);
  if (!dec_GstQuery (kms_fragmenter_get_encoding_rules (fragmenter), data.buf,
          data.size, &query, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GstQuery");
    g_error_free (err);
  } else {
    GST_DEBUG ("Decoded query %" GST_PTR_FORMAT, query);
  }

  g_free (data.buf);

  if (query != NULL)
    gst_query_unref (query);
}

static void
test_query_KmsFragmenter (KurentoMarshalRules rules)
{
  KmsFragmenter *fragmenter = NULL;
  GError *err = NULL;
  GstQuery *query;
  GstCaps *caps;
  const gchar non_simple_caps_string[] =
      "video/x-raw, format=(string)I420, framerate=(fraction)[ 1/100, 100 ], "
      "width=(int)[ 16, 4096 ], height=(int)[ 16, 4096 ]; video/x-raw, "
      "format=(string)YUY2, framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
      "height=(int)[ 16, 4096 ]; video/x-raw, format=(string)RGB8_PALETTED, "
      "framerate=(fraction)[ 1/100, 100 ], width=(int)[ 16, 4096 ], "
      "height=(int)[ 16, 4096 ]; video/x-raw, "
      "format=(string){ I420, YUY2, YV12 }, width=(int)[ 16, 4096 ], "
      "height=(int)[ 16, 4096 ], framerate=(fraction)[ 1/100, 100 ]";

  caps = gst_caps_from_string (non_simple_caps_string);
  query = gst_query_new_caps (caps);

  fragmenter = kms_fragmenter_new (rules, MAX_BUFFER);

  kms_fragmenter_set_message_type (fragmenter, GST_MARSHALL_REQUEST);

  if (!kms_fragmenter_query (fragmenter, 0, query, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed fragmenting query");
    goto end;
  }

  fail_unless (GST_QUERY_IS_SERIALIZED (query) ==
      kms_fragmenter_is_serialized (fragmenter));

  check_fragmented_query (fragmenter);

end:
  if (err != NULL)
    g_error_free (err);

  if (fragmenter != NULL)
    kms_fragmenter_unref (fragmenter);

  gst_query_unref (query);
  gst_caps_unref (caps);
}

GST_START_TEST (test_kmsfragmenter_query)
{
  test_query_KmsFragmenter (KURENTO_MARSHALL_BER);
  test_query_KmsFragmenter (KURENTO_MARSHALL_PER);
}

GST_END_TEST static void
check_fragmented_buffer (KmsFragmenter * fragmenter)
{
  struct composed_data data;
  GError *err = NULL;
  GstBuffer *buffer = NULL;

  data.size = kms_fragmenter_coded_data_size (fragmenter);
  data.buf = g_malloc (data.size);
  data.index = 0;

  kms_fragmenter_for_each_message (fragmenter, (GFunc) append_data, &data);
  if (!dec_GstBuffer (kms_fragmenter_get_encoding_rules (fragmenter), data.buf,
          data.size, &buffer, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed decoding GstBuffer");
    goto done;
  } else {
    GST_DEBUG ("Decoded buffer %" GST_PTR_FORMAT, buffer);
  }

  if (gst_buffer_memcmp (buffer, 0, TEST_BUFFER, strlen (TEST_BUFFER)) != 0) {
    fail ("Buffer coded and decoded has not the same content");
  }

done:
  g_free (data.buf);

  if (err != NULL)
    g_error_free (err);

  if (buffer != NULL)
    gst_buffer_unref (buffer);
}

static void
test_buffer_KmsFragmenter (KurentoMarshalRules rules)
{
  KmsFragmenter *fragmenter = NULL;
  GError *err = NULL;
  GstBuffer *buf;
  gsize size;

  size = strlen (TEST_BUFFER);
  buf = gst_buffer_new_wrapped (g_memdup (TEST_BUFFER, size), size);

  fragmenter = kms_fragmenter_new (rules, MAX_BUFFER);
  kms_fragmenter_set_message_type (fragmenter, GST_MARSHALL_REQUEST);

  if (!kms_fragmenter_buffer (fragmenter, 0, buf, &err)) {
    GST_ERROR ("%s", err->message);
    fail ("Failed fragmenting buffer");
    goto end;
  }

  GST_DEBUG ("Buffer fragmented in %d messages",
      kms_fragmenter_n_messages (fragmenter));

  check_fragmented_buffer (fragmenter);

end:
  if (err != NULL)
    g_error_free (err);

  if (fragmenter != NULL)
    kms_fragmenter_unref (fragmenter);

  gst_buffer_unref (buf);
}

GST_START_TEST (test_kmsfragmenter_buffer)
{
  test_buffer_KmsFragmenter (KURENTO_MARSHALL_BER);
  test_buffer_KmsFragmenter (KURENTO_MARSHALL_PER);
}

GST_END_TEST
/***************************************/
/* kmsfragmenter_tests test suite */
/***************************************/
static Suite *
kmsfragmenter_tests_suite (void)
{
  Suite *s = suite_create ("fragmenter");
  TCase *tc_chain = tcase_create ("fragmenter");

  suite_add_tcase (s, tc_chain);

  tcase_add_test (tc_chain, test_kmsfragmenter_event);
  tcase_add_test (tc_chain, test_kmsfragmenter_query);
  tcase_add_test (tc_chain, test_kmsfragmenter_buffer);

  return s;
}

GST_CHECK_MAIN (kmsfragmenter_tests);
